/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** \file Pmn_construct.c

\brief Compute point values of P^m_n Associated Legendre Functions.

Summary:
- Pmncos evaluates a normalized P^m_n(cos(phi)).
- Pmncos_resave_init initializes for bulk value computation.
- next_rtsinPmncos does one step of the bulk value computation.
- Pmncos_resave_destroy frees memory used by Pmncos_resave_init.

*/

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

#ifndef LNOF2
#define LNOF2 0.693147180559945309417232121458
#endif

REAL Pmncos(int order,
	    int index,
	    REAL phi
	    ){
  /** Evaluate an associated Legendre function P^m_n at any point.
 
  This routine is super stabilized for all conditions.

  INPUTS:  
  - order -- the order of the associated legendre function you want.
    In P^m_n notation this is m. order>=0
  - index -- Which frequency to use, starting from the lowest legal.
    In P^m_n notation, n is m+index.  index >=0
  - phi -- The point at which to evaluate. 0< phi < PI
  
  OUTPUTS:  returns \\(cP^m_n(\\cos(\\phi))\\) with c to make L^2 normalized.

                                      m
                                     P (cos(phi)) * const
		                      n

  The constant is set so we are L^2 normalized.
  Since the number of points to be used is not known, we cannot
  adjust for the discretization.

   NOTES: 
   - We pass type REAL to several functions which expect double.
     This should not cause trouble but still might.
   - Just like dbrtsinPmncos except for sqrt(sin) factor.

*/
  REAL oneback,twoback;
  int s,i,sizetemp,size;
  REAL sofar[2],ex,why;

#if ENTEREXIT
  printf("entering Pmncos\n");
#endif

  /* initialize */
  /* here we generate the correct starting constant (for P^m_m).
  \sqrt{\Gamma(2 order+2) \over 2} {1 \over 2^{order}\Gamma(order+1)}
  ={1\over \sqrt{2}}\prod_{j=1}^{m} \sqrt{1 + {1\over 2j}} */
  twoback=1.0;
  for(s=1;s<=order;s++)
    twoback*=1.0 + 0.5/s;
  twoback=sqrt(twoback/2);

  /* now initialize ex, why, and the recurrence: sofar*/
  ex=cos(phi);
  why=sin(phi);
  /* We are going to be very careful in our representation of the values of our
   functions. Although the recurrence is stable, it is prone to overflows
   and underflows. Rather than let the type REAL try to deal with it, we
   remove a power of 2, leaving us with values of size 1.  The exponent
   is stored in  size .*/

  /* this gives us our sin() to the power order *//* 1/2 removed **/
  /* here is the size (nearest power of 2) of the power of sin*/
  if(why==0){
    if(order==0){
      sofar[0]=twoback;
      size=0;
    }else
      return 0.0;    	
  }else{/* okay to log */
    size= (int) (order)* log(why) /LNOF2;/* 1/2 removed **/
    /*here is the remnant, including coefficient*/
    sofar[0]=twoback * exp((order)*log(why) -size*LNOF2);/* 1/2 removed **/
    /* re-extract the size, probably unecessary*/
    sofar[0]=frexp(sofar[0],&sizetemp);
    size+=sizetemp;
  }

  if(index==0)              /* index==0 we are done */
    return ldexp(sofar[0],size);

  /* another special case, requires some initialization*/
  /* this is the oneback recurrence coefficient */
  oneback=2.0*sqrt((0.5+order)*(0.5+1.0/(2.0*order+1.0)));
  sofar[1]=oneback* sofar[0]*ex;

  /* This renormalizes so that, on average we have size 1 */
  frexp(fabs(sofar[0])+fabs(sofar[1]),&sizetemp);
  size+=sizetemp;
  sofar[0]=ldexp(sofar[0],-sizetemp);
  sofar[1]=ldexp(sofar[1],-sizetemp);

  if(index==1)              /* index==1 we are done */
    return ldexp(sofar[1],size);

  /* general case */
  for(i=2;i<=index;i++){
    /* This is the main recurrence.  First we compute the coefficients,
       then compute and load in the values */
    oneback=2*sqrt((1.0+(order-0.5)/i)*(1.0-(order-0.5)/(2.0*order+i)));
    twoback=-sqrt((1-1.0/i)*(1-1.0/(2*order+i))
		  *(1+4.0/(2*i+2*order-3)));
    sofar[i%2]=oneback*ex*sofar[(i-1)%2]+twoback*sofar[(i-2)%2];
    /* This renormalizes so that, on average we have size 1 */
    frexp(fabs(sofar[0])+fabs(sofar[1]),&sizetemp);
    size+=sizetemp;
    sofar[0]=ldexp(sofar[0],-sizetemp);
    sofar[1]=ldexp(sofar[1],-sizetemp);
  }

#if ENTEREXIT
  printf("returning from Pmncos\n");
#endif

  return ldexp(sofar[index%2],size);
}

void Pmncos_resave_init(Pmncos_Recurrence_Save *pmn_save,
			REAL *node,
			int numpts
			){
  /** Converts a set of nodes (often gaussian) to the initialization needed 
      for the associated legendre function recurrence.

      INPUTS: 
      - pmn_save -- a structure to hold the outputs
      - node -- and array holding the sampling nodes.
        The array is of length (numpts+1)/2 and the values
	should be in [0,PI/2]. We assume the sampling on
	(PI/2,PI] is symmetric
      - numpts -- the number of sample points. We use half of these
        (plus 1 if odd)

      OUTPUTS : This function creates memory for the arrays in pmn_save
      and initializes some values.
  
  */

  
  register int i,hemipts;

#if ENTEREXIT
  printf("entering Pmncos_resave_init\n");
#endif

  /* load numpts into structure */
  pmn_save->numpts=numpts;

  /* find half (rounding up) the number of points */
  hemipts=(numpts+1)/2;

  /* Initialize memory **********/
  
  /* first dimension, length 2 */
  pmn_save->sofar=(REAL **)calloc(2,sizeof(REAL *));
  assert(pmn_save->sofar != NULL);

  /* main memory: 2*hemipts for sofar, hemipts for ex, hemipts for why */
  pmn_save->sofar[0]=(REAL *)calloc(4*hemipts,sizeof(REAL));
  assert(pmn_save->sofar[0] != NULL);

  /* second row, leave hemipts for first */
  pmn_save->sofar[1]=pmn_save->sofar[0]+hemipts;

  /* ex, leaving hemipts for sofar[1] */
  pmn_save->ex=pmn_save->sofar[1]+hemipts;

  /* why, leaving hemipts for ex */
  pmn_save->why=pmn_save->ex+hemipts;

  /* integer array length hemipts */
  pmn_save->size=(int *)calloc(hemipts,sizeof(int));
  assert(pmn_save->size !=NULL);

  /* now initialize ex=cos(node) and why=sin(node)*/

  for(i=0;i<hemipts;i++){
    pmn_save->ex[i]=cos(node[i]);
    pmn_save->why[i]=sin(node[i]);
  }

#if ENTEREXIT
  printf("exiting Pmncos_resave_init\n");
#endif

}

void Pmncos_resave_destroy(Pmncos_Recurrence_Save *pmn_save
			   ){
  /** destroy (free) a structure Pmncos_Recurrence_Save

  INPUTS: pmn_save -- the structure to free

  OUTPUTS : The memory held by pmn_save is freed
  
  */

#if ENTEREXIT
  printf("entering Pmncos_resave_destroy\n");
#endif

  /* Free memory **********/
  
  /* main REAL memory. All the others are actually part of this memory */
  free(pmn_save->sofar[0]);

  /* first dimension of sofar */
  free(pmn_save->sofar);

  /* separate integer array */
  free(pmn_save->size);

#if ENTEREXIT
  printf("exiting Pmncos_resave_destroy\n");
#endif

}

void next_rtsinPmncos(REAL *out,
		      Pmncos_Recurrence_Save *pmn_save,
		      int halfpts,
		      int order,
		      int index
		      ){
  /** Do one step in the recurrence process then output the results.  

  This saves on memory if we are trying to compress all associated
  Legendre fcns of some order.  This version is super stabilized for
  all conditions.

  INPUTS: 
  - out-- the results will be put in this array of length halfpts, in format 
    out[i]=C^m_n P^m_n(pmn_save->ex[i])*sqrt(pmn_save->why[i]).
    Generally pmn_save->ex[i] is cos(quadrature_point[i])
    pmn_save->ex[i] is sin(quadrature_point[i])
  - pmn_save -- a structure that holds some precomputed trig values
    and the results of previous recurrences.
    Must be initialized with Pmncos_resave_init.
  - halfpts-- the number of points to use. Generally this is the
    number of points in [0,PI/2] and we use parity for
    [PI/2,PI]. This routine does not care, however,
    because the location of the points is encoded in pmn_save.
  - order -- the order of the associated legendre function desired.
    In P^m_n notation this is m
  - index--- the index desired minus m. In P^m_{m+k} notation this is k.

  OUTPUTS : results are output in outfcn, above
  
  USAGE: You must call this function for successive values of
  index, starting at 0.  Order must remain fixed.

  NOTES:  
  - We pass type REAL to several functions which expect double.
    This should not cause trouble but still might.
  - Is this the correct norm fix ??   
  */
  int s,sizetemp;
  /* keep these as double for safety */
  double oneback,twoback;

#if ENTEREXIT
  printf("entering next_rtsinPmncos\n");
#endif

  if(index==0){
     
    /* here we generate the correct starting constant (for P^m_m).
       \sqrt{\Gamma(2 order+2) \over 2} {1 \over 2^{order}\Gamma(order+1)}
       ={1\over \sqrt{2}}\prod_{j=1}^{m} \sqrt{1 + {1\over 2j}} */
    twoback=1.0;
    for(s=1;s<=order;s++)
      twoback*=1.0 + 0.5/s;
    twoback=sqrt(twoback/2);
#if 0
    /* now part of the gaussian weights */

    /* since we are on [0,PI], fix the norm */
    twoback*=sqrt(PI);
    /* fix norm for discritization */
    twoback/=sqrt(2*halfpts);
#endif
    /* now initialize the recurrence: sofar*/
    /* We are going to be very careful in our representation of the
       values of our functions. Although the recurrence is stable, it is
       prone to overflows and underflows. Rather than let the type REAL
       try to deal with it, we remove a power of 2, leaving us with values
       of size 1.  The exponent is stored in size .*/
    for(s=0;s<halfpts;s++){
      /* loop through the points */

      /* this gives us our sin() to the power order +1/2 */
      /* here is the size (nearest power of 2) of the power of sin*/
      (pmn_save->size)[s]= (int) (order+0.5)* log((pmn_save->why)[s]) /LNOF2;

      /*here is the remnant, including coefficient*/
      (pmn_save->sofar)[0][s]=twoback 
	* exp((order+0.5)*log((pmn_save->why)[s]) -(pmn_save->size)[s]*LNOF2);

      /* re-extract the size, probably unecessary*/
      (pmn_save->sofar)[0][s]=frexp((pmn_save->sofar)[0][s],&sizetemp);
      (pmn_save->size)[s]+=sizetemp;

      /* now output this */
      out[s]=ldexp((pmn_save->sofar)[0][s],(pmn_save->size)[s]);
    }
  }
  
  else if(index==1){  /* another special case, requires some initialization*/

    /* this is the oneback recurrence coefficient */
    oneback=2.0*sqrt((0.5+order)*(0.5+1.0/(2.0*order+1.0)));

    for(s=0;s<halfpts;s++){
      /* loop through the points */

      /* main recurrence step, but only index 1 */
      (pmn_save->sofar)[1][s]=
	oneback*(pmn_save->sofar)[0][s]*(pmn_save->ex)[s];

      /* This renormalizes so that, on average we have size 1 */
      frexp(fabs((pmn_save->sofar)[0][s])+fabs((pmn_save->sofar)[1][s]),
	    &sizetemp);
      (pmn_save->size)[s]+=sizetemp;
      (pmn_save->sofar)[0][s]=ldexp((pmn_save->sofar)[0][s],-sizetemp);
      (pmn_save->sofar)[1][s]=ldexp((pmn_save->sofar)[1][s],-sizetemp);
      
      /* now output this */
      out[s]=ldexp((pmn_save->sofar)[1][s],(pmn_save->size)[s]);
    }
  }

  else{         /* general case */
 /* This is the main recurrence.  First we compute the coefficients,
     then compute and load in the values for each point*/

    oneback=2*sqrt((1.0+(order-0.5)/index)
		   *(1.0-(order-0.5)/(2.0*order+index)));
    twoback=-sqrt((1-1.0/index)*(1-1.0/(2*order+index))
		 *(1+4.0/(2*index+2*order-3)));

    for(s=0;s<halfpts;s++){
      /* loop through the points */

      /* main recurrence step */
      (pmn_save->sofar)[index%2][s]=
	oneback*(pmn_save->ex)[s]*(pmn_save->sofar)[(index-1)%2][s]
	+twoback*(pmn_save->sofar)[(index-2)%2][s];

      /* This renormalizes so that, on average we have size 1 */
      frexp(fabs((pmn_save->sofar)[0][s])+fabs((pmn_save->sofar)[1][s]),
	    &sizetemp);
      (pmn_save->size)[s]+=sizetemp;
      (pmn_save->sofar)[0][s]=ldexp((pmn_save->sofar)[0][s],-sizetemp);
      (pmn_save->sofar)[1][s]=ldexp((pmn_save->sofar)[1][s],-sizetemp);

      /* now output this*/
      out[s]=ldexp((pmn_save->sofar)[index%2][s],(pmn_save->size)[s]);
    }
  }
  
#if ENTEREXIT
  printf("exiting next_rtsinPmncos\n");
#endif

}




  
