/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file analyse_d1.c

\brief Analyze (expand) using the direct1 algorithm.

This program does the analysis (expansion)) portion of the direct
transform for spherical harmonics, using the plan created by
initialize_d1

This is essentially a demo version, since in a real program one would
not want to read/write to files this often.

We only do the Associated Legendre Transform part of the SH transform.
To do a full transform, you also need to fourier transform in
longitude.

Usage:
call this routine with 3 command-line arguments:
d1_file fcn_file out_file
- d1_file -- The file created by initialize_d1, 
  for the size problem you want
- fcn_file -- the file from which to read the function values. 
  These are not true function values, because we do not do the
  fourier transform in longitude. The ordering is 
  m= 0, 1, -1, 2, -2 ...
  For a given m, there are numpts REAL values, in binary.
  (numpts is the number of latitude points, read from f1_file.)
  - out_file -- The file in which to put the SH coefficients. 
    They must agree with the size determined by band_limit,
    which is read from f1_file. The ordering is 
    m= 0, 1, -1, 2, -2 ...
    For a given m, there are (band_limit -abs(m)) REAL values,
    in binary.
	
OPTIONS:
set DOTIME to 1 to turn on timing 
    
**/
#define DOTIME 1

#include "../libftsh.h"

/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
int main(int argc,
          char *argv[]
          ){


  register int m;
  int numpts,band_limit,order_limit;
  REAL *node,*weight;
  FILE *fp;
  Pm_Direct_d *pmn_matrix;
  REAL **hat, **m_data;
  Pm_1D_Workspace pmnwork;
#if DOTIME
  REAL prealg;
#endif

  /* make sure the correct number of arguments were given
   (program name counts as one) */
  assert(argc==4);

  /********************************************************/
  /* load in d1_file data **/
  fp=fopen(argv[1],"rb");
  assert(fp != NULL);

  /* note what we are doing */
  printf("%s : reading initialization from %s\n",argv[0],argv[1]);

  /* load precomputed initialization and set up workspace and such */
  direct1_fromfile(&numpts,&band_limit,&node,&weight,
		   &pmn_matrix,&pmnwork,fp);

  /* close file */
  fclose(fp);
  /********************************************************/


  /********************************************************/
  /* read in value data ***/
  fp=fopen(argv[2],"rb");
  assert(fp != NULL);

  /* note what we are doing */
  printf("band_limit=%i, numpts=%i, reading values from %s\n",
	 band_limit,numpts,argv[2]);

  /* set order_limit */
  order_limit=band_limit; 

  /* holder for the function data */
  m_data=(REAL **)calloc(2*order_limit-1,sizeof(REAL *));
  assert(m_data != NULL);

  /* shift to center */
  m_data+=order_limit-1;

  /* main memory: (2*order_limit-1) * numpts */
  m_data[0]=(REAL *)calloc((2*order_limit-1)*numpts,sizeof(REAL));
  assert(m_data[0] != NULL);

  for(m=1;m<order_limit;m++){
    /* loop through other orders in positive/negative pairs */

    /* base +m on previous -m */
    m_data[m]=m_data[-(m-1)]+numpts;
    
    /* base -m on +m */
    m_data[-m]=m_data[m]+numpts;
  }

  /* load in the function values */
  fread(m_data[0],sizeof(REAL),(2*order_limit-1)*numpts,fp);

  /* close file */
  fclose(fp);
  /***********************************************************/


  /***********************************************************/
  /* get memory for hat, the transform coefficients */
  hat=(REAL **)calloc(2*order_limit-1,sizeof(REAL *));
  assert(hat != NULL);

  /* shift to center */
  hat+=order_limit-1;

  /* main memory: order_limit * order_limit*/
  hat[0]=(REAL *)calloc(order_limit*order_limit,sizeof(REAL));
  assert(hat[0] != NULL);

  for(m=1;m<order_limit;m++){
    /* loop through other orders in positive/negative pairs */

    /* base +m on previous -m */
    hat[m]=hat[-(m-1)]+(order_limit -(m-1));
    
    /* base -m on +m */
    hat[-m]=hat[m]+(order_limit -m);
  }
  /********************************************************/

#if DOTIME
  /* note time */
  prealg=second();
#endif

  /**********************************************************************/
  /* Do the transform itself (latitude only). */
  for(m=-order_limit+1;m<order_limit;m++){
    /* loop through the order */
    /* printf("m=%i\n",m);*/


  /* Used to seg fault here when -O flag used; seems okay now*/

    /* analyse for that m */
    Pm_direct1_analyse(hat[m],pmn_matrix+abs(m),m_data[m],&pmnwork);

  } 
  /**********************************************************************/

#if DOTIME
  printf("\t Core took %.3e seconds\n",second()-prealg);
#endif    


  /**********************************************************************/
  /* output values to out_file */
  fp=fopen(argv[3],"wb");
  
  /* note what we are doing */
  printf("writing coefficients to %s\n",argv[3]);

  /* write coefficients */
  fwrite(hat[0],sizeof(REAL),order_limit*order_limit,fp);

  /* close file */
  fclose(fp);
  /*************************************************************************/

  /*************************************************************************/
  /* clean up */
  free(m_data[0]);
  free(m_data-(order_limit-1));
  free(hat[0]);
  free(hat-(order_limit-1));

  /* we don't yet have a free-er for the pmn_matrices */

}
/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
