/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file assess_f1.c

\brief Assess the performance of the fast1 algorithm.

This program does an assessment of the fast transform for spherical
harmonics, using the plan created by initialize_f1

We determine the number of coefficients in the compressed
representation and time the synthesis and analysis transforms for each
individual order.  Note that totalling the time for each order tends
to overestimate the time for the total tranform, for a reason not
clear.

We only do the Associated Legendre Transform part of the SH transform.
To do a full transform, you also need to fourier transform in
longitude.

Usage:
call this routine with 3 command-line arguments:
f1_file iterations out_file
- f1_file -- The file created by initialize_f1, 
  for the size problem you want
- iterations -- the number of iterations to do >0.
  On small problems our timing routine is not precise enough,
  and will tell us most things take 0 time. To compensate we
  time some number of iterations of the same process and 
  divide by the number of iterations.
  This parameter sets the number of iterations for the m==0
  case. For other m, we increase the number of iteration to
  compensate for the smaller problem size
- out_file -- the file in which to put the results
  We first output global results, in lines beginning with #
  (so gnuplot can ignore them). Then we give the main data
  in the form:
	
       order  coefficient_count  synthesis_time analysis_time

  This column format can then be plotted
	
**/

#include "../libftsh.h"


/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
int main(int argc,
          char *argv[]
          ){


  register int m,iter;
  int numpts,band_limit,order_limit,iterations,num_iter;
  REAL *node,*weight;
  FILE *fp;
  Pm_1D *pmn_matrix;
  REAL **hat, **m_data;
  Pm_1D_Workspace pmnwork;
  REAL prealg;
  REAL *syn_time,*anal_time,syn_total,anal_total;
  int *coef_count;
  int coef_total;

  /* make sure the correct number of arguments were given
   (program name counts as one) */
  assert(argc==4);

  /********************************************************/
  /* load in f1_file data **/
  fp=fopen(argv[1],"rb");
  assert(fp != NULL);

  /* note what we are doing */
  printf("%s : reading initialization from %s\n",argv[0],argv[1]);

  /* load precomputed initialization and set up workspace and such */
  fast1_fromfile(&numpts,&band_limit,&node,&weight,
		 &pmn_matrix,&pmnwork,fp);

  /* close file */
  fclose(fp);
  /********************************************************/

  /* get iterations */
  iterations=atoi(argv[2]);

  /* note what we are doing */
  printf("iterations=%i numpts=%i band_limit=%i, dummy data\n",
	 iterations,numpts,band_limit);

  /********************************************************/
  /* set dummy coefficent data ***/

  /* set order_limit */
  order_limit=band_limit; 

  /* The user is encouraged to have their own structure for holding the
     transform coefficients. All we need is to be able to pass an array
     of coefficients for a given m */

  /* get memory for hat, the transform coefficients */
  hat=(REAL **)calloc(2*order_limit-1,sizeof(REAL *));
  assert(hat != NULL);

  /* shift to center */
  hat+=order_limit-1;

  /* main memory: order_limit * order_limit*/
  hat[0]=(REAL *)calloc(order_limit*order_limit,sizeof(REAL));
  assert(hat[0] != NULL);

  for(m=1;m<order_limit;m++){
    /* loop through other orders in positive/negative pairs */

    /* base +m on previous -m */
    hat[m]=hat[-(m-1)]+(order_limit -(m-1));
    
    /* base -m on +m */
    hat[-m]=hat[m]+(order_limit -m);
  }
  
  /********************************************************/


  /***********************************************************************/
  /* set dummy function data */
  m_data=(REAL **)calloc(2*order_limit-1,sizeof(REAL *));
  assert(m_data != NULL);

  /* shift to center */
  m_data+=order_limit-1;

  /* main memory: (2*order_limit-1) * numpts */
  m_data[0]=(REAL *)calloc((2*order_limit-1)*numpts,sizeof(REAL));
  assert(m_data[0] != NULL);

  for(m=1;m<order_limit;m++){
    /* loop through other orders in positive/negative pairs */

    /* base +m on previous -m */
    m_data[m]=m_data[-(m-1)]+numpts;
    
    /* base -m on +m */
    m_data[-m]=m_data[m]+numpts;
  }

  /**********************************************************************/


  /**********************************************************************/
  /* get memory for the time and count holders */
  
  /* time for the synthesis step for each order */
  syn_time=(REAL *)calloc(order_limit,sizeof(REAL));
  assert(syn_time != NULL);

  /* time for the analysis step for each order */
  anal_time=(REAL *)calloc(order_limit,sizeof(REAL));
  assert(anal_time != NULL);

  /* coefficient counts for each order */
  coef_count=(int *)calloc(order_limit,sizeof(int));
  assert(coef_count != NULL);

  /**********************************************************************/


  /**********************************************************************/
  /* Do the transform itself (lattitude only), m>=0 only. */
  for(m=0;m<order_limit;m++){
    /* loop through the order */

    /* set the coefficient counts */
    coef_count[m]=pmn_matrix[m].matrix[0].non_zero
      +pmn_matrix[m].matrix[1].non_zero;

    /* compute the number of iterations to use for this order.
       For m==0 this agrees with the iterations input,
       but for higher m we compensate for the smaller problem size */
    num_iter=(iterations*band_limit)/(band_limit-m);

    /* note time */
    prealg=second();

    for(iter=0;iter<num_iter;iter++){
      /* do a number of iterations */

      /* synthesize for that m */
      Pm_1d_synthesize(m_data[m],pmn_matrix+abs(m),hat[m],&pmnwork);

    }
    
    /* compute and note synthesis time */
    syn_time[m]=(second()-prealg)/num_iter;

    /********** analysis **************/

    /* note time */
    prealg=second();

    for(iter=0;iter<num_iter;iter++){
      /* do a number of iterations */

      /* analyse for that m */
      Pm_1d_analyse(hat[m],pmn_matrix+abs(m),m_data[m],&pmnwork);

    }
    
    /* compute and note analysis time */
    anal_time[m]=(second()-prealg)/num_iter;

  } 
  /**********************************************************************/

  /**********************************************************************/
  /* accumulate total times and counts.
     We pretend the m<0 values are present too. This is valid for time
     purposes, but not for memory.*/

  /* m==0 counts once */
  coef_total=coef_count[0];
  syn_total=syn_time[0];
  anal_total=anal_time[0];

  for(m=0;m<order_limit;m++){
    /* loop through remaining m */

    /* other m count twice (+/-) */
    coef_total+=2*coef_count[m];
    syn_total+=2*syn_time[m];
    anal_total+=2*anal_time[m];
  }
  /**********************************************************************/

  /**********************************************************************/
  /* output values to out_file */
  fp=fopen(argv[3],"w");
  
  /* note what we are doing */
  printf("Writing assesment to %s\n",argv[3]);

  /* first general parameters*/
  fprintf(fp,"#numpts=%i band_limit=%i iterations=%i\n",
	  numpts,band_limit,iterations);

  /* now the column headers*/
  fprintf(fp,"#order\t count\t synthesis\t analysis\n");

  /* now the totals */
  fprintf(fp,"#total\t %i\t %.3e\t %.3e\n",
	  coef_total,syn_total,anal_total);

  /* now the main data */
  
  for(m=0;m<order_limit;m++){
    /* loop through the orders */

    /* output data */
    fprintf(fp,"%i\t %i\t %.3e\t %.3e\n",
	    m,coef_count[m],syn_time[m],anal_time[m]);
  }

  /* close file */
  fclose(fp);
  /*************************************************************************/

  /*************************************************************************/
  /* clean up */
  free(m_data[0]);
  free(m_data-(order_limit-1));
  free(hat[0]);
  free(hat-(order_limit-1));
  /* we don't yet have a free-er for the pmn_matrices */
  free(syn_time);
  free(anal_time);
  free(coef_count);

}
/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
