/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file diff_coef.c

\brief Compare two sets of coefficients.

This program compares two sets of coefficients as used in a spherical
harmonic transform. It outputs the \\(L^2\\) error between them

For instance we can compare initial coefficients with the result after
a synthesis and an analysis.

Usage:
call this routine with 3 command-line arguments:
band_limit coef_file_1 coef_file_2
- band_limit -- The restriction on the degree of the expansion 
  (n in \\(P^m_n\\) notation).
- coef_file_1 -- the file from which to read the first set of coefficients.
  We read band_limit*band_limit REALs.
- coef_file_2 -- the file from which to read the second set of 
  coefficients. We read band_limit*band_limit REALs.
  
**/

#include "../libftsh.h"

/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
int main(int argc,
          char *argv[]
          ){


  register int i;
  int band_limit,order_limit;
  FILE *fp;
  REAL *hat_1,*hat_2,norm,error;

  /* make sure the correct number of arguments were given
   (program name counts as one) */
  assert(argc==4);

  /* read in the band_limit */
  band_limit=atoi(argv[1]);

  /* we really use band_limit as an order_limit */
  order_limit=band_limit;

  /* note what we are doing */
  printf("%s : band_limit=%i, comparing %s and %s\n",
	 argv[0],band_limit,argv[2],argv[3]);

  /******************************************/
  /* read the first set */

  /* open file */
  fp=fopen(argv[2],"rb");
  assert(fp != NULL);

  /* get main memory */
  hat_1=(REAL *)calloc(order_limit*order_limit,sizeof(REAL));
  assert(hat_1 != NULL);

  /* read in the values */
  fread(hat_1,sizeof(REAL),order_limit*order_limit,fp);

  /* close file */
  fclose(fp);
  /*********************************************/


  /******************************************/
  /* read the second set */

  /* open file */
  fp=fopen(argv[3],"rb");
  assert(fp != NULL);

  /* get main memory */
  hat_2=(REAL *)calloc(order_limit*order_limit,sizeof(REAL));
  assert(hat_2 != NULL);

  /* read in the values */
  fread(hat_2,sizeof(REAL),order_limit*order_limit,fp);

  /* close file */
  fclose(fp);
  /*********************************************/


  /*********************************************/
  /* compare, accumulating error and norm */

  norm=0.0;
  error=0.0;
  for(i=0;i<order_limit*order_limit;i++){
    /* loop through all coefficients */
    
    /* error */
    error+=(hat_1[i]-hat_2[i])*(hat_1[i]-hat_2[i]);

    /* norm of hat_1. hat_2 should be comparable */
    norm+=hat_1[i]*hat_1[i];
  }

  /* sqrt error and norm */
  norm=sqrt(norm);
  error=sqrt(error);

  /* print results */
  printf("error %e out of norm %e, relative error %e\n",
	 error,norm,error/norm);
  /********************************************************/


  /********************************************************/
  /* clean up **/
  free(hat_1);
  free(hat_2);

}
/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
