/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file diff_fcn.c

\brief Compare two sets of function values.

This program compares two sets of function values as used in a
spherical harmonic transform. It outputs the \\(L^2\\) error between
them.

For instance we can compare function values after two different
versions of the transform.

We do not use the quadrature weight, so this is only a rough estimate
of the proper error.

Usage:
call this routine with 4 command-line arguments:
numpts band_limit fcn_file_1 fcn_file_2
- numpts --  the number of latitude points
- band_limit -- The restriction on the degree of the expansion 
  (n in \\(P^m_n\\) notation).
- fcn_file_1 -- the file from which to read the first set of fcn values.
  We read numpts*(2*band_limit-1) REALs.
- fcn_file_2 -- the file from which to read the second set of fcn values.
  We read numpts*(2*band_limit-1)  REALs.
	       
**/

#include "../libftsh.h"

/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
void main(int argc,
          char *argv[]
          ){


  register int i;
  int band_limit,order_limit,numpts;
  FILE *fp;
  REAL *fcn_1,*fcn_2,norm,error;

  /* make sure the correct number of arguments were given
   (program name counts as one) */
  assert(argc==5);

  /* read in numpts */
  numpts=atoi(argv[1]);

  /* read in the band_limit */
  band_limit=atoi(argv[2]);

  /* we really use band_limit as an order_limit */
  order_limit=band_limit;

  /* note what we are doing */
  printf("%s : numpts=%i, band_limit=%i, comparing %s and %s\n",
	 argv[0],numpts,band_limit,argv[3],argv[4]);

  /******************************************/
  /* read the first set */

  /* open file */
  fp=fopen(argv[3],"rb");
  assert(fp != NULL);

  /* get main memory */
  fcn_1=(REAL *)calloc(numpts*(2*order_limit-1),sizeof(REAL));
  assert(fcn_1 != NULL);

  /* read in the values */
  fread(fcn_1,sizeof(REAL),numpts*(2*order_limit-1),fp);

  /* close file */
  fclose(fp);
  /*********************************************/


  /******************************************/
  /* read the second set */

  /* open file */
  fp=fopen(argv[4],"rb");
  assert(fp != NULL);

  /* get main memory */
  fcn_2=(REAL *)calloc(numpts*(2*order_limit-1),sizeof(REAL));
  assert(fcn_2 != NULL);

  /* read in the values */
  fread(fcn_2,sizeof(REAL),numpts*(2*order_limit-1),fp);

  /* close file */
  fclose(fp);
  /*********************************************/


  /*********************************************/
  /* compare, accumulating error and norm */

  norm=0.0;
  error=0.0;
  for(i=0;i<numpts*(2*order_limit-1);i++){
    /* loop through all fcn values */
    
    /* error */
    error+=(fcn_1[i]-fcn_2[i])*(fcn_1[i]-fcn_2[i]);

    /* norm of fcn_1. fcn_2 should be comparable */
    norm+=fcn_1[i]*fcn_1[i];
  }

  /* sqrt error and norm */
  norm=sqrt(norm);
  error=sqrt(error);

  /* print results */
  printf("error %e out of norm %e (no quadrature),\n\t relative error %e\n",
	 error,norm,error/norm);
  /********************************************************/


  /********************************************************/
  /* clean up **/
  free(fcn_1);
  free(fcn_2);

}
/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
