/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file initialize_d1.c

\brief Precompute data for the direct1 algorithm.

This program initializes the direct transform for spherical harmonics,
for comparison with "fast" transforms. The direct transform is
ordinary dense matrix times vector multiplication. We do however take
advantage of the even/odd symmetry about the equator.

Usage:
call this routine with 3 command-line arguments:
numpts band_limit out_file
- numpts -- the number of latitude points.
  We construct a gaussian quadrature for this number of points,
  although other node/weight combinations could be used.
  In this version, the number of points must be divisible by 2.
- band_limit -- The restriction on the degree of the expansion 
  (n in \\(P^m_n\\) notation).
  For gaussian nodes this must be less than numpts.
- out_file -- The name of the file in which to put results of the 
  initialization

**/

#include "../libftsh.h"

/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
void main(int argc,
          char *argv[]
          ){


  int numpts,band_limit;
  REAL g_epsilon,epsilon;
  REAL *gnode,*gweight;
  FILE *fp;

  /* make sure the correct number of arguments were given
   (program name counts as one) */
  assert(argc==4);

  /* get numpts */
  numpts=atoi(argv[1]);

  /* get band_limit */
  band_limit=atoi(argv[2]);

  /* sanity checks */
  assert(numpts > 0);
  assert(band_limit >0 );
  assert(band_limit < numpts );

  /* print parameters */
  printf("%s : numpts=%i band_limit=%i\n",
	 argv[0],numpts,band_limit);

  /*************************************************************************/
  /*************************************************************************/
  /* construct the gaussian nodes and weights */
  /* The user is encouraged to substitute their own node/weight maker here.
     The algorthim was only proven for equispaced nodes, but works fine
     for gaussian nodes as well. It is expected to work fine for any node
     set that is a smooth distortion of equispaced. A random perturbation
     will mess it up however. The nodes must be symmetric about the equator
     and not include the poles.

     Note that we construct Arcsin( the usual gaussian nodes).
     Other nodes must also be translated into the angular coordinate.
     */

  /* g_epsilon is the target accuracy for the gaussian nodes */
#if USE_double
  g_epsilon=1e-15;
#else
  g_epsilon=1e-5;
#endif

  /* get memory */
  
  /* for the nodes */
  gnode=(REAL *)calloc(numpts,sizeof(REAL));
  assert(gnode!=NULL);

  /* for the weights */
  gweight=(REAL *)calloc(numpts,sizeof(REAL));
  assert(gweight!=NULL);

  /* actually make them */
  mk_gnode_gweight(gnode,gweight,g_epsilon,numpts);
  /*************************************************************************/
  /*************************************************************************/

  /* open file*/
  fp=fopen(argv[3],"wb");

  /*************************************************************************/

  /* initialize, dump to file, and clean up */
  direct1_tofile(numpts,band_limit,gnode,gweight,fp);

  /*************************************************************************/

  /* close file */
  fclose(fp);

  /* final diagnostics */
  printf("Wrote to file %s .\n",argv[3]);

  /* free nodes and weights */
  free(gnode);
  free(gweight);
    
}
/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
