/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file initialize_f1.c

\brief Precompute data for the fast1 algorithm.

This program initializes the fast transform for spherical harmonics

A Fast Transform for Spherical Harmonics
Martin J. Mohlenkamp 
Journal of Fourier Analysis and Applications, 5(2/3):159-184, 1999
   
This version is the simplest, both from an algorithmic standpoint and
a user standpoint. Algorithmicly, it does the single partition
version, which should have the lowest break-even point, but will be
slower asymptotically. For the user, it provides routines with the
fewest parameters to worry about. This is easier, but does not allow
tuning.

This initialization algorithm is slow. For a degree N spherical
harmonic representation it will take \(O(N^3 log(N))\). The idea is to
initialize only once (for some set of parameters) and save it to a
file.

Usage:
call this routine with 4 command-line arguments:
numpts band_limit epsilon out_file
- numpts -- the number of latitude points.
  We construct a gaussian quadrature for this number of points,
  although other node/weight combinations could be used.
  In this version, the number of points must be divisible by 4.
  Additional factors of 2 will improve performance. In general
  you should have a power of 2 times a factor smaller than 10.
- band_limit -- The restriction on the degree of the expansion (n in P^m_n
  notation).
  For gaussian nodes this must be less than numpts.
- epsilon -- The target precision. 1 > epsilon >0
  This algorithm gives an approximate transform, but allows you
  to control the error. The idea is to choose the error here 
  slightly smaller than the errors in the rest of your program.
  The larger the allowed error, the faster the algorithm will run. 
  As you make the error smaller, the slow-down is not dramatic, so
  do not be afraid to ask for somewhat smaller error than you think
  you need.
- out_file -- The name of the file in which to put results of the 
  initialization

Notes:
- epsilon directly controls the level of truncation of entries.
  By fiddling with the normalization, we think we now have it bounding
  the overall, accumulated \(L^2\) error.

**/

#include "../libftsh.h"

/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
int main(int argc,
          char *argv[]
          ){


  int numpts,band_limit;
  REAL g_epsilon,epsilon,ratio;
  REAL *gnode,*gweight;
  FILE *fp;

  /* make sure the correct number of arguments were given
   (program name counts as one) */
  assert(argc==5);

  /* get numpts */
  numpts=atoi(argv[1]);

  /* get band_limit */
  band_limit=atoi(argv[2]);

  /* get epsilon */
  epsilon=atof(argv[3]);

  /* sanity checks */
  assert(numpts > 0);
  assert(band_limit >0 );
  assert(band_limit < numpts );
  assert(epsilon < 1);
  /* epsilon < 0 could be used to disable all truncation */

  /* print parameters */
  printf("%s : numpts=%i band_limit=%i epsilon=%e\n",
	 argv[0],numpts,band_limit,epsilon);

  /*************************************************************************/
  /*************************************************************************/
  /* construct the gaussian nodes and weights */
  /* The user is encouraged to substitute their own node/weight maker here.
     The algorthim was only proven for equispaced nodes, but works fine
     for gaussian nodes as well. It is expected to work fine for any node
     set that is a smooth distortion of equispaced. A random perturbation
     will mess it up however. The nodes must be symmetric about the equator
     and not include the poles.

     Note that we construct Arcsin( the usual gaussian nodes).
     Other nodes must also be translated into the angular coordinate.
     */

  /* g_epsilon is the target accuracy for the gaussian nodes */
#if USE_double
  g_epsilon=1e-15;
#else
  g_epsilon=1e-5;
#endif

  /* get memory */
  
  /* for the nodes */
  gnode=(REAL *)calloc(numpts,sizeof(REAL));
  assert(gnode!=NULL);

  /* for the weights */
  gweight=(REAL *)calloc(numpts,sizeof(REAL));
  assert(gweight!=NULL);

  /* actually make them */
  mk_gnode_gweight(gnode,gweight,g_epsilon,numpts);
  /*************************************************************************/
  /*************************************************************************/

  /* open file*/
  fp=fopen(argv[4],"wb");

  /*************************************************************************/

  /* initialize, dump to file, and clean up */
  ratio=fast1_tofile(numpts,band_limit,gnode,gweight,epsilon,fp);

  /*************************************************************************/

  /* close file */
  fclose(fp);

  /* final diagnostics */
  printf("Wrote to file %s . Compression ratio = %5.2f\n",
	 argv[4],ratio);

  /* the compression ratio gives an upper bound on the speed-up achieved
     by this algorithm. There is some additional overhead, however. */

  /* free nodes and weights */
  free(gnode);
  free(gweight);
    
}
/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
