/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file random_coef.c

\brief Construct random spherical harmonic coefficients.

This program constructs random coefficients for use in a spherical
harmonic transform, such as synthesize_f1. The are normalized to total
norm 1.

Usage:
call this routine with 2 command-line arguments:
band_limit out_file
- band_limit -- The restriction on the degree of the expansion (n in P^m_N
  notation).
- out_file -- the file in which to put the function values. 
  We write band_limit*band_limit random REALs, with total norm 1.

**/

#include "../libftsh.h"

/*************************************************************************/
/*************************************************************************/
/*************************************************************************/
int main(int argc,
          char *argv[]
          ){


  register int i;
  int band_limit,order_limit;
  FILE *fp;
  REAL *hat,norm;
  time_t now;

  /* make sure the correct number of arguments were given
   (program name counts as one) */
  assert(argc==3);

  /* read in the band_limit */
  band_limit=atoi(argv[1]);

  /* we really use band_limit as an order_limit */
  order_limit=band_limit;

  /* note what we are doing */
  printf("%s : band_limit=%i\n",argv[0],band_limit);

  /* get main memory */
  hat=(REAL *)calloc(order_limit*order_limit,sizeof(REAL));
  assert(hat != NULL);

  /******************************************/
  /* fill with random values of norm 1 */

  /* random seed */
  now = time(NULL);
  assert(now!=-1); /* -1 means can't read the time */
  srand((unsigned)((unsigned) now));
  
  /* fill hat with randoms, accumulating norm */
  norm=0.0;
  for(i=0;i<order_limit*order_limit;i++){
    hat[i]=(2.0* rand()/(RAND_MAX + 1.0)) -1.0;
    /* accumulate norm */
    norm+=hat[i]*hat[i];
  }

  /* scale to norm 1*/
  norm=1.0/sqrt(norm);
  for(i=0;i<order_limit*order_limit;i++)
    hat[i]*=norm;
  /********************************************************/

  /********************************************************/
  /* write to file */
  fp=fopen(argv[2],"wb");

  /* note what we are doing */
  printf("writing random coefficients of norm 1 on %s\n",argv[2]);

  /* write */
  fwrite(hat,sizeof(REAL),order_limit*order_limit,fp);

  /* close file */
  fclose(fp);
  /********************************************************/


  /********************************************************/
  /* clean up **/
  free(hat);

}

