/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

/** \file bdmat.c
    
\brief Contains simple functions based on the Banded_Matrix struct.

Summary:
- init_bdmat_memory allocates a block of memory.
- give_bdmat_memory passes extra memory to the next Banded_Matrix.
- free_bdmat_memory frees a block of memory.
- bdmat_copy makes a copy
- dump_bin_bdmat writes to file in binary.
- load_bin_bdmat reads back what dump_bin_bdmat wrote.

*/


Banded_Matrix init_bdmat_memory(int max_size,
				int max_rows,
				int max_bands
				){
  /** This routine gets a large amount of memory for Banded_Matrix's,
      and gives it to a single Banded_Matrix to hold. This special
      Banded_Matrix acts as a pointer to this block of memory, and
      also keeps track of how large this memory is.

      INPUT:  
      - max_size -- the amount of memory to get for matrix
      - max_rows -- the amount of memory to get for band_p
      - max_bands -- (1/3) the amount of memory to get for band_list

      OUPUT: returns a Banded_Matrix with
      - .max_size=max_size
      - .max_rows=max_rows+1
      - .max_bands=max_bands;
      - .non_zero=0;
      - .n_row=-1 This ruse allows give_bdmat_memory to pass along
        all the memory instead of losing one row
      - .n_col=0
      - .band_p pointing to max_rows+1 ints
      - .band_list pointing to 3*max_bands ints
      - .matrix pointing to max_size REALs

  */

  Banded_Matrix bdmat;

#if ENTEREXIT
  printf("enter init_bdmat_memory\n");
#endif

  /* empty matrix */
  bdmat.n_col=0; 
  bdmat.non_zero=0;
  bdmat.n_row=0; 

  /* note the max_rows */
  /* one row is wasted when we first pass memory*/
  bdmat.max_rows=max_rows+1;

  /* memory for band_p */
  /* one row is wasted when we first pass memory*/
  bdmat.band_p=(int *)calloc(max_rows+1,sizeof(int));
  assert(bdmat.band_p != NULL);

  /* set no bands for this memory holder */
  bdmat.band_p[0]=0;

  /* note the max_bands */
  bdmat.max_bands=max_bands;

  /* memory for band_list, 3*max_bands */
  bdmat.band_list=(int *)calloc(3*max_bands,sizeof(int));
  assert(bdmat.band_list != NULL);

  /* note the max_size */
  bdmat.max_size=max_size;

  /* memory for matrix */
  bdmat.matrix=(REAL *)calloc(max_size,sizeof(REAL));
  assert(bdmat.matrix != NULL);

#if ENTEREXIT
  printf("return from init_bdmat_memory\n");
#endif

  return bdmat;
}

void give_bdmat_memory(Banded_Matrix *current,
		       Banded_Matrix *previous
		       ){
  /** This routine sets up memory for the next Banded_Matrix we wish to use.
     
  It looks at the previous Banded_Matrix we used, and increments fields

  INPUT: previous -- the last Banded_Matrix we used

  OUTPUT: current -- the Banded_Matrix we wish to use next
                   
  We set the .band_p, .band_list, and .matrix fields to 
  the next available ints and REAL.
  .max_size, .max_rows, and .max_bands are set to account for 
  the amount used in previous.
  also set:
  .non_zero=0;
  .n_row=-1
  .n_col=0

  */

  Banded_Matrix bdmat;

#if ENTEREXIT
  printf("entering give_bdmat_memory\n");
#endif

  /* empty matrix */
  current->n_col=0; 
  current->non_zero=0;
  current->n_row=-1; 

  /* advance band_p */
  current->max_rows=previous->max_rows - previous->n_row -1;
  assert(current->max_rows >=0);
  current->band_p=previous->band_p +previous->n_row +1;
    
  /* advance band_list */
  current->max_bands=previous->max_bands - 
    ((previous->band_p[previous->n_row])/3);
  assert(current->max_bands >=0);
  current->band_list=previous->band_list 
    +(previous->band_p[previous->n_row]);

  /* advance matrix */
  current->max_size=previous->max_size -previous->non_zero;
  assert(current->max_size >= 0);
  current->matrix=previous->matrix + (previous->non_zero);

#if ENTEREXIT
  printf("exiting give_bdmat_memory\n");
#endif

}

void free_bdmat_memory(Banded_Matrix *in
		       ){
  /** This routine frees the memory that was given to a Banded_Matrix
      by init_bdmat_memory.

      INPUT: in -- the Banded_Matrix to free.  it must have been
      initialized by init_bdmat_memory, or we will attempt to free
      things not gotten by calloc, probaby causing catastrophic
      failure.  

      NOTES: 
      - if used incorrectly, this routine may yield nasty memory bugs
    */

#if ENTEREXIT
  printf("enter free_bdmat_memory\n");
#endif

  /* free memory from band_p */
  free(in->band_p);

  /* free memory from band_list */
  free(in->band_list);

  /* free memory from matrix */
  free(in->matrix);

#if ENTEREXIT
  printf("exiting free_bdmat_memory\n");
#endif

}

void bdmat_copy(Banded_Matrix *out,
		Banded_Matrix *in
		){
  /** This routine copies a Banded_Matrix.

  INPUTS: in -- the Banded_Matrix to copy

  OUTPUT: out -- the values of in are copied to out
  Should be initialized as follows
  - .max_rows is the amount of row memory available
  - .band_p points to .max_rows ints
  - .max_bands is the amount of band memory available
  - .band_list points to 3*.max_bands ints
  - .max_size is the amount of main memory available
  - .matrix points to .max_size REAL

  */

  register int i, top_i;

#if ENTEREXIT
  printf("entering bdmat_copy\n");
#endif

  /* set sizes for out*/
  out->n_col=in->n_col;
  out->n_row=in->n_row;

  /* make sure there is enough row memory */
  assert(out->n_row < out->max_rows );

  /* set non_zero */
  out->non_zero=in->non_zero;

  /* make sure there is enough room */
  assert(out->non_zero <= out->max_size );

  /* make sure there is enough band memory.
     The number of bands (*3) is held in in->band_p[n_row] */
  assert(in->band_p[in->n_row] <= 3*(out->max_bands));


  /* actual copying *********************/

  /* copy band_p. there are n_row+1 values */
  for(i=0;i<= out->n_row ; i++)
    out->band_p[i]=in->band_p[i];

  /* copy band_list. There are (in->band_p[in->n_row]) values */
  for(i=0;i<(in->band_p[in->n_row]);i++)
    out->band_list[i]=in->band_list[i];

  /*copy matrix values. There are non_zero of them */
  top_i=(out->non_zero);
  for(i=0;i<top_i;i++)
    out->matrix[i]=in->matrix[i];


#if ENTEREXIT
  printf("exiting bdmat_copy\n");
#endif
}

void dump_bin_bdmat(Banded_Matrix *x,
		    FILE *out
		    ){
  /** This routine dumps a Banded_Matrix to a file in binary format.
     
  INPUTS: 
  - x -- the Banded_Matrix
  - out -- A pointer to the file to write into. 
    Setting this to stdout makes it write to the screen.
    If out==NULL we use stdout.
    
  OUTPUT: writes into the file given by out. The writing mode ("wb" or "ab")
          is determined outside of this routine.

  */
#if ENTEREXIT
  printf("entering dump_bin_bdmat\n");
#endif

  /* default is stdout (send to screen)*/
  if(out==NULL)
    out=stdout;

  /* max_size, max_rows, and max_bands will not be relevent
     when we reread this so we don't write them*/

  /* number of rows and columns in the matrix */
  fwrite(&(x->n_row),sizeof(int),1,out);
  fwrite(&(x->n_col),sizeof(int),1,out);

  /* number of non_zero entries */
  fwrite(&(x->non_zero),sizeof(int),1,out);

  /* write band_p. there are n_row+1 of them */
  fwrite(x->band_p,sizeof(int),(x->n_row)+1,out);

  /* write band_list. there are 3*num_bands==band_p[n_row] of them */
  fwrite(x->band_list,sizeof(int),x->band_p[x->n_row],out);

  /* write the matrix entries. there are non_zero of them */
  fwrite(x->matrix,sizeof(REAL),(x->non_zero),out);


#if ENTEREXIT
  printf("exiting dump_bin_bdmat\n");
#endif
}

void load_bin_bdmat(Banded_Matrix *x,
		    FILE *in
		    ){
  /** This routine boads a Banded_Matrix to a file in binary format,
      as written by dump_bin_bdmat
     
      INPUTS: 
      - x -- the Banded_Matrix. Should be initialized as follows
        + .max_rows is the amount of row memory available
	+ .band_p points to .max_rows ints
	+ .max_bands is the amount of band memory available
	+ .band_list points to 3*.max_bands ints
	+ .max_size is the amount of main memory available
	+ .matrix points to .max_size REAL      
      - in -- A pointer to the file to read from. 
        Setting this to stdin makes it read from the screen.
	If in==NULL we use stdin.
	The reading mode ("rb") 
	is determined outside of this routine.
     
      OUTPUT: x -- is filled in.

     */
#if ENTEREXIT
  printf("entering load_bin_bdmat\n");
#endif

  /* default is stdin (read from screen)*/
  if(in==NULL)
    in=stdin;

  /* max_size, max_rows, and max_bands were not written */

  /* number of rows and columns in the matrix */
  fread(&(x->n_row),sizeof(int),1,in);
  fread(&(x->n_col),sizeof(int),1,in);

  /* number of non_zero entries */
  fread(&(x->non_zero),sizeof(int),1,in);

  /* check for sufficient sizefor band_p. we use one extra row */
  assert(x->n_row < x->max_rows);

  /* read band_p. there are n_row+1 of them */
  fread(x->band_p,sizeof(int),(x->n_row)+1,in);

  /* check for sufficient size for band_list */
  assert(x->band_p[x->n_row] <= 3*(x->max_bands));

  /* read band_list. there are 3*num_bands==band_p[n_row] of them */
  fread(x->band_list,sizeof(int),x->band_p[x->n_row],in);

  /* check for sufficient size for number of entries */
  assert(x->non_zero <= x->max_size);

  /* write the matrix entries. there are non_zero of them */
  fread(x->matrix,sizeof(REAL),(x->non_zero),in);


#if ENTEREXIT
  printf("exiting load_bin_bdmat\n");
#endif
}











