/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

/** \file bdmat_ops.c
    
\brief Contains functions interacting the Banded_Matrix struct with others.

Summary:
- bdmat_dvect_multiply applies to a Dense_Vector
- bdmat_t_dvect_multiply applies the transpose to a Dense_Vector
- dmat_to_bdmat converts from a Dense_Matrix

*/

/** Switch on old or new method to loop.
 New method tries to localize variables and should be faster.
 In my tests it made synthesis twice as fast but did not help analysis.
 The compiler optimization may be already doing some localization to the 
 old loops. 
*/
#define OLDLOOPS 0

void bdmat_dvect_multiply(Dense_Vector *out,
			  Banded_Matrix *inmat,
			  Dense_Vector *invect
			  ){
  /** This routine applies a Banded_Matrix to a Dense_Vector to produce a 
      Dense_Vector.
      The matrix may be square or rectangular.

      INPUTS: 
      - out -- the Dense_Vector to put the result in.
        Should be initialized as follows:
	+ .max_size is the amount of memory available.
	+ .vector points to .max_size REAL
      - inmat -- the Banded_Matrix. 
      - invect -- the Dense_Vector


      OUTPUT: out is modified so out = inmat * invect

      NOTES:
      - might want to hold some things as local registers, rather than 
        repeatedly accessing arrays.
      - looping structure may be inefficient
      
  */
  register int irow, j,k;
  register REAL tmpreal;
#if OLDLOOPS
#else
  int jtop,ktop;
  REAL *matrixbit,*vectorbit;
#endif

#if ENTEREXIT
  printf("enter bdmat_dvect_multiply\n");
#endif

  /* check for compatible sizes **/
  assert(inmat->n_col == invect->n_row);

  /* set the dimensions of out **/
  out->n_row=inmat->n_row;

  /* check for enough room */
  assert(out->n_row <= out->max_size);

  for (irow=0; irow<(inmat->n_row); irow++){
    /* loop through the rows in inmat */

    /* We essentially take the inner product of the given row with invect.*/

    /* erase out to start*/
    tmpreal=0.0;
#if OLDLOOPS
    for(j=inmat->band_p[irow];j<inmat->band_p[irow+1];j+=3){
      /* loop through the bands on that row (by 3s) **/
	
      for(k=inmat->band_list[j];k<inmat->band_list[j+1];k++){
	/* loop within the band */

	/* multiply the inmat entry by the corresponding invect entry and
	   add the result to the proper out entry*/

	tmpreal+=
	  inmat->matrix[inmat->band_list[j+2]+k-inmat->band_list[j]]
	  *invect->vector[k];

      }
    }
#else
    jtop=inmat->band_p[irow+1];
    for(j=inmat->band_p[irow];j<jtop;j+=3){
      /* loop through the bands on that row (by 3s) **/

      ktop=inmat->band_list[j+1]-inmat->band_list[j];
      matrixbit=inmat->matrix+inmat->band_list[j+2];
      vectorbit=invect->vector+inmat->band_list[j];
      for(k=0;k<ktop;k++){
	/* loop within the band */

	/* multiply the inmat entry by the corresponding invect entry and
	   add the result to the proper out entry*/

	tmpreal+=matrixbit[k]*vectorbit[k];

      }
    }
#endif
    /* set out value */
    out->vector[irow]=tmpreal;

  }/* row loop */

#if ENTEREXIT
  printf("exit from  bdmat_dvect_multiply\n");
#endif

}


void bdmat_t_dvect_multiply(Dense_Vector *out,
			    Banded_Matrix *inmat,
			    Dense_Vector *invect
			    ){
  /** This routine applies the TRANSPOSE of a Banded_Matrix 
      to a Dense_Vector to produce a Dense_Vector.
      The matrix may be square or rectangular.

      INPUTS: 
      - out -- the Dense_Vector to put the result in.
        Should be initialized as follows
	+ .max_size is the amount of memory available
	+ .vector points to .max_size REAL
      - inmat -- the Banded_Matrix. 
      - invect -- the Dense_Vector


      OUTPUT: out is modified so out = TRANSPOSE(inmat) * invect
      
      NOTES:
      - might want to hold some things as local registers, rather than 
        repeatedly accessing arrays
      - looping structure may be inefficient
  */
  register int irow, j,k;
  register REAL tmp_real;
#if OLDLOOPS
#else
  int jtop,ktop;
  REAL *matrixbit,*vectorbit;
#endif

#if ENTEREXIT
  printf("enter bdmat_t_dvect_multiply\n");
#endif

  /* check for compatible sizes; remember transposing **/
  assert(inmat->n_row == invect->n_row);

  /* set the dimensions of out; remember transposing  **/
  out->n_row=inmat->n_col;

  /* check for enough room */
  assert(out->n_row <= out->max_size);

  /* clear out */
  for(j=0;j<out->n_row;j++){
    out->vector[j]=0.0;
  }

  for (irow=0; irow<(inmat->n_row); irow++){
    /* loop through the rows in inmat */

    /* We now multiply the given row of inmat with the corresponding row
       value from invect, and add onto out.
       If we are using blocks then we do block matrix-vector multiplies
       instead of regular multiplication.*/
    /* store the vector value as a register */
    tmp_real=invect->vector[irow];
#if OLDLOOPS
    for(j=inmat->band_p[irow];j<inmat->band_p[irow+1];j+=3){
      /* loop through the bands on that row (by 3s) **/
	
      for(k=inmat->band_list[j];k<inmat->band_list[j+1];k++){
	/* loop within the band */

	/* multiply the inmat entry by the corresponding invect entry and
	   add the result to the proper out entry*/
	out->vector[k]+=inmat->matrix[inmat->band_list[j+2]+k
				     -inmat->band_list[j]]
          *tmp_real; /* invect->vector[irow] */

      }
    }
#else
    jtop=inmat->band_p[irow+1];
    for(j=inmat->band_p[irow];j<jtop;j+=3){
      /* loop through the bands on that row (by 3s) **/
      ktop=inmat->band_list[j+1]-inmat->band_list[j];
      matrixbit=inmat->matrix+inmat->band_list[j+2];
      vectorbit=out->vector+inmat->band_list[j];
      for(k=0;k<ktop;k++){
	/* loop within the band */

	/* multiply the inmat entry by the corresponding invect entry and
	   add the result to the proper out entry*/
	vectorbit[k]+=matrixbit[k]*tmp_real;

      }
    }
#endif
  }/* row loop */

#if ENTEREXIT
  printf("exit from  bdmat_t_dvect_multiply\n");
#endif

}

void dmat_to_bdmat(Banded_Matrix *out,
		   Dense_Matrix *in,
		   REAL cutoff,
		   int min_gap
		   ){
  /** Convert a matrix in Dense_Matrix form to Banded_Matrix form.

  INPUT:
  - out -- Where to put the result in Banded_Matrix format. 
    Should be initialized as follows
    + .max_rows is the amount of row memory available
    + .band_p points to .max_rows ints
    +.max_bands is the amount of band memory available
    + .band_list points to 3*.max_bands ints
    + .max_size is the amount of main memory available
    + .matrix points to .max_size REAL
  - in -- the Dense_Matrix
  - cutoff -- the truncation threshold to use in the conversion.
    Entries less than cutoff are discarded. To disable
    truncation, set cutoff <0 .
  - min_gap -- the minimal allowed gap between bands. Coefficients
    less than cutoff are retained if removing them 
    makes a gap of less than min_gap.
    To disable, set to 0.

  OUTPUT: out

  NOTES:
  - not checking for sufficient memory

  */


  register int i,j,k,row;
  int kept,bands_kept,local_min_gap;

#if ENTEREXIT
  printf("enter dmat_to_bdmat\n");
#endif

  /* check for rational min_gap */
  if(min_gap <0)
    min_gap=0;

  /* set the size for the output */
  out->n_col= in->n_col;
  out->n_row= in->n_row;
#if 0
  printf("in->n_row=%i, out->n_row=%i ",in->n_row,out->n_row);
#endif

  /* make sure we have enough row memory */
  assert(out->n_row < out->max_rows);

  /* kept is the total number of entries we have kept */
  kept=0;

  /* bands_kept is the total number of bands we have kept */
  bands_kept=0;

  for(row=0;row< out->n_row;row++){
    /* loop through the rows */

    /* mark the beginning of this row in band_p */
    out->band_p[row]=3*bands_kept;

    for(i=0;i<in->n_col;i++){
      /* loop through the row */

      if(fabs(in->matrix[row*(in->n_col)+i]) > cutoff){
	/* we have found the beginning of a band */

	/* note start of band */
	out->band_list[3*bands_kept]=i;

	/* note offset within matrix */
	out->band_list[3*bands_kept+2]=kept;

	for(j=i;j<in->n_col;j++){
	  /* continue looping, now within the band*/

	  if(fabs(in->matrix[row*(in->n_col)+j]) > cutoff){
	    /* keeper */

	    /* copy the coefficient */
	    out->matrix[kept]=in->matrix[row*(in->n_col)+j];
	    
	    /* note another kept */
	    kept++;
	    
	  }else{
	    /* we found a candidate for the end of the band */

	    /* fix min_gap so we don't go off the edge */
	    local_min_gap=(min_gap <(in->n_col -j))? min_gap : in->n_col -j;

	    for(k=1;k<local_min_gap;k++){
	      /* loop through the minimal gap length */

	      if(fabs(in->matrix[row*(in->n_col)+j+k]) > cutoff){
		/* found a big one, gap too small */
		
		/* update j */
		j+=k;

		for(;k>=0;k--){
		  /* loop back through the gap, filling in */

		  /* copy the coefficient */
		  out->matrix[kept]=in->matrix[row*(in->n_col)+j-k];
	  
		  /* note another kept */
		  kept++;
		}

		/* short out k loop. min_gap-1 also would short, but this
		 way we can determine natural exits*/
		k=local_min_gap;
	      }
	    }

	    if(k==local_min_gap){
	      /* didn't find another big one, so end the band */

	      /* mark the band end */
	      out->band_list[3*bands_kept+1]=j;
	      
	      /* reset i at end of band */
	      i=j;

	      /* short out j loop. in->n_col-1 would also short the loop,
		 but this way we can distinguish natural from shorted exits*/
	      j=in->n_col;

	      /* note another band kept*/
	      bands_kept++;
	    }
	  } 
	  /* else unnatural exit, continue the band. j was updated above */
	    
	} /* j loop */

	if(j==in->n_col){
	  /* we hit the right edge during a band**/
	  
	  /* mark the band end */
	  out->band_list[3*bands_kept+1]=j;
	  
	  /* reset i at end of band, shorting i loop */
	  i=j;

	  /* note another band kept*/
	  bands_kept++;
	}

      } /* band started */
    }/* i loop */
  } /* row loop */

  /* mark the end of the last row in band_p[row] */
  out->band_p[row]=3*bands_kept;

  /* set non_zero */
  out->non_zero=kept;

#if 0
  printf("--> out->n_row=%i \n",out->n_row);
#endif

#if ENTEREXIT
  printf("exiting dmat_to_bdmat\n");
#endif
}
/************************************************************************/









