/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file dctw.c

\brief Discrete Cosine Transform based on FFTW.

Summary:
- dctw_init initializes; do once. 
- dctw does the transform.
- rfftw_to_NR_format is auxiliary routine for dctw

*/

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void dctw(REAL *fcn,
	  Ftrigtw_Plan *main_plan,
	  int dir
	  ){
  /** Computes the disrete cosine transform of fcn and puts the coefficients 
      back into fcn 

      This is the DCT at half integer points.

      INPUTS:
      - fcn --  array containing the data
      - main_plan -- a structure holding precomputed trig functions,
        the plan used by fftw,and the number of points
	initialized by dctw_init
      - dir  --  the direction to transform.  1 is expansion and 
	-1 evaluation.

      OUTPUT: fcn.  
      - In forward mode (dir==1), puts the coefficients in fcn.
        \\[f(k) = norm \\sum_{j=0}^{numpts-1} 
        f(j) \\cos\\left(\\pi\\frac{k(j+1/2)}{numpts}\\right)
        \\]
        set for $L^2$ normalization.

                 numpts-1              k  (j+0.5)
            fcn[k]= SUM   fcn[j] cos( PI ----------------  ) *norm
	           j=0                         numpts   
                   set for L^2 normalization.
      - In reverse mode inverts this process.

  NOTES:
  - comment better
  - requires even numpts. checked by dctw_init
  - precompute sqrt(2.0)?
  - slow and awkward
*/
  register int i,halfn,numpts;
  REAL *short_save,*long_save,*work_out;
  register REAL sum,sum1,y1,y2,ytemp;

#if ENTEREXIT
  printf("entering dctw\n");
#endif

  /* note number of points */
  numpts=main_plan->numpts;

  /* note half the number of points */
  halfn=numpts>>1;

  /* align the halfn length  trig functions */
  short_save=main_plan->trig_save;

  /* align the numpts length  trig functions */
  long_save=short_save+halfn;

  /* align the workspace */
  work_out=long_save+numpts;

  /* forward   *********************************************************/
  if(dir==1){

    /* some nonsense using short_save */
    for (i=0;i<halfn;i++) {
      y1=0.5*(fcn[i]+fcn[numpts-i-1]);
      y2=short_save[i]*(fcn[i]-fcn[numpts-i -1]);
      fcn[i]=y1+y2;
      fcn[numpts-i-1]=y1-y2;
    }

    /* call the real fft. fcn --> work_out */
    rfftw_one(main_plan->rw_plan_forward,fcn,work_out);

    /* re-order coefficients into NR format; work_out --> fcn */
    rfftw_to_NR_format(fcn,work_out,numpts,1);

    /* some nonsense using long_save */
    for (i=2;i<numpts;i+=2) {
      y1=fcn[i]*long_save[i]-fcn[i+1]*long_save[i+1];
      y2=fcn[i+1]*long_save[i]+fcn[i]*long_save[i+1];
      fcn[i]=y1;
      fcn[i+1]=y2;
    }

    /* include norm with fcn[1]*/
    sum=0.5*fcn[1]*long_save[0];
    for (i=numpts-1;i>0;i-=2) {
      sum1=sum;
      sum += fcn[i];
      fcn[i]=sum1;
    }

    /* include norm with fcn[0]*/
    fcn[0]*=long_save[0]*sqrt(0.5);

    /* reverse   *********************************************************/
  }else if(dir==-1){

    ytemp=fcn[numpts -1];/**/
    for (i=numpts-1;i>2;i-=2) 
      fcn[i]=fcn[i-2]-fcn[i];
    fcn[1]=2.0*ytemp;

    /* include norm with fcn[0/1]*/
    /* ??? remember extra sqrt(2)*/
    fcn[0]*=long_save[0]*sqrt(2.0);
    fcn[1]*=long_save[0];

    /* some nonsense with long_save*/
    for (i=2;i<numpts;i+=2) {
      /* dir caused a sign change here*/
      y1=fcn[i]*long_save[i]+fcn[i+1]*long_save[i+1];
      y2=fcn[i+1]*long_save[i]-fcn[i]*long_save[i+1];
      fcn[i]=y1;
      fcn[i+1]=y2;
    }
       
    /* re-order coefficients into fftw format; fcn --> work_out */
    rfftw_to_NR_format(fcn,work_out,numpts,-1);

    /* call the real fft in reverse; work_out --> fcn */
    rfftw_one(main_plan->rw_plan_reverse,work_out,fcn);

    /* some nonsense with short_save */
    for (i=0;i<halfn;i++) {
      y1=fcn[i]+fcn[numpts-i-1];
      y2=(0.5/short_save[i])*(fcn[i]-fcn[numpts-i-1]);
      /* incorporated an extra factor of 1/2 here to get norm right */
      fcn[i]=0.25*(y1+y2);
      fcn[numpts-i-1]=0.25*(y1-y2);
    }

    /* error   *********************************************************/
  }else
    fprintf(stderr,"dctw: illegal dir=%d\n",dir);

#if ENTEREXIT
  printf("exiting dctw\n");
#endif

}

void rfftw_to_NR_format(REAL *data_nr,
			REAL *data_rfftw,
			int numpts,
			int dir
			){
  /** Translates between the formats for the real discrete
      fourier transform used by the rfftw package and the Numerical Recipies
      routines. This helps allow the NR DCT routine to call the rfftw.

      INPUTS:  
      - data_nr -- The data in the nr format:
        real[0],real[n/2],real[1],imag[1]...
	...real[n/2-1],imag[n/2-1]
      - data_rfftw--The data in the rfftw format:
        real[0],real[1],...real[n/2],
	imag[(n+1)/2-1],....,imag[1]
      - numpts -- the length n of the vectors; MUST be even
      - dir 
        +  1 for rfftw -> nr
	+ -1 for nr -> rfftw
    
      OUTPUTS: 
      - if dir==1, data_nr is written onto
      - if dir==-1, data_rfftw is written onto

      NOTES:
      - we must have REAL==fftw_real
      - there is a sign convention difference on the imaginary parts
        realfft_pre(,,,,1); versus rfftw_one(rw_plan_forward,,);

  **/


  register int i,halfpts;

#if ENTEREXIT
  printf("entering rfftw_to_NR_format\n");
#endif

  /* numpts must be even */
  assert(numpts%2 != 1);

  /* half the number of points */
  halfpts=numpts/2;

  if(dir==1){
    /* data_rfftw -> data_nr */

    /* main real part */
    for(i=0;i<halfpts;i++)
      data_nr[2*i]=data_rfftw[i];

    /* main inaginary part */
    for(i=1;i<halfpts;i++)
      data_nr[2*i+1]=-data_rfftw[numpts-i];

    /* one extra real one */
    data_nr[1]=data_rfftw[halfpts];


  }else if(dir==-1){
    /* data_nr -> data_rfftw */

    /* main real part */
    for(i=0;i<halfpts;i++)
      data_rfftw[i]=data_nr[2*i];

    /* main inaginary part */
    for(i=1;i<halfpts;i++)
      data_rfftw[numpts-i]=-data_nr[2*i+1];

    /* one extra real one */
    data_rfftw[halfpts]=data_nr[1];

  }else
    fprintf(stderr,"rfftw_to_NR_format: illegal dir=%i\n",dir);

#if ENTEREXIT
  printf("exiting rfftw_to_NR_format\n");
#endif

}

void dctw_init(Ftrigtw_Plan *main_plan,
		int flags,
		int numpts
		){
  /** This routine initializes for dctw

     We initialize trigonometic things used by the lct and dct portions,
     and call the initailizer for the fftw real transform (forward).

     INPUTS: 
     - main_plan -- a structure that will hold the things we generate
       Memory for its fields is allocated in this routine.
     - flags -- flags passed directly to the rfftw initializer
     - numpts -- the number of points on which we will transform
       MUST be even

     OUTPUTS: main_plan is filled in:
     

     NOTES:
     - comment better
     
  ********/
  
  register int i,halfn;
  register REAL *local_save;
  /** keep these always as double so we don't lose precision **/
  register double arg, norm;

#if ENTEREXIT
  printf("entering dctw_init\n");
#endif

  /* make sure numpts is even */
  assert(numpts%2==0);

  /* note length of the transform */
  main_plan->numpts=numpts;

  /* some things are half the length */
  halfn=numpts>>1;

  /* does not use complex_work */
  main_plan->complex_work=NULL;

  /* does not use w_plan */
  main_plan->w_plan=NULL;

  /* create rfftw_plan for forward*/
  main_plan->rw_plan_forward=
    rfftw_create_plan(numpts,FFTW_REAL_TO_COMPLEX,flags);

  /* create rfftw_plan for reverse*/
  main_plan->rw_plan_reverse=
    rfftw_create_plan(numpts,FFTW_COMPLEX_TO_REAL,flags);

  /* get memory for trig_save
     numpts/2 real things
     numpts/2 real/imaginary pairs
     numpts workspace
     */
  local_save=(REAL *)calloc(2*numpts+halfn,sizeof(REAL));

  /* align trig_save with local_save */
  main_plan->trig_save=local_save;

  /* intialize trig functions.
     trig_save[2*i]=cos( i*PI/numpts) *norm
     trig_save[2*i+1]=sin( i*PI/numpts) * norm*/

  /* first a shorter sin only part **/
  arg=PI/numpts;
  for(i=0;i<halfn;i++)
    local_save[i]=sin((i+0.5)*arg);

  /* shift local_save to new portion*/
  local_save+=halfn;

  /* compute normalizing factor */
  norm=sqrt(2.0/numpts);
#if 0
  norm=1.0;
#endif

  /* now a real/imaginary pair; include norm**/
  for(i=0;i<numpts;i+=2){
    arg=i*0.5*PI/numpts;
    local_save[i]=cos(arg)*norm;
    local_save[i+1]=sin(arg)*norm;
  }

  /* there is an extra numpts left over as workspace */

#if ENTEREXIT
  printf("exiting dctw_init\n");
#endif
}





  

