/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file direct1_fromfile.c

\brief Load precomputed data for direct1 transform.

Summary:
- load_direct1 loads from file. 
- direct1_fromfile interprets the output of load_direct1 and sets up.

*/

#include "libftsh.h"
 
#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void direct1_fromfile(int *numpts_ptr,
		      int *band_limit_ptr,
		      REAL **node_ptr,
		      REAL **weight_ptr,
		      Pm_Direct_d **pmn_matrix_ptr,
		      Pm_1D_Workspace *pmnwork,
		      FILE *fp
		      ){
  /** This routine loads the initialization dumped by dump_direct1,
      interprets it and does all the set-up necessary to do the
      synthesis or analysis transform.

      INPUTS: fp -- the objects are read from the file pointed to by
      fp, in binary format. The file must be opened before
      calling this routine.

      OUTPUTS: 
      - numpts --  the number of points on the full interval.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
      - node --  an array of length numpts with the nodes,
        usually Arcsin of the usual gaussian nodes.
      - weight -- an array of length numpts with the quadrature weights
      - pmn_matrix -- an array of band_limit (==order_limit)
        Pm_Direct_d structures, each representing the matrix for one m.
      - pmn_work -- a structure holding workspace and other precomputed
        values used by the transform, independent of m

      NOTES:
      - node and weight are not needed for the synthesis/analysis routines
        but are provided here for other uses.

    */

#if ENTEREXIT
  printf("entering direct1_fromfile\n");
#endif

  /* load from file */
  load_direct1(numpts_ptr,band_limit_ptr,node_ptr,weight_ptr,
	       pmn_matrix_ptr,fp);

  /* initialize workspace for synthesizer/analizer */
  init_Pm_1d_workspace(pmnwork,*numpts_ptr,*band_limit_ptr,
		       *node_ptr,*weight_ptr);

#if ENTEREXIT
  printf("exiting direct1_fromfile\n");
#endif
}

void load_direct1(int *numpts_ptr,
		  int *band_limit_ptr,
		  REAL **node_ptr,
		  REAL **weight_ptr,
		  Pm_Direct_d **pmn_matrix_ptr,
		  FILE *fp
		  ){
  /** This routine loads the initialization dumped by dump_direct1 We
      interpret what we read and re-align pointers and such to get
      things ready to use

      INPUTS: fp -- the objects are read from the file pointed to by
      fp, in binary format. The file must be opened before calling
      this routine.

      OUTPUTS: 
      - numpts --  the number of points on the full interval.
      - band_limit -- the bound on the degree of the expansion.
        In P^m_n notation this bounds n.
      - node --  an array of length numpts with the nodes,
        usually Arcsin of the usual gaussian nodes.
      - weights -- an array of length numpts with the quadrature weights
      - pmn_matrix -- an array of band_limit (==order_limit)
        Pm_Direct_d structures, each representing the matrix for one m.

    */

  register int m,order_limit;
  int numpts,band_limit,halfpts;
  Dense_Matrix dmat_mem, *dmat_used;
  Pm_Direct_d *pmn_matrix;
  REAL *node,*weight;

#if ENTEREXIT
  printf("entering load_direct1\n");
#endif
  /* read numpts and band_limit */
  fread(&numpts,sizeof(int),1,fp);
  fread(&band_limit,sizeof(int),1,fp);

  /* order_limit is the same as band_limit */
  order_limit=band_limit;

  /* halfpts is half of numpts */
  halfpts=numpts/2;

  /* read nodes and weights */
  node=(REAL *)calloc(numpts,sizeof(REAL));
  assert(node != NULL);
  fread(node,sizeof(REAL),numpts,fp);
  weight=(REAL *)calloc(numpts,sizeof(REAL));
  assert(weight != NULL);
  fread(weight,sizeof(REAL),numpts,fp);

  /* get all dense matrix  memory */
  dmat_mem=init_dmat_memory(halfpts*(order_limit*(order_limit+1))/2);
  dmat_used=&dmat_mem;

  /* get the pmn_matrix memory */
  pmn_matrix=(Pm_Direct_d *)calloc(order_limit,sizeof(Pm_Direct_d));
  assert(pmn_matrix != NULL);

  /* main martix input.*/
  for(m=0;m<order_limit;m++){
    /* loop through the orders */

    /* give memory to matrix[0] */
    give_dmat_memory(&(pmn_matrix[m].matrix[0]),dmat_used);
    /* load [0] */
    load_bin_dmat(&(pmn_matrix[m].matrix[0]),fp);
    /* re-aim memory */
    dmat_used=&(pmn_matrix[m].matrix[0]);

    /* give memory to matrix[1] */
    give_dmat_memory(&(pmn_matrix[m].matrix[1]),dmat_used);
    /* load [1] */
    load_bin_dmat(&(pmn_matrix[m].matrix[1]),fp);
    /* re-aim memory */
    dmat_used=&(pmn_matrix[m].matrix[1]);
  }

  /* set parameters for output */
  (*numpts_ptr)=numpts;
  (*band_limit_ptr)=band_limit;
  (*node_ptr)=node;
  (*weight_ptr)=weight;
  (*pmn_matrix_ptr)=pmn_matrix;

#if ENTEREXIT
  printf("exiting load_direct1\n");
#endif
}
