/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** \file direct1_tofile.c

\brief Precompute and save data for direct1 spherical harmonic transform.

Summary:
- direct1_tofile drives the initialization and saving.
- init_direct1 does the actual initialization.
- dump_direct1 does the actual saving.

*/

#include "libftsh.h"
 
#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void direct1_tofile(int  numpts,
		    int band_limit,
		    REAL *node,
		    REAL *weight,
		    FILE *fp
		    ){
  /** This routine initializes the direct transform for spherical harmonics
      (version 1) and dumps it to a file.

      INPUTS: 
      - numpts --  the number of points on the full interval. 
        Must be divisible by 2.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
      - node --  an array of length numpts with the nodes,
        usually Arcsin of the usual gaussian nodes.
      - weight -- an array of length numpts with the quadrature weights

      OUTPUTS: fp - the objects created are written onto the file pointed to by
      fp, in binary format. The file must be opened before
      calling this routine.

      NOTES:
      - node and weight are input so the user can select.
  */
  int order_limit,halfpts;
  Pm_Direct_d *pmn_matrix;

#if ENTEREXIT
  printf("entering direct1_tofile\n");
#endif

  /* set some global parameters*/

  /* set halfpts */
  halfpts=numpts/2;

  /* set order_limit to band_limit. */
  order_limit=band_limit;

  /* main initialization ************************************************/

  init_direct1(&pmn_matrix,halfpts, band_limit,node);

  /* Dump to file */

  /* dump using subroutine */
  dump_direct1(numpts,band_limit,node,weight,pmn_matrix,fp);

  /* free everything that we can**/

  /* free the matrices. This depends on init_direct1 giving the main 
     memory to pmn_matrix[0].matrix[0] */
  free_dmat_memory(&(pmn_matrix[0].matrix[0]));
  
  /* free the global holder */
  free(pmn_matrix);


#if ENTEREXIT
  printf("exiting direct1_tofile\n");
#endif

}

void init_direct1(Pm_Direct_d **out,
		  int halfpts,
		  int band_limit,
		  REAL *node
		  ){
  /** This routine initializes for the transform for spherical
      harmonics, direct version 1.  This version simple precomputes the
      Pm matrices and stores them in dense matrix format.

      INPUTS: 
      - halfpts -- half the number of points on the full interval.
        This implies the full number of points is even,
	and we also assume the nodes are symmetric.
	Currrently halfpts must also be even for our trig
	transforms to work.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
	For gaussian nodes this must be < 2*halfpts
	and for other nodes it should be even smaller.
	We confirn band_limit < 2*halfpts but leave the rest to the user.
      - node --  an array with the nodes, usually Arcsin of the usual
        gaussian nodes. We use halfpts of these, representing 
	half of the interval.
 
     OUTPUTS: out -- memory is allocated for band_limit
     (==order_limit) Pm_Direct_d structures, each representing the
     matrix for one m. Various bits of memory are allocated for these,
     and values filled in.

  */

  register int m,i,parity,order_limit;
  int numbas[2];
  Pm_Direct_d *pmn_matrix;
  Pmncos_Recurrence_Save pmn_save;
  Dense_Matrix dmat_mem,*dmat_used;

#if ENTEREXIT
  printf("entering init_direct1\n");
#endif

  /* order_limit should be the same as band_limit */
  order_limit=band_limit;

  /* memory for main compressed matrix holder */
  pmn_matrix=(Pm_Direct_d *)calloc(order_limit,sizeof(Pm_Direct_d));
  assert(pmn_matrix != NULL);

  /* get memory for and initialize Pmn recurrence */
  Pmncos_resave_init(&pmn_save,node,2*halfpts);

  /* memory holder for all the Dense_Matrices */
  /* we need halfpts*(order_limit*(order_limit+1))/2 total */
  dmat_mem=init_dmat_memory(halfpts*(order_limit*(order_limit+1))/2);
  
  /* align our memory passer */
  dmat_used=&dmat_mem;

  /*** begin main (order) loop *********************************************/
  for(m=0;m<order_limit;m++){
    /* loop through the order */
#if 0
    printf("m=%i\n",m);
#endif
    /* construct the matrix for that order */
    
    /* set the number of basis elements. There may be one less odd one*/
    numbas[0]=(order_limit-m+1)/2;
    numbas[1]=(order_limit-m)/2;

    for(parity=0;parity<2;parity++){
      /* loop through the two parities */

      /* give memory to the [parity] matrix */
      give_dmat_memory(&(pmn_matrix[m].matrix[parity]),dmat_used);

      /* set [parity] matrix size */
      pmn_matrix[m].matrix[parity].n_col=halfpts;
      pmn_matrix[m].matrix[parity].n_row=numbas[parity];

      /* re-align dmat_used */
      dmat_used=&(pmn_matrix[m].matrix[parity]);
    }

	
    for(i=0;i<numbas[1];i++){
      /* loop through the basis elements, by two's*/

      /* construct the even basis function*/
      next_rtsinPmncos(pmn_matrix[m].matrix[0].matrix+i*halfpts,
		       &pmn_save,halfpts,m,2*i);

      /* construct the odd basis function*/
      next_rtsinPmncos(pmn_matrix[m].matrix[1].matrix+i*halfpts,
		       &pmn_save,halfpts,m,2*i+1);
    }

    if(numbas[0] > numbas[1]){
      /*  there is one more even function*/

      /* construct the even basis function*/
      next_rtsinPmncos(pmn_matrix[m].matrix[0].matrix+i*halfpts,
		       &pmn_save,halfpts,m,2*i);
    }

    /* optionally allow change of sign convention */

  } /* end order loop */

  /* free the things we don't need any more */

  /* don't need pmn_save */
  Pmncos_resave_destroy(&pmn_save);

  /* align out */
  (*out)=pmn_matrix;

#if ENTEREXIT
  printf("exiting init_direct1\n");
#endif
}

void dump_direct1(int numpts,
		  int band_limit,
		  REAL *node,
		  REAL *weight,
		  Pm_Direct_d *pmn_matrix,
		  FILE *fp
		  ){
  /** This routine dumps the initialization constructed by init_direct1 and
      general parameters for the direct transform for spherical harmonics,
      dense version.

      INPUTS: 
      - numpts --  the number of points on the full interval.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
      - node --  an array of length numpts with the nodes,
        usually Arcsin of the usual gaussian nodes.
      - weights -- an array of length numpts with the quadrature weights
      - pmn_matrix -- an array of band_limit (==order_limit)
        Pm_Direct_d structures, each representing the 
	matrix for one m.

      OUTPUTS: fp -- the objects above are written into the file
      pointed to by fp, in binary format. The file must be opened
      before calling this routine.

  */

  register int m,order_limit;

#if ENTEREXIT
  printf("entering dump_direct1\n");
#endif

  /* order_limit is the same as band_limit */
  order_limit=band_limit;

  /* dump numpts and band_limit */
  fwrite(&numpts,sizeof(int),1,fp);
  fwrite(&band_limit,sizeof(int),1,fp);

  /* dump nodes and weights */
  /* we could probably just dump halfpts of them */
  fwrite(node,sizeof(REAL),numpts,fp);
  fwrite(weight,sizeof(REAL),numpts,fp);

  /* main martix output.*/

  for(m=0;m<order_limit;m++){
    /* loop through the orders */

    /* dump matrix[0] and [1] */
    dump_bin_dmat(&(pmn_matrix[m].matrix[0]),fp);
    dump_bin_dmat(&(pmn_matrix[m].matrix[1]),fp);

  }

#if ENTEREXIT
  printf("exiting dump_direct1\n");
#endif
}
