/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** \file direct1_transform.c

\brief Perform the direct1 spherical harmonic transform.

Summary:
- Pm_direct1_analyse transforms function values to coefficients.
- Pm_direct1_synthesize transforms coefficients to function values.

*/

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void Pm_direct1_analyse(REAL *outcoef,
			Pm_Direct_d *inmat,
			REAL *infcn,
			Pm_1D_Workspace *pmnwork
			){
  /** Expand (analyses) a function into spherical harmonic coefficients. 

  This routine only does one order (m) so it is really only an
  Associated Legendre Function analyser.

  INPUTS: 
  - inmat -- The precomuted matrix of Associated Legendre functions
    of the desired order in Pm_Direct_d format
  - infcn -- an array of REALs to hold the function to expand.
    Its length is encoded in the number of columns in inmat
  - pmnwork -- a structure that holds workspace and a few global
    precomputed values
    
  OUTPUTS: outcoef -- an array of the coefficients computed. The
  information on how many of these to compute is encoded by the number
  of rows in inmat.

  NOTES:
  - We could check that (numbas[0]==numbas[1])||(numbas[0]==numbas[1]+1)
  - ?UNROLL the parity LOOP ONCE DEBUGGED? 

  */
  register int parity,i;
  int numbas[2],numpts;

#if ENTEREXIT
  printf("entering Pm_direct1_analyse\n");
#endif

  /* deduce number of basis functions. There may be one less odd one */
  numbas[0]=inmat->matrix[0].n_row;
  numbas[1]=inmat->matrix[1].n_row;

  /* deduce numpts */
  numpts=inmat->matrix[0].n_col+inmat->matrix[1].n_col;

  /* check for compatability with pmnwork.*/
  assert(numpts == pmnwork->numpts);

  /* we must have 
     pmnwork->dvect_fcn[0].n_row=(numpts+1)/2;
     pmnwork->dvect_fcn[1].n_row=numpts/2;
     This is set when dvect_fcn is initialized and
     confirmed within dmat_dvect_multiply.*/

  /* check with pmnwork for sufficient space in dvect_coef */
  assert(numbas[0]+numbas[1] <= pmnwork->band_limit);


  /* reflect infcn into even and odd components */
  reflect_evenodd(pmnwork->dvect_fcn[0].vector,
		  pmnwork->dvect_fcn[1].vector,
		  infcn,numpts,1);

  /* UNROLL THIS LOOP ONCE DEBUGGED */
  for(parity=0;parity<2;parity++){
    /* loop through even versus odd basis functions */
    
    /* incorporate (modified) gaussian weights into function */
    /* adjust size since [0] may be 1 longer than [1] */
    pointwise_multiply(pmnwork->dvect_fcn[parity].vector,
		       pmnwork->dvect_fcn[parity].vector,
		       pmnwork->mod_weight,
		       (numpts+1-parity)/2);
    
    /* apply dense matrix to function, matching parities */
    dmat_dvect_multiply(pmnwork->dvect_coef+parity,
			 &(inmat->matrix[parity]),
			 pmnwork->dvect_fcn+parity);
  }

  /* reorganize even and odd parity coefficients into desired output format*/
  for(i=0;i<numbas[1];i++){
    /* loop through the basis elements, by two's*/

    /* note the even coefficient*/
    outcoef[2*i]=pmnwork->dvect_coef[0].vector[i];
    
    /* note the odd coefficient*/
    outcoef[2*i+1]=pmnwork->dvect_coef[1].vector[i];
  }
  
  if(numbas[0] > numbas[1]){
    /*  there is one more even function*/
    
    /* note the even coefficient*/
    outcoef[2*i]=pmnwork->dvect_coef[0].vector[i];
  }

#if ENTEREXIT
  printf("exiting Pm_direct1_analyse\n");
#endif  
}

void Pm_direct1_synthesize(REAL *outfcn,
			   Pm_Direct_d *inmat,
			   REAL *incoef,
			   Pm_1D_Workspace *pmnwork
			   ){
  /** Evaluates (synthesizes) a function from spherical harmonic coefficients.

  This routine only does one order (m) so it is really only an
  Associaded Legendre Function synthesizer.

  INPUTS: 
  - inmat -- The precomuted matrix of Associated Legendre functions
    of the desired order in Pm_Direct_d format
  - incoef -- an array of the coefficients to use. The information
    on how many of these to use is encoded by the number of rows in inmat.
  - pmnwork -- a structure that holds workspace and a few global
    precomputed values
    
  OUTPUTS: outfcn -- an array of REALs to hold the computed function.
  Its length is encoded in the number of columns in inmat

  NOTES:
  - We could check that (numbas[0]==numbas[1])||(numbas[0]==numbas[1]+1)
  */
  register int parity,i;
  int numbas[2],numpts;

#if ENTEREXIT
  printf("entering Pm_direct1_synthesize\n");
#endif

  /* deduce number of basis functions. There may be one less odd one */
  numbas[0]=inmat->matrix[0].n_row;
  numbas[1]=inmat->matrix[1].n_row;

  /* deduce numpts */
  numpts=inmat->matrix[0].n_col+inmat->matrix[1].n_col;

  /* check for compatability with pmnwork.*/
  assert(numpts == pmnwork->numpts);

  /* reorganize even and odd parity coefficients into desired input format*/
  
  /* set the sizes for the auxillary dense vectors */ 
  pmnwork->dvect_coef[0].n_row=numbas[0];
  pmnwork->dvect_coef[1].n_row=numbas[1];

  /* check with pmnwork for sufficient space */
  assert(numbas[0]+numbas[1] <= pmnwork->band_limit);

  for(i=0;i<numbas[1];i++){
    /* loop through the basis elements, by two's*/
    
    /* note the even coefficient*/
    pmnwork->dvect_coef[0].vector[i]=incoef[2*i];

    /* note the odd coefficient*/
    pmnwork->dvect_coef[1].vector[i]=incoef[2*i+1];
  }

  if(numbas[0] > numbas[1]){
    /*  there is one more even function*/

    /* note the even coefficient*/
    pmnwork->dvect_coef[0].vector[i]=incoef[2*i];
  }

  /* UNROLL THIS LOOP ONCE DEBUGGED */
  for(parity=0;parity<2;parity++){
    /* loop through even versus odd basis functions */
    
    /* apply TRANSPOSE of the dense matrix to the coefficients,
       matching parities. */
    dmat_t_dvect_multiply(pmnwork->dvect_fcn+parity,
			  &(inmat->matrix[parity]),
			  pmnwork->dvect_coef+parity);
  }
  
  /* un-reflect data from even and odd components */
  reflect_evenodd(pmnwork->dvect_fcn[0].vector,
		  pmnwork->dvect_fcn[1].vector,outfcn,numpts,-1);
  
  /* remove excess sqrt(sin)*/
  pointwise_multiply(outfcn,outfcn,pmnwork->over_rootsin,numpts);
  
#if ENTEREXIT
  printf("exiting Pm_direct1_synthesize\n");
#endif  
}
