/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

/** \file dmat.c
    
\brief Contains simple functions based on the Dense_Matrix struct.

Summary:
- init_dmat_memory allocates a block of memory.
- give_dmat_memory passes extra memory to the next Dense_Matrix.
- free_dmat_memory frees a block of memory.
- dump_bin_dmat writes to file in binary.
- load_bin_dmat reads back what dump_bin_dmat wrote.

*/

Dense_Matrix init_dmat_memory(unsigned  int n
			       ){
  /** This routine gets a large amount of memory for Dense_Matrix's,
      and gives it to a single Dense_Matrix to hold. This special
      Dense_Matrix acts as a pointer to this block of memory, and
      also keeps track of how large this memory is.

      INPUT:  n -- the amount of memory to get

      OUPUT: returns a Dense_Matrix with
      - .max_size=n
      - .n_row=0
      - .n_col=0
      - .matrix pointing to memory for n REAL's

  */

  Dense_Matrix dmat;

#if ENTEREXIT
  printf("enter init_dmat_memory\n");
#endif

  /* set max_size*/
  dmat.max_size=n;

  /* empty Matrix */
  dmat.n_row=0;
  dmat.n_col=0;

  /* main memory for the Matrix*/
  dmat.matrix=(REAL *)calloc(n,sizeof(REAL));
  assert(dmat.matrix != NULL);

#if ENTEREXIT
  printf("return from init_dmat_memory\n");
#endif

  return dmat;
}

void give_dmat_memory(Dense_Matrix *current,
		      Dense_Matrix *previous
		      ){
  /** This routine sets up memory for the next Dense_Matrix we wish to use.
     
  It looks at the previous Dense_Matrix we used, and increments fields

  INPUT: previous -- the last Dense_Matrix we used.

  OUTPUT: current -- the Dense_Matrix we wish to use next.
          We align current->matrix to the first space not used
	  by previous->matrix, and set current->max_size to 
	  account for the amount used in previous.
    
  */
#if ENTEREXIT
  printf("entering give_dmat_memory\n");
#endif

  current->n_row=0; /* not used */
  current->n_col=0; /* not used */

  /* align matrix to next unused memory */
  current->matrix=previous->matrix + (previous->n_row)*(previous->n_col);

  /* adjust max_size to account for memory used */
  current->max_size=previous->max_size - (previous->n_row)*(previous->n_col);

#if ENTEREXIT
  printf("exiting give_dmat_memory\n");
#endif
}

void free_dmat_memory(Dense_Matrix *in
		      ){
  /** This routine frees the memory that was given to a Dense_Matrix
      by init_dmat_memory.

      INPUT: in -- the Dense_Matrix to free.  it must have been
      initialized by init_dmat_memory, or we will attempt to free
      things not gotten by calloc, probaby causing catastrophic
      failure.  

      NOTES: 
      - If used incorrectly, this routine may yield nasty memory bugs.
  */

#if ENTEREXIT
  printf("enter free_dmat_memory\n");
#endif

  /* free memory from matrix */
  free(in->matrix);

#if ENTEREXIT
  printf("exiting free_dmat_memory\n");
#endif

}

void dump_bin_dmat(Dense_Matrix *x,
		   FILE *out
		   ){
  /** This routine dumps a Dense_Matrix to a file in binary format.
     
  This is efficient, but not human readable.
     
  INPUTS: 
  - x -- the Dense_Matrix
  - out -- A pointer to the file to write into. 
    Setting this to stdout makes it write to the screen.
    If out==NULL we use stdout.
    
  OUTPUT: writes into the file given by out. The writing mode ("wb" or "ab")
          is determined outside of this routine.

  */

  register int i,j;

#if ENTEREXIT
  printf("entering dump_bin_dmat\n");
#endif

  /* default is stdout (send to screen)*/
  if(out==NULL)
    out=stdout;

  /* max_size will not be relevent, so we don't write it.*/

  /* number of rows and columns in the matrix */
  fwrite(&(x->n_row),sizeof(int),1,out);
  fwrite(&(x->n_col),sizeof(int),1,out);

  /* output the matrix entries */
  fwrite(x->matrix,sizeof(REAL),(x->n_row)*(x->n_col),out);

#if ENTEREXIT
  printf("exiting dump_bin_dmat\n");
#endif
}

void load_bin_dmat(Dense_Matrix *x,
		   FILE *in
		   ){
  /** This routine loads a Dense_Matrix from a file in binary format,
      as written by dump_bin_dmat.
     
      INPUTS: 
      - x -- the Dense_Matrix to be filled in. Should be initialized with
        + .max_size indicating the amount of memory available
	+ .matrix pointing to max_size REALs
      - in -- A pointer to the file to read from. 
        Setting this to stdin makes it read from the screen.
	If in==NULL we use stdin.
	The reading mode ("rb") 
	is determined outside of this routine.
	
      OUTPUT: x -- is filled in.

  */

  register int i,j;

#if ENTEREXIT
  printf("entering load_bin_dmat\n");
#endif

  /* default is stdin (read from screen)*/
  if(in==NULL)
    in=stdin;

  /* number of rows and columns in the matrix */
  fread(&(x->n_row),sizeof(int),1,in);
  fread(&(x->n_col),sizeof(int),1,in);

  /* make sure there is enough room */
  assert(x->max_size >= (x->n_row)*(x->n_col));

  /* input the matrix entries */
  fread(x->matrix,sizeof(REAL),(x->n_row)*(x->n_col),in);

#if ENTEREXIT
  printf("exiting load_bin_dmat\n");
#endif
}





