/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

/** \file dmat_ops.c
    
\brief Contains functions interacting_the Dense_Matrix struct with others.

Summary:
- dmat_dvect_multiply applies to a Dense_Vector
- dmat_t_dvect_multiply applies the transpose to a Dense_Vector

*/

void dmat_dvect_multiply(Dense_Vector *out,
			 Dense_Matrix *inmat,
			 Dense_Vector *invect
			 ){
  /** Apply a matrix to a vector to get a new vector.
      The matrix is in Dense_Matrix format, 
      and the vectors in Dense_Vector format.

      INPUTS: 
      - out -- the Dense_Vector to put the result in.
        Should be initialized as follows:
	+ .max_size is the amount of memory available.
	+ .vector points to .max_size REAL
      - invect -- an input vector in Dense_Vector format
      - inmat -- the input matrix in Dense_Matrix format

      OUTPUT: out -- inmat * invect = out

      NOTES:
      - ordering of loops not optimized
      - ? use registers for n_row n_col?
      - blocking could help speed
  */

  register int i,k;
  register REAL tmp_real;

#if ENTEREXIT
  printf("entering dmat_dvect_multiply\n");
#endif

  /* check compatibility of sizes*/
  assert(inmat->n_col==invect->n_row);

  /* set sizes for out*/
  out->n_row=inmat->n_row;

  /* make sure there is room in out **/
  assert(out->max_size >= out->n_row);

  /* main multiplication ***/

  for(i=0;i<inmat->n_row;i++){
    /* loop through rows **/
    /* clear register */
    tmp_real=0;

    for(k=0;k<inmat->n_col;k++){
      /* loop through columns  **/
      /* add contribution to the register accumulant **/
      tmp_real+=inmat->matrix[k + i*inmat->n_col]*
	invect->vector[k];
    }
    /* transfer accumulated value to proper place */
    out->vector[i]=tmp_real;
  }


#if ENTEREXIT
  printf("exiting dmat_dvect_multiply\n");
#endif
}

void dmat_t_dvect_multiply(Dense_Vector *out,
			   Dense_Matrix *inmat,
			   Dense_Vector *invect
			   ){
  /** Apply the TRANSPOSE of a matrix to a vector to get a new vector.
      The matrix is in Dense_Matrix format, 
      and the vectors in Dense_Vector format.

      INPUTS:
      - out -- the Dense_Vector to put the result in.
        Should be initialized as follows:
	+ .max_size is the amount of memory available.
	+ .vector points to .max_size REAL
      - invect -- an input vector in Dense_Vector format
      - inmat -- the input matrix in Dense_Matrix format

      OUTPUT: out -- TRANSPOSE(inmat) * invect = out
       
      NOTES:
      - ordering of loops not optimized
      - blocking may improve speed
  */

  register int i,k;
  register REAL tmp_real;

#if ENTEREXIT
  printf("entering dmat_t_dvect_multiply\n");
#endif

  /* check compatibility of sizes; remember transposing*/
  assert(inmat->n_row==invect->n_row);

  /* set sizes for out; remember transposing*/
  out->n_row=inmat->n_col;

  /* make sure there is room in out **/
  assert(out->max_size >= out->n_row);


  /* clear out */
  for(i=0;i<out->n_row;i++){
    out->vector[i]=0.0;
  }

  /* main multiplication ***/

  for(i=0;i<inmat->n_row;i++){
    /* loop through rows **/

    /* store the vector value as a register */
    tmp_real=invect->vector[i];

    for(k=0;k<inmat->n_col;k++){
      /* loop through columns  **/
      out->vector[k]+=inmat->matrix[k + i*inmat->n_col]
	*tmp_real; /* invect->vector[i] */
    }
  }
#if ENTEREXIT
  printf("exiting dmat_t_dvect_multiply\n");
#endif
}



