/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

/** \file dyadic_gsearch.c

\brief Best non-decreasing dyadic partition search.

Summary:
- dyadic_gsearch is the actual search
- dyadic_gsearch_init initializes a structure
- dyadic_gsearch_free frees this structure

The memory for all the objects used in the search are held in a
structure of type Dyadic_Gsearch_Save. Declare one of these:

    Dyadic_Gsearch_Save dy_gs_save;

To initialize this structure, call 

    dyadic_gsearch_init(&dy_gs_save,num_levels);

with num_levels the number of dyadic levels to use.

Then compute whatever basis you are using and load in the costs:
costfull[level][within] should contain the full bell cost of the
within^th interval the level^th level (2^level intervals on level).
Similarly for costasym.

Then call this routine to do the search itself. Returns the minimal
cost, which is also stored in
dy_gs_save.costfull[dy_gs_save.best_level][0].

    dyadic_gsearch(&dy_gs_save);

You can then do other basis expansions (with the same num_levels) and
searchs.

When you are ready to free the memory, call

    dyadic_gsearch_free(&dy_gs_save);

*/

int dyadic_gsearch(Dyadic_Gsearch_Save *dy_gs_save
		   ){
  /** This routine implements a `best non-decreasing dyadic partition'
      search.  It finds the partition with lowest cost under the
      restriction that an interval's right neighbor is either the same
      size or one size (2x) larger.  Generally this is to be used with
      local cosine. When an interval's left neighbor is the same size,
      we use a regular full sized bell and put the cost of this
      expansion in .cost_full[][]. When the left neighbor is smaller
      we shrink the left edge of the bell, making it asymmetric and
      yielding .cost_asym[][].
 
      The search tree is grown from the left, with an interval having
      children its allowed right neighbors. The cost of a partition is
      the cost of a path from the left edge to the right edge. Since
      each interval has two costs, the path into an interval is also
      important.
                                                                        
                                                                        
             | _--------------O                   |        
             |/                                   |                     
            /|  _-----O-------|--------->O        |                       
           O-|-/           _--|--------/          |
            \|       |    /   |         |         |                     
             |~--O---|-->O----|--->O----|---->O   |                        
                                                                      
      The generic decision step is: assume the current node's children
      contain the minimal cost of a path from them to the
      right. Compare these costs and ADD the smaller to the full and
      asym costs of the current node.  We process intervals in order
      of the x-coordinate of their center, starting at the right.

      INPUTS: dy_gs_save -- a structure holding memory for the costs, branching
      and ordering of interval. See the declaration of
      the structure Dyadic_Gsearch_Save for the full 
      description of its fields.
    
      OUTPUTS: returns the minimal cost, and also stores it in
      dy_gs_save.costfull[dy_gs_save.best_level][0]. To determine the
      partition chosen, start at [best_level][0] and follow
      .branching[][], which has values one of {RIGHTEDGE=0,
      SAMESIZE=1, UPSIZE=2}

  */

  register int s,numlevels;
  int bestlev;
  unsigned int owiths,olevs;
  enum {RIGHTEDGE=0, SAMESIZE=1, UPSIZE=2};
  unsigned short int **branching;
  unsigned int **costfull,**costasym,*orderlev,*orderwithin;

#if ENTEREXIT
  printf("entering dyadic_gsearch\n");
#endif

  /* align local variables with structure variables.
     This makes the code easier to read and might be more efficient*/
  costfull=dy_gs_save->cost_full;
  costasym=dy_gs_save->cost_asym;
  branching=dy_gs_save->branching;
  orderwithin=dy_gs_save->order_within;
  orderlev=dy_gs_save->order_level;

  /* note number of levels */
  numlevels=dy_gs_save->num_levels;

  /* We now do all the comparisons, building paths and cutting them off.
     We begin at the end of order__ and work back.  For each block we:
     1. find its 'parent', the interval of the same size to the left
     2. decide if it has a big brother
     3. compare: add best cost to parent's cost
                  set branch to 1 if same size won, 2 if big bro' won
     4. do the same for its asymmetric parent
     Although costasym[0] and costasym[numlevels-1] are dead ends,
     we compute them also, to simplify the algorithm*/

  for(s=(1<<numlevels)-2;s>=0;s--){
    /* loop through all intervals in the search order, starting at the right*/

    /* determine the within value of the interval*/
    owiths=orderwithin[s];

    if(owiths){   
      /* isn't at left edge  (or top) */
      
      /* determine the level value of the interval*/
      olevs=orderlev[s];

      if(owiths%2){   /* has no big brother */ 

	costfull[olevs][owiths-1]+=costfull[olevs][owiths];
	costasym[olevs][owiths-1]+=costfull[olevs][owiths];
	branching[olevs][owiths-1]=SAMESIZE;

      }else{                 /* has big brother */
	if(costfull[olevs][owiths]
	   <costasym[olevs-1][owiths>>1]){ /*beats big bro'*/

	  costfull[olevs][owiths-1]+=costfull[olevs][owiths];
	  costasym[olevs][owiths-1]+=costfull[olevs][owiths];
	  branching[olevs][owiths-1]=SAMESIZE;

	}else{     /* brother wins */
	  costfull[olevs][owiths-1]+=costasym[olevs-1][owiths>>1];
	  costasym[olevs][owiths-1]+=costasym[olevs-1][owiths>>1];
	  branching[olevs][owiths-1]=UPSIZE;
	}
      }
    }/* end not at left edge*/
  }

  /* This has done all the comparisons, except comparing the trees 
     which begin at different levels.  Their best results are in 
     costfull[level][0]*/
  bestlev=0;
  for(s=1;s<numlevels;s++){
    /* loop through the rest of the levels */
    
    if(costfull[bestlev][0]>costfull[s][0])
      /*best one so far has been beaten*/
      bestlev=s;
  }

  /* set best_level**/
  dy_gs_save->best_level=bestlev;

#if ENTEREXIT
  printf("returning from dyadic_gsearch\n");
#endif

  /* return our minimal cost **/
  return costfull[bestlev][0];
}

void dyadic_gsearch_init(Dyadic_Gsearch_Save *dy_gs_save,
			 int num_levels
			 ){
  /** This function initializes the structure Dyadic_Gsearch_Save
      for use by dyadic_gsearch.  Memory can be freed by dyadic_gsearch_free.

      INPUTS:  
      - dy_gs_save -- a pointer to the structure to fill in
      - num_levels -- the number of dyadic levels to use

      OUTPUT:  dy_gs_save -- memory is allocated and some things filled in.
  */

  register int level,s,k;

#if ENTEREXIT
  printf("entering dyadic_gsearch_init\n");
#endif

  /* note the number of levels */
  dy_gs_save->num_levels=num_levels;

  /**************************************************************************/
  /* initialize memory **/
  /* this is memory and set up for the best basis chooser.*/

  /* first index for branching */
  dy_gs_save->branching=(unsigned short int **)
    calloc(num_levels,sizeof(unsigned short int*));
  assert(dy_gs_save->branching != NULL);

  /* first index for cost_full */
  dy_gs_save->cost_full=(unsigned int **)
    calloc(num_levels,sizeof(unsigned int*));
  assert(dy_gs_save->cost_full != NULL);

  /* first index for cost_asym */
  dy_gs_save->cost_asym=(unsigned int **)
    calloc(num_levels,sizeof(unsigned int*));
  assert(dy_gs_save->cost_asym != NULL);

  /* second index for all three.
     It would be better to get a large chunk of memory and divide it up,
     but that would be confusing */

  for(level=0;level<num_levels;level++){
    /* loop through the levels */

    /* note 2^level, the memory for that level*/
    k=1<<level;

    dy_gs_save->cost_full[level]=(unsigned int *)
      calloc(k,sizeof(unsigned int));
    assert(dy_gs_save->cost_full[level] != NULL);
    dy_gs_save->cost_asym[level]=(unsigned int *)
      calloc(k,sizeof(unsigned int));
    assert(dy_gs_save->cost_asym[level] != NULL);
    dy_gs_save->branching[level]=(unsigned short int *)
      calloc(k,sizeof(unsigned short int));
    assert(dy_gs_save->branching[level] != NULL);
  } 
  

  /* We also have an array to tell us in which order to do the comparisons
     Take coef_full, the matrix of coeficients.  Represent each interval of
     coef by the point in its center.  Project this to a horizontal line.
     The intervals are now in the correct order.*/

  /* memory for order_level */
  dy_gs_save->order_level=(unsigned int *)
    calloc((1<<num_levels)-1,sizeof(unsigned int));
  assert(dy_gs_save->order_level != NULL);
    
  /* memory for order_within */
  dy_gs_save->order_within=(unsigned int *)
    calloc((1<<num_levels)-1,sizeof(unsigned int));
  assert(dy_gs_save->order_within != NULL);
  
  /* fill in both */

  for(level=0;level<num_levels;level++){
    /* loop through the levels */
    
    for(s=0;s<(1<<level);s++){
      /* loop through the intervals on that level */

      /* write the current level and within on the appropriate address */
      dy_gs_save->order_level[s*(1<<(num_levels-level))
			     +(1<<(num_levels-level-1))-1]
	=level;

      dy_gs_save->order_within[s*(1<<(num_levels-level))
			   +(1<<(num_levels-level-1))-1]
	=s;
    }
  }

#if ENTEREXIT
  printf("exiting dyadic_gsearch_init\n");
#endif
}

void dyadic_gsearch_free(Dyadic_Gsearch_Save *dy_gs_save
			 ){
  /** This function frees the structure Dyadic_Gsearch_Save
      used by dyadic_gsearch and initialized by dyadic_gsearch_init.

      INPUTS:  dy_gs_save -- a pointer to the structure to free

      OUTPUT:  none
  */

  register int level;

#if ENTEREXIT
  printf("entering dyadic_gsearch_free\n");
#endif

  /**************************************************************************/
  /* free memory **/

  /* second index for branching, cost_full, cost_asym*/

  for(level=0;level<dy_gs_save->num_levels;level++){
    /* loop through the levels */

    /* free memory for that level*/
    free(dy_gs_save->cost_full[level]);
    free(dy_gs_save->cost_asym[level]);
    free(dy_gs_save->branching[level]);
  } 
  
  /* first index for branching, cost_full, cost_asym*/
  free(dy_gs_save->branching);
  free(dy_gs_save->cost_full);
  free(dy_gs_save->cost_asym);

  /* only index for order_level, order_within */
  free(dy_gs_save->order_level);
  free(dy_gs_save->order_within);

#if ENTEREXIT
  printf("exiting dyadic_gsearch_free\n");
#endif
}





