/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file dyadic_trig_gsearch.c

\brief Local trig version (wrapper) for dyadic_gsearch.c.

Summary:
- dyadic_g_ilist_init initializes all the intervals needed.
- ltrigint_init initializes a single interval.
- gsearch_to_ilist unpacks the results of the search.

*/

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void dyadic_g_ilist_init(Local_Trig_Interval *ilist[2],
			 int *ints_used,
			 int *coefs_used,
			 REAL **bell_handle,
			 int *bell_length,
			 int **bell_offsets,
			 int halfpts,
			 int num_levels,
			 REAL (*bell_choice) (int,int),
			 int flags
			 ){
  /** This routine constructs the interval list for a dyadic partition
      suitable for the growing interval size search (gsearch).

      INPUTS:  
      - ilist -- an array of 2 pointers to Local_Trig_Interval,
        to be filled in.
      - halfpts -- the number of points on which to construct these
        intervals. For the growing search, this usually 
	represents half of the number of points in the full interval.
      - num_levels -- the number of dyadic levels to use. We need
        halfpts >= (1<<num_levels) and halfpts % (1<<num_levels) == 0
      - bell_choice-- the bell function to use
      - flags -- integer flags passed to the fftw initializer


      OUTPUTs: 
      - ilist[0/1] -- has memory allocated and is filled in
        + ilist[0] -- gets intervals with trig functions for even
	  parity at the right edge
	+ ilist[1] -- gets intervals with trig functions for odd
	  parity at the right edge
      - ints_used -- is filled in with the length of the ilist's
      - coefs_used -- is filled in with the total number of possible
        coefficients this ilist can produced. This is used 
	to allocate memory and set offsets within whatever
	array will eventually hold these coefficients
      - bell_handle -- is pointed to an array which contains all
        the bell values. This is used when we want to 
	save the bells to disc efficiently.
      - bell_length -- is set to the length of bell_handle.
      - bell_offsets -- is pointed to an array containing integer offsets
      withing bell_handle, indicating the location
      of particular bells.
   
      NOTES:
      - coefs_used includes offsets for the left edge asymmetric bells,
        which are not used.
      - We now use the bell memory efficiently, with no repeated computations.
        For the trig transforms, we do not repeat the trig value or fftw
	computations. There is an inefficiency in that we have multiple 
	versions of the workspace. This adds trivial memory, but may 
	have cache implications
      - There was a mysterious segmentation fault possibly related
        to free(plan_holder) so that is now disabled. 
	This may cause a memory leak.

*/

  register int level,within,interval;
  int num_int_dyadic,winlength,n_int_lev;

  REAL ***bell_holder,*bell_mem;
  int bell_total, *offset;
  Ftrigtw_Plan *plan_holder;

#if ENTEREXIT
  printf("entering dyadic_g_ilist_init\n");
#endif

  /* make sure halfpts can handle that many levels.
     we need 2^(num_levels-1) to divide halfpts so we can have
     num_levels worth of dyadic splits. (num_levels==1 means no splits).
     We also need one extra factor of 2 for lctw, and so the bells can spill
     as designed. */
  assert(halfpts >= (1<<num_levels));
  assert(halfpts % (1<<num_levels) == 0);

  /* we need 2^(num_levels)-1 full intervals and
     2^(num_levels-1)-1 -(num_levels-1) asymmetric intervals,
     times two parities.
     (The asymmetric does not use the top and bottom levels,
     or intervals that hit the left edge.)*/
  num_int_dyadic=(1<<num_levels) -1 + (1<<(num_levels-1)) -1*num_levels;
  ilist[0]=(Local_Trig_Interval *)calloc(2*num_int_dyadic,
					 sizeof(Local_Trig_Interval));
  assert(ilist[0]!=NULL);

  /* give half to the odd parity */
  ilist[1]=ilist[0]+num_int_dyadic;

  /* note the number of intervals as an output */
  (*ints_used)=num_int_dyadic;

  /* we use the following indexing convention for the coefficient matrix:
     A given level has 2^level intervals, of length halfpts/2^level.
     coefficients for [level][within] start at 
     level*halfpts+within*halfpts/2^level
     For asymmetric coefficients, there is and additional offset of
     (num_levels-1)*halfpts. (Note there is no asymmetric level 0, so
     we can collapse one level)
     Although the asymmetric doesn't use intervals at the left edge,
     we leave space for them so the indexing doesn't get too complicated. 
     */

  /* calculate the total number of possible coefficients,
     to use later as an offset.
     The symmetric bell portion has halfpts*num_levels
     The asymmetric portion has halfpts*(num_levels-2) if num_levels>2*/
  (*coefs_used)=halfpts*num_levels;
  if(num_levels >2)
    (*coefs_used)+=halfpts*(num_levels-2);

#if 1
  /* initialize bell_holder. It holds all the bell values */

  /* first index, the level.
   We have twice as many so we can hold dual bells too*/
  bell_holder=(REAL ***)calloc(num_levels*2,sizeof(REAL **));
  assert(bell_holder != NULL);
  
  /* second index, which bell at that level. convention:
     0 -- symmetric left edge
     1 -- symmetric generic
     2 -- symmetric right edge
     3 -- asymmetric generic
     4 -- asymmetric right edge
     Not all levels will have all of these, but for simplicity we get
     memory for all of them.
     We also get twice as much for the dual bells*/
  bell_holder[0]=(REAL **)calloc(5*num_levels*2,sizeof(REAL *));
  assert(bell_holder[0] != NULL);
  
  /* memory for the offsets. It holds the same information as bell_holder[0],
     but in the form of offsets rather than pointers */
  offset=(int *)calloc(5*num_levels*2,sizeof(int));
  assert(offset != NULL);

  for(level=1;level<num_levels*2;level++){
    /* loop through the  levels, including extra for dual bells*/

    /* align the next one */
    bell_holder[level]=bell_holder[level-1]+5;
  }

  /* main bell memory. We have
    level 0 -- halfpts (constant function)
    level 1 -- 2 edge symmetric at (halfpts/2+halfpts/4) each = halfpts*(3/2)
               if not the last level, then 1 edge asym at halfpts*(5/8)
    level _ -- 2 edge symmetrc at winlength*3/2 = winlength*3 
               1 generic symmetric at winlength*2 = winlength*2
	       1 generic asymmetric at winlength*7/4
	       1 edge asymmetric at winlength*5/4
	       total= winlength*8
    level last -- 2 edge symmetrc at winlength*3/2 = winlength*3 
                  1 generic symmetric at winlength*2 = winlength*2
		  total winlength*5 
    We then DOUBLE this, so we can have dual bells as well
    **/
  
  /* level zero */
  bell_total=halfpts;

  if( num_levels >1){

    /* level 1 symmetrics */
    bell_total+=(halfpts*3)/2;

    if( num_levels >2){

      /* extra level 1 asymmetric */
      bell_total+=((halfpts/2)*5)/4;

      for(level=2; level < num_levels-1; level++){
	/* generics */

	/* add on bell lengths,  winlength*8*/
	bell_total+=(halfpts>>level)*8;
      }

      /* last level */
      bell_total+=(halfpts>>level)*5;
    }
  }

  /* get the memory, doubling so we can have dual bells */
  bell_holder[0][0]=(REAL *)calloc(bell_total*2,sizeof(REAL));
  assert(bell_holder[0][0] != NULL);

  /* create bells and align pointers 
   The dual bells parallel the regular bells, but are offset by bell_total*/

  /* next free space */
  bell_mem= bell_holder[0][0];

  /* level 0*/
  bell_1size_init(bell_holder[0][0],bell_choice,0,halfpts,0,0);
  /*dual bell*/
  bell_holder[num_levels][0]=bell_mem+bell_total;
  bell_1size_init(bell_holder[num_levels][0],bell_choice,0,halfpts,0,1);
  /* update bell_mem*/
  bell_mem+=halfpts;

  /* level 0 all offset are 0*/
  offset[0]=0;
  offset[num_levels*5+0]=0;
  offset[1]=0;
  offset[num_levels*5+1]=0;
  offset[2]=0;
  offset[num_levels*5+2]=0;
  offset[3]=0;
  offset[num_levels*5+3]=0;
  offset[4]=0;
  offset[num_levels*5+4]=0;

  if(num_levels >1 ){
    /********* level 1 *****************/
 
    level=1;
   
    /* set winlength */
    winlength=halfpts>>level;

    /* left edge symmetric */
    /* align memory */
    bell_holder[level][0]=bell_mem;
    bell_holder[num_levels+level][0]=bell_mem+bell_total;
    /* set offsets */
    offset[level*5]=bell_holder[level][0]-bell_holder[0][0];
    offset[(num_levels+level)*5]=offset[level*5]+bell_total;
    /* make bells */
    bell_1size_init(bell_holder[level][0],bell_choice,
		    0,winlength,winlength/2,0);
    bell_1size_init(bell_holder[num_levels+level][0],bell_choice,
		    0,winlength,winlength/2,1);
    /* update bell_mem*/
    bell_mem+=(winlength*3)/2;

    /* no generic symmetric, but exists as an offset */
    offset[level*5+1]=0;
    offset[(num_levels+level)*5+1]=0;

    /* right edge symmetric */
    /* align memory */
    bell_holder[level][2]=bell_mem;
    bell_holder[num_levels+level][2]=bell_mem+bell_total;
    /* set offsets */
    offset[level*5+2]=bell_holder[level][2]-bell_holder[0][0];
    offset[(num_levels+level)*5+2]=offset[level*5+2]+bell_total;
    /* make bells */
    bell_1size_init(bell_holder[level][2],bell_choice,
		    winlength/2,winlength,0,0);
    bell_1size_init(bell_holder[num_levels+level][2],bell_choice,
		    winlength/2,winlength,0,1);
    /* update bell_mem*/
    bell_mem+=(winlength*3)/2;

    /* no generic asymmetric */
    offset[level*5+3]=0;
    offset[(num_levels+level)*5+3]=0;

    /* may not be right edge asymmetic. If there is, it will overwrite this */
    offset[level*5+4]=0;
    offset[(num_levels+level)*5+4]=0;

    if( num_levels >2 ){

      /* extra level 1 asymmetric right edge. retain level==1 from above */
      /* align memory */
      bell_holder[level][4]=bell_mem;
      bell_holder[num_levels+level][4]=bell_mem+bell_total;
      /* set offsets */
      offset[level*5+4]=bell_holder[level][4]-bell_holder[0][0];
      offset[(num_levels+level)*5+4]=offset[level*5+4]+bell_total;
      /* make bells */
      bell_1size_init(bell_holder[level][4],bell_choice,
		      winlength/4,winlength,0,0);
      bell_1size_init(bell_holder[num_levels+level][4],bell_choice,
		      winlength/4,winlength,0,1);
      /* update bell_mem*/
      bell_mem+=(winlength*5)/4;

      for(level=2; level < num_levels-1; level++){
	/* generics */

	/* set winlength*/
	winlength=(halfpts>>level);

	/* left edge symmetric */
	/* align memory */
	bell_holder[level][0]=bell_mem;
	bell_holder[num_levels+level][0]=bell_mem+bell_total;
	/* set offsets */
	offset[level*5]=bell_holder[level][0]-bell_holder[0][0];
	offset[(num_levels+level)*5]=offset[level*5]+bell_total;
	/* make bells */
	bell_1size_init(bell_holder[level][0],bell_choice,
			0,winlength,winlength/2,0);
	bell_1size_init(bell_holder[num_levels+level][0],bell_choice,
			0,winlength,winlength/2,1);
	/* update bell_mem*/
	bell_mem+=(winlength*3)/2;

	/* generic symmetric */
	/* align memory */
	bell_holder[level][1]=bell_mem;
	bell_holder[num_levels+level][1]=bell_mem+bell_total;
	/* set offsets */
	offset[level*5+1]=bell_holder[level][1]-bell_holder[0][0];
	offset[(num_levels+level)*5+1]=offset[level*5+1]+bell_total;
	/* make bells */
	bell_1size_init(bell_holder[level][1],bell_choice,
			winlength/2,winlength,winlength/2,0);
	bell_1size_init(bell_holder[num_levels+level][1],bell_choice,
			winlength/2,winlength,winlength/2,1);
	/* update bell_mem*/
	bell_mem+=winlength*2;

	/* right edge symmetric */
	/* align memory */
	bell_holder[level][2]=bell_mem;
	bell_holder[num_levels+level][2]=bell_mem+bell_total;
	/* set offsets */
	offset[level*5+2]=bell_holder[level][2]-bell_holder[0][0];
	offset[(num_levels+level)*5+2]=offset[level*5+2]+bell_total;
	/* make bells */
	bell_1size_init(bell_holder[level][2],bell_choice,
			winlength/2,winlength,0,0);
	bell_1size_init(bell_holder[num_levels+level][2],bell_choice,
			winlength/2,winlength,0,1);
	/* update bell_mem*/
	bell_mem+=(winlength*3)/2;

	/* generic asymmetric */
	/* align memory */
	bell_holder[level][3]=bell_mem;
	bell_holder[num_levels+level][3]=bell_mem+bell_total;
	/* set offsets */
	offset[level*5+3]=bell_holder[level][3]-bell_holder[0][0];
	offset[(num_levels+level)*5+3]=offset[level*5+3]+bell_total;
	/* make bells */
	bell_1size_init(bell_holder[level][3],bell_choice,
			winlength/4,winlength,winlength/2,0);
	bell_1size_init(bell_holder[num_levels+level][3],bell_choice,
			winlength/4,winlength,winlength/2,1);
	/* update bell_mem*/
	bell_mem+=(winlength*7)/4;
	
	/* right edge asymmetric */
	/* align memory */
	bell_holder[level][4]=bell_mem;
	bell_holder[num_levels+level][4]=bell_mem+bell_total;
	/* set offsets */
	offset[level*5+4]=bell_holder[level][4]-bell_holder[0][0];
	offset[(num_levels+level)*5+4]=offset[level*5+4]+bell_total;
	/* make bells */
	bell_1size_init(bell_holder[level][4],bell_choice,
			winlength/4,winlength,0,0);
	bell_1size_init(bell_holder[num_levels+level][4],bell_choice,
			winlength/4,winlength,0,1);
	/* update bell_mem*/
	bell_mem+=(winlength*5)/4;

      }

      /* last level, use level value from loop */

      /* set winlength */
      winlength=halfpts>>level;

      /* left edge symmetric */
      /* align memory */
      bell_holder[level][0]=bell_mem;
      bell_holder[num_levels+level][0]=bell_mem+bell_total;
      /* set offsets */
      offset[level*5]=bell_holder[level][0]-bell_holder[0][0];
      offset[(num_levels+level)*5]=offset[level*5]+bell_total;
      /* make bells */
      bell_1size_init(bell_holder[level][0],bell_choice,
			0,winlength,winlength/2,0);
      bell_1size_init(bell_holder[num_levels+level][0],bell_choice,
		      0,winlength,winlength/2,1);
      /* update bell_mem*/
      bell_mem+=(winlength*3)/2;
      
      /* generic symmetric */
      /* align memory */
      bell_holder[level][1]=bell_mem;
      bell_holder[num_levels+level][1]=bell_mem+bell_total;
      /* set offsets */
      offset[level*5+1]=bell_holder[level][1]-bell_holder[0][0];
      offset[(num_levels+level)*5+1]=offset[level*5+1]+bell_total;
      /* make bells */
      bell_1size_init(bell_holder[level][1],bell_choice,
		      winlength/2,winlength,winlength/2,0);
      bell_1size_init(bell_holder[num_levels+level][1],bell_choice,
		      winlength/2,winlength,winlength/2,1);
      /* update bell_mem*/
      bell_mem+=winlength*2;
      
      /* right edge symmetric */
      /* align memory */
      bell_holder[level][2]=bell_mem;
      bell_holder[num_levels+level][2]=bell_mem+bell_total;
      /* set offsets */
      offset[level*5+2]=bell_holder[level][2]-bell_holder[0][0];
      offset[(num_levels+level)*5+2]=offset[level*5+2]+bell_total;
	/* make bells */
      bell_1size_init(bell_holder[level][2],bell_choice,
		      winlength/2,winlength,0,0);
      bell_1size_init(bell_holder[num_levels+level][2],bell_choice,
		      winlength/2,winlength,0,1);
      /* update bell_mem*/
      bell_mem+=(winlength*3)/2;
    
      /* no generic or right edge asymmetric */
      offset[level*5+3]=0;
      offset[(num_levels+level)*5+3]=0;
      offset[level*5+4]=0;
      offset[(num_levels+level)*5+4]=0;


    }
  }

  /* make all the lct and dct saves */

  /* for each level there is a lct and a dct */
  plan_holder=(Ftrigtw_Plan *)calloc(2*num_levels,sizeof(Ftrigtw_Plan));
  assert(plan_holder != NULL);

  for(level=0;level<num_levels; level++){
    /* loop through the levels */

    /* set winlength */
    winlength=halfpts>>level;

    /* make lctw plan */
    lctw_init(plan_holder+level,flags,winlength);

    /* make dctw plan. Offset by num_levels */
    dctw_init(plan_holder+level+num_levels,flags,winlength);
  }

  /* start interval count at 0 */
  interval=0;

  /* first we do all the symmetric, full bell intervals ****/

  /* level 0 initialize */
  /*even*/
  /* initialize with given values*/
  ilist[0][interval].numpts=halfpts;
  ilist[0][interval].left_extra=0;
  ilist[0][interval].right_extra=0;
  ilist[0][interval].fcn_offset=0;
  ilist[0][interval].coef_offset=0;
  /* align bells */
  ilist[0][interval].bell=bell_holder[0][0];
  ilist[0][interval].dualbell=bell_holder[num_levels+0][0];
  /* set which_transform */
  ilist[0][interval].which_transform=1;/*dct*/
  /* copy plan (pointers) */
  ilist[0][interval].transform_plan=plan_holder[num_levels+0];

  /*odd. which transform and transform plan are the only differences*/
  ilist[1][interval]=ilist[0][interval];
  ilist[1][interval].which_transform=0;/* lct*/
  ilist[1][interval].transform_plan=plan_holder[0];

  /* increment interval */
  interval++;

  for(level=1; level<num_levels;level++){
    /* loop through the other levels */

    /* set interval size */
    winlength=halfpts>>level;

    /* note number of intervals on this level */
    n_int_lev=1<<level;

    /* left edge */
    /*even and odd the same (use lct)*/
    /* initialize with given values*/
    ilist[0][interval].numpts=winlength;
    ilist[0][interval].left_extra=0;
    ilist[0][interval].right_extra=winlength/2;
    ilist[0][interval].fcn_offset=0;
    ilist[0][interval].coef_offset=level*halfpts;
    /* align bells */
    ilist[0][interval].bell=bell_holder[level][0];
    ilist[0][interval].dualbell=bell_holder[num_levels+level][0];
    /** set which_transform */
    ilist[0][interval].which_transform=0;/*lct*/
    /* copy plan (pointers) */
    ilist[0][interval].transform_plan=plan_holder[level];

    /* odd is the same */
    ilist[1][interval]=ilist[0][interval];

    /* increment interval */
    interval++;

    /* generic */
    for(within=1;within<n_int_lev-1;within++){

      /*even and odd the same (use lct)*/
      /* initialize with given values*/
      ilist[0][interval].numpts=winlength;
      ilist[0][interval].left_extra=winlength/2;
      ilist[0][interval].right_extra=winlength/2;
      ilist[0][interval].fcn_offset=within*winlength;
      ilist[0][interval].coef_offset=level*halfpts+within*winlength;
      /* align bells */
      ilist[0][interval].bell=bell_holder[level][1];
      ilist[0][interval].dualbell=bell_holder[num_levels+level][1];
      /* set which_transform */
      ilist[0][interval].which_transform=0;/*lct*/
      /* copy plan (pointers) */
      ilist[0][interval].transform_plan=plan_holder[level];

      /* odd is the same */
      ilist[1][interval]=ilist[0][interval];

      /* increment interval */
      interval++;

    }

    /* right edge. use inherited within from loop.
       here even and odd differ */
    /* even */
    /* initialize with given values*/
    ilist[0][interval].numpts=winlength;
    ilist[0][interval].left_extra=winlength/2;
    ilist[0][interval].right_extra=0;
    ilist[0][interval].fcn_offset=within*winlength;
    ilist[0][interval].coef_offset=level*halfpts+within*winlength;
    /* align bells */
    ilist[0][interval].bell=bell_holder[level][2];
    ilist[0][interval].dualbell=bell_holder[num_levels+level][2];
    /* set which_transform */
    ilist[0][interval].which_transform=1;/*dct*/
    /* copy plan (pointers) */
    ilist[0][interval].transform_plan=plan_holder[num_levels+level];

    /* odd differs in which_transform and in plan */
    ilist[1][interval]=ilist[0][interval];
    ilist[1][interval].which_transform=0;
    ilist[1][interval].transform_plan=plan_holder[level];

    /* increment interval */
    interval++;

  }/* level loop */

  /* now we do the asymmetric bells, which spill half as much to the left.
    otherwise they behave the same.
    There is no level==0 or level==num_levels-1
    We don't compute the left edge intervals, but for now we include them
    in the coef_offsets.
    */

  for(level=1; level<num_levels-1;level++){
    /* loop through the levels 1 through num_levels-2*/

    /* set interval size */
    winlength=halfpts>>level;

    /* note number of intervals on this level */
    n_int_lev=1<<level;

    /* left edge SKIPPED, but for now included in the coef_offsets*/

    /* generic */
    for(within=1;within<n_int_lev-1;within++){

      /* same as symmetric case except left_extra,
	 bell_holder[][3]
	 (level-1) in coef_offset,
	 and extra coef offset of num_levels*halfpts*/

      /*even and odd the same (use lct)*/
      /* initialize with given values*/
      ilist[0][interval].numpts=winlength;
      ilist[0][interval].left_extra=winlength/4;
      ilist[0][interval].right_extra=winlength/2;
      ilist[0][interval].fcn_offset=within*winlength;
      ilist[0][interval].coef_offset=num_levels*halfpts+
	(level-1)*halfpts+within*winlength;
      /* align bells */
      ilist[0][interval].bell=bell_holder[level][3];
      ilist[0][interval].dualbell=bell_holder[num_levels+level][3];
      /* set which_transform */
      ilist[0][interval].which_transform=0;/*lct*/
      /* copy plan (pointers) */
      ilist[0][interval].transform_plan=plan_holder[level];

      /* odd is the same */
      ilist[1][interval]=ilist[0][interval];

      /* increment interval */
      interval++;

    }
    /* right edge. use inherited within from loop.
       here even and odd differ */
    /* same as symmetric case except left_extra ,
       bell_holder[][4]
       (level-1) in coef_offset
       and extra coef offset of num_levels*halfpts*/
    /* even */
    /* initialize with given values*/
    ilist[0][interval].numpts=winlength;
    ilist[0][interval].left_extra=winlength/4;
    ilist[0][interval].right_extra=0;
    ilist[0][interval].fcn_offset=within*winlength;
    ilist[0][interval].coef_offset=num_levels*halfpts+
	(level-1)*halfpts+within*winlength;
    /* align bells */
    ilist[0][interval].bell=bell_holder[level][4];
    ilist[0][interval].dualbell=bell_holder[num_levels+level][4];
    /* set which_transform */
    ilist[0][interval].which_transform=1;/*dct*/
    /* copy plan (pointers) */
    ilist[0][interval].transform_plan=plan_holder[num_levels+level];

    /* odd differs in which_transform and in plan */
    ilist[1][interval]=ilist[0][interval];
    ilist[1][interval].which_transform=0;
    ilist[1][interval].transform_plan=plan_holder[level];

    /* increment interval */
    interval++;

  }/* level loop */

  /* align bell_handle */
  (*bell_handle)=bell_holder[0][0];

  /* free top levels of bell_holder */
  free(bell_holder[0]);
  free(bell_holder);

  /* align offsets */
  (*bell_offsets)=offset;

  /* set bell_length; account for doubling due to dual bells */
  (*bell_length)=bell_total*2;

  /* free plan_holder. The things it pointed to are held by the intervals.
     note that this may cause a memory leak since we have no good way 
     of freeing the plans */
#if 0
  free(plan_holder);
#endif

  /*************************************************************************/
  /*************************************************************************/
#else
  /*************************************************************************/

  /* start interval count at 0 */
  interval=0;

  /*** first we do all the symmetric, full bell intervals ****/

  /* level 0 initialize */
  /*even*/
  ltrigint_init(ilist[0]+interval,halfpts,0,0,0,0,bell_choice,1,flags);
  /*odd*/
  ltrigint_init(ilist[1]+interval,halfpts,0,0,0,0,bell_choice,0,flags);

  /* increment interval */
  interval++;

  for(level=1; level<num_levels;level++){
    /* loop through the other levels */

    /* set interval size */
    winlength=halfpts>>level;

    /* note number of intervals on this level */
    n_int_lev=1<<level;

    /* left edge */
    /*even and odd the same (use lct)*/
    ltrigint_init(ilist[0]+interval,winlength,0,winlength/2,
		  0,level*halfpts,bell_choice,0,flags);
    ilist[1][interval]=ilist[0][interval];

    /* increment interval */
    interval++;

    /* generic */
    for(within=1;within<n_int_lev-1;within++){

      /*even and odd the same (use lct)*/
      ltrigint_init(ilist[0]+interval,winlength,winlength/2,winlength/2,
		    within*winlength,level*halfpts+within*winlength,
		    bell_choice,0,flags);
      ilist[1][interval]=ilist[0][interval];

      /* increment interval */
      interval++;

    }

    /* right edge. use inherited within from loop.
       here even and odd differ */
    ltrigint_init(ilist[0]+interval,winlength,winlength/2,0,
		  within*winlength,level*halfpts+within*winlength,
		  bell_choice,1,flags);
    ltrigint_init(ilist[1]+interval,winlength,winlength/2,0,
		  within*winlength,level*halfpts+within*winlength,
		  bell_choice,0,flags);

    /* increment interval */
    interval++;

  }/* level loop */

  /** now we do the asymmetric bells, which spill half as much to the left.
    otherwise they behave the same.
    There is no level==0 or level==num_levels-1
    We don't compute the left edge intervals, but for now we include them
    in the coef_offsets.
    */

  for(level=1; level<num_levels-1;level++){
    /* loop through the levels 1 through num_levels-2*/

    /* set interval size */
    winlength=halfpts>>level;

    /* note number of intervals on this level */
    n_int_lev=1<<level;

    /* left edge SKIPPED, but for now included in the coef_offsets*/

    /* generic */
    for(within=1;within<n_int_lev-1;within++){

      /* same as symmetric case except left_extra,
	 (level-1),
	 and extra coef offset of num_levels*halfpts*/

      /*even and odd the same (use lct)*/
      ltrigint_init(ilist[0]+interval,winlength,winlength/4,winlength/2,
		    within*winlength,
		    num_levels*halfpts+(level-1)*halfpts+within*winlength,
		    bell_choice,0,flags);
      ilist[1][interval]=ilist[0][interval];

      /* increment interval */
      interval++;
    }

    /* right edge. use inherited within from loop.
       here even and odd differ */
    /* same as symmetric case except left_extra ,
       (level-1)
	 and extra coef offset of num_levels*halfpts*/
    ltrigint_init(ilist[0]+interval,winlength,winlength/4,0,
		  within*winlength,
		  num_levels*halfpts+(level-1)*halfpts+within*winlength,
		  bell_choice,1,flags);
    ltrigint_init(ilist[1]+interval,winlength,winlength/4,0,
		  within*winlength,
		  num_levels*halfpts+(level-1)*halfpts+within*winlength,
		  bell_choice,0,flags);

    /* increment interval */
    interval++;

  }/* level loop */

#endif
  /********************************************************************/

#if 0
  /* diagnostic: check that we used the number of intervals we predicted */
  assert(num_int_dyadic==interval);
#endif

#if 0
  /* diagnostic:
     disable the bells and transforms.
     If num_levels==1, this is the direct method */
  for(level=0;level<num_int_dyadic;level++){
    /** loop through the intervals*/

    /* set the tranform to 9, which is simple copy */
    ilist[0][level].which_transform=9;
    ilist[1][level].which_transform=9;
  }
  printf("dyadic_g_ilist_init: WARNING: disabling trig transforms.\n");
#endif



#if ENTEREXIT
  printf("exiting dyadic_g_ilist_init\n");
#endif
}
/****************************************************************/

void ltrigint_init(Local_Trig_Interval *out,
		   int numpts,
		   int left_extra,
		   int right_extra,
		   int fcn_offset,
		   int coef_offset,
		   REAL (*bell_choice)(int,int),
		   int which_transform,
		   int fftw_flags
		   ){
  /** This routine initializes a Local_Trig_Interval.

  INPUTS: 
  - numpts
  - left_extra
  - right_extra
  - fcn_offset
  - coef_offset
  - which_transform -- are copied into the structure
  - bell_choice -- a pointer to the bell function to use
  - fftw_flags -- flags passed to the fftw initializer

  OUTPUTS: out -- is written onto with the above info.
  - Memory is obtained for .bell and .dualbell and
    they are initialized.
  - .transform_plan is initialized for the transform
    indicated by which_transform.

  NOTES:
  - document better
  */

#if ENTEREXIT
  printf("entering ltrigint_init\n");
#endif

  /* initialize with given values*/
  out->numpts=numpts;
  out->left_extra=left_extra;
  out->right_extra=right_extra;
  out->fcn_offset=fcn_offset;
  out->coef_offset=coef_offset;

  /*********** initialize bell ************/

  /* memory */
  out->bell=(REAL *)calloc(2*(left_extra+numpts+right_extra),sizeof(REAL));
  assert(out->bell != NULL);
  /* give half to the dual */
  out->dualbell=out->bell+left_extra+numpts+right_extra;

  /* bell*/
  bell_1size_init(out->bell,bell_choice,left_extra,numpts,right_extra,0);

  /*dual bell*/
  bell_1size_init(out->dualbell,bell_choice,left_extra,numpts,right_extra,1);

  /********** set which_transform **************/
  out->which_transform=which_transform;

  if(which_transform==0)
    /* lct */
    lctw_init(&(out->transform_plan),fftw_flags,numpts);

  else if(which_transform==1)
    /* dct */
    dctw_init(&(out->transform_plan),fftw_flags,numpts);

  else
    /* error */
    fprintf(stderr,"ltrigint_init: Error: illegal which_transform=%i\n",
	    which_transform);
#if ENTEREXIT
  printf("exiting ltrigint_init\n");
#endif
}

void gsearch_to_ilist(Local_Trig_Interval *ilist_out,
		      int *intervals_kept,
		      int max_intervals,
		      Local_Trig_Interval *ilist_in,
		      Dyadic_Gsearch_Save *dy_gs_save,
		      int halfpts
		      ){
  /** This function converts the choices made by the dyadic_gsearch
      to a list of intervals.

      INPUTS: 
      - max_intervals -- the largest number of interval ilist_out can hold 
      - ilist_in -- all possible intervals, in Local_Trig_Interval format
      - dy_gs_save -- a pointer to gsearch structure
        From its best_level and branching fields we extract the partition.
      - halfpts -- the numpts of data points this search covered
        In the usual case where the grow search is for half
	the true function interval, this is half the number of points.

      OUTPUT:  
      - ilist_out -- written onto with the list of interval.
        Note that we copy the structure Local_Trig_Interval,
	which includes pointers to other things. We do not
	copy the arrays they point to
      - intervals_kept -- the number of intervals kept is written here

      NOTES:
      - It is possible to run over ilist_in
      - In 2007 fixed bug i=i++ --> i++

  */

  register int i,j;
  int level,within,currentbranch,winlength,left_extra,fcn_offset;

#if ENTEREXIT
  printf("entering gsearch_to_ilist\n");
#endif

  /*** extract the partition choosen ***/

  /* start at best level*/
  level=dy_gs_save->best_level;

  /* start off left edge */
  within=-1;
  
  /* aim to move to right at same level */
  currentbranch=1;

  /* note current window length */
  winlength=halfpts>>level;

  /* at (really off) the left edge */
  left_extra=0;

  /* no intervals so far */
  i=0;

  /* branching -- the list of decisions made by growsearch.
                  The first index is level, and the second is interval
		  within that level.  The branching values are:
		  0 - stop, we are at the right edge
		  1 - stay at the same scale and move right
		  2 - go up 1 scale and move right */

  while(currentbranch){
    /* not at end, look for next interval */
    
    if(currentbranch==2){      /* go up level */

      /* up level */
      level--;

      /* move one to the right, then change scales */
      within=(within+1)>>1;

      /* reset branching*/
      currentbranch=dy_gs_save->branching[level][within];
      
      /* note window length at this scale, twice as much as before */
      winlength= winlength<<1;

      /* set left_extra for asymmetric bell*/
      left_extra=(winlength>>2);

    }else{
      /* stay at same level */

      /* move to the right */
      within++;

      /* reset branching*/
      currentbranch=dy_gs_save->branching[level][within];
      
      /* since we are at the same scale, winlength is unchanged*/

      /* note the left spilling. The [0] interval has zero spill,
	 inherited from the initialization.
	 Other intervals must be set.*/
      if(within)
	left_extra=(winlength>>1);
    }

    /* note offset within main fcn interval */
    fcn_offset=within * winlength;

    /* We now have determined an interval, but still must find it in 
       ilist_in and copy it to ilist_out.
       We simply run through ilist_in until we find it. This is not efficient,
       but to do better we would need to assume something about the order
       of the intervals, which seems dangerous.
       Note that is there is no match at all, this will overrun 
       */
    /*
    printf("winlength=%i fcn_offset=%i left_extra=%i\n",
	   winlength,fcn_offset,left_extra);
	   */
    j=0;
    while((ilist_in[j].numpts != winlength) /* compare level */
	  ||(ilist_in[j].fcn_offset != fcn_offset) /* compare within */
	  ||(ilist_in[j].left_extra != left_extra) /* compare full/asym */
	  ){
      /*      printf("[%i]\n",j);*/
      /* no match try next j*/
      j++;
    }
    /* make sure there is room to keep another */
    assert(i<max_intervals);

    /* we now have a match, which we copy */
    ilist_out[i]=ilist_in[j];

    /* kept an interval */
    i++;
    
  }

  (*intervals_kept)=i;

#if ENTEREXIT
  printf("exiting gsearch_to_ilist\n");
#endif
}

