/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file fast1_fromfile.c

\brief Load precomputed data for fast1 transform.

Summary:
- load_fast1 loads from file. 
- fast1_fromfile interprets the output of load_fast1 and sets up.

*/

#include "libftsh.h"
 
#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void fast1_fromfile(int *numpts_ptr,
		    int *band_limit_ptr,
		    REAL **node_ptr,
		    REAL **weight_ptr,
		    Pm_1D **pmn_matrix_ptr,
		    Pm_1D_Workspace *pmnwork,
		    FILE *fp
		    ){
  /** This routine loads the initialization dumped by dump_fast1,
      interprets it and does all the set-up necessary to do the
      synthesis or analysis transform

      INPUTS: fp -- the objects are read from the file pointed to by
      fp, in binary format. The file must be opened before
      calling this routine.

      OUTPUTS: 
      - numpts --  the number of points on the full interval.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
      - node --  an array of length numpts with the nodes,
        usually Arcsin of the usual gaussian nodes.
      - weight -- an array of length numpts with the quadrature weights
      - pmn_matrix -- an array of band_limit (==order_limit)
        Pm_1D structures, each representing the compressed 
	matrix for one m.
      - pmn_work -- a structure holding workspace and other precomputed
        values used by the transform, independent of m

      NOTES:
      - node and weight are not needed for the synthesis/analysis routines
        but are provided here for other uses.
  */
  int num_levels;
  Fast1_Save saves_fordump;

#if ENTEREXIT
  printf("entering fast1_fromfile\n");
#endif

  /* load from file */
  load_fast1(numpts_ptr,band_limit_ptr,&num_levels,node_ptr,weight_ptr,
	     &saves_fordump,pmn_matrix_ptr,fp);

  /* free bell_offsets from saves_fordump */
  free(saves_fordump.bell_offsets);

  /* the memory for saves_fordump.bell_handle is not freed, as it is still
     in use. It can be freed at the end if we can find a level==0 bell */

  /* num_levels is forgotten */

  /* initialize workspace for synthesizer/analizer */
  init_Pm_1d_workspace(pmnwork,*numpts_ptr,*band_limit_ptr,
		       *node_ptr,*weight_ptr);

#if ENTEREXIT
  printf("exiting fast1_fromfile\n");
#endif
}

void load_fast1(int *numpts_ptr,
		int *band_limit_ptr,
		int *num_levels_ptr,
		REAL **node_ptr,
		REAL **weight_ptr,
		Fast1_Save *saves_fordump,
		Pm_1D **pmn_matrix_ptr,
		FILE *fp
		){
  /** This routine loads the initialization dumped by dump_fast1.

  We interpret what we read and re-align pointers and such to get
  things ready to use.

  INPUTS: fp -- the objects are read from the file pointed to by fp,
  in binary format. The file must be opened before calling this
  routine

  OUTPUTS: 
  - numpts --  the number of points on the full interval.
  - band_limit -- the bound on the degree of the expansion.
    In P^m_n notation this bounds n.
  - num_levels -- the number of dyadic levels used in the adaption   
  - node --  an array of length numpts with the nodes,
    usually Arcsin of the usual gaussian nodes.
  - weights -- an array of length numpts with the quadrature weights
  - saves_fordump -- a structure containing the information needed
    for the bells. Constructed by init_fast1
  - pmn_matrix -- an array of band_limit (==order_limit)
    Pm_1D structures, each representing the compressed matrix for one m.


  NOTES:
  - num_levels is not needed by the synthesis/analysis routines. It may
    be useful to keep it so we could re-dump.
  - saves_for dump is also not needed. we could free it here.
  - we free plan_holder, thus losing our handle on the trig plans
    This may cause memory leaks since we can't free them easily
  - We now have separate workspace for each trig plan, but these could
    be combined. 
  */

  register int i,m,order_limit;
  int num_rows,num_elements,num_bands,num_intervals;
  int numpts,band_limit,num_levels;
  REAL *node,*weight;
  Pm_1D *pmn_matrix;
  Banded_Matrix bdmat_mem, *bdmat_used;
  Local_Trig_Interval *ilist_mem,*ilist_used;
  int level,winlength,flags,halfpts;
  REAL ***bell_holder;
  Ftrigtw_Plan *plan_holder;
  register int parity;

#if ENTEREXIT
  printf("entering load_fast1\n");
#endif
  /* read numpts and band_limit */
  fread(&numpts,sizeof(int),1,fp);
  fread(&band_limit,sizeof(int),1,fp);

  /* order_limit is the same as band_limit */
  order_limit=band_limit;

  /* we could dump epsilon for diagnostics */

  /*  read num_levels so we can recreate trig stuff */
  fread(&num_levels,sizeof(int),1,fp);

  /* read nodes and weights */
  node=(REAL *)calloc(numpts,sizeof(REAL));
  assert(node != NULL);
  fread(node,sizeof(REAL),numpts,fp);
  weight=(REAL *)calloc(numpts,sizeof(REAL));
  assert(weight != NULL);
  fread(weight,sizeof(REAL),numpts,fp);

  /* read saves_fordump, which contains the bells */
  /* get a new name */
  fread(&(saves_fordump->bell_length),sizeof(int),1,fp);
  saves_fordump->bell_handle=
    (REAL *)calloc(saves_fordump->bell_length,sizeof(REAL));
  assert(saves_fordump->bell_handle != NULL);
  fread(saves_fordump->bell_handle,sizeof(REAL),saves_fordump->bell_length,fp);
  saves_fordump->bell_offsets=(int *)calloc(10*num_levels,sizeof(int));
  assert(saves_fordump->bell_offsets != NULL);
  fread(saves_fordump->bell_offsets,sizeof(int),10*num_levels,fp);

  /* read num_rows,num_elements,num_bands,num_intervals */
  fread(&num_rows,sizeof(int),1,fp);
  fread(&num_elements,sizeof(int),1,fp);
  fread(&num_bands,sizeof(int),1,fp);
  fread(&num_intervals,sizeof(int),1,fp);
  
  /* get all banded matrix  memory */
  bdmat_mem=init_bdmat_memory(num_elements,num_rows,num_bands);
  bdmat_used=&bdmat_mem;

  /* get all intervallist memory */
  ilist_mem=(Local_Trig_Interval *)
    calloc(num_intervals,sizeof(Local_Trig_Interval));
  assert(ilist_mem != NULL);
  ilist_used=ilist_mem;

  /* get the pmn_matrix memory */
  pmn_matrix=(Pm_1D *)calloc(order_limit,sizeof(Pm_1D));
  assert(pmn_matrix != NULL);

  /* main martix input.
     we read everything needed for the banded matrices,
     but only the parameters for the Local_Trig_Intervals */
  for(m=0;m<order_limit;m++){
    /* loop through the orders */

    /* give memory to matrix[0] */
    give_bdmat_memory(&(pmn_matrix[m].matrix[0]),bdmat_used);
    /* load [0] */
    load_bin_bdmat(&(pmn_matrix[m].matrix[0]),fp);
    /* re-aim memory */
    bdmat_used=&(pmn_matrix[m].matrix[0]);

    /* give memory to matrix[1] */
    give_bdmat_memory(&(pmn_matrix[m].matrix[1]),bdmat_used);
    /* load [1] */
    load_bin_bdmat(&(pmn_matrix[m].matrix[1]),fp);
    /* re-aim memory */
    bdmat_used=&(pmn_matrix[m].matrix[1]);

    /* read n_interval[0/1]*/
    fread(&(pmn_matrix[m].n_interval[0]),sizeof(int),1,fp);
    fread(&(pmn_matrix[m].n_interval[1]),sizeof(int),1,fp);

    /* align ilist[0]*/
    pmn_matrix[m].ilist[0]=ilist_used;
    /* read [0]; remember pointers are gibberish */
    for(i=0;i<pmn_matrix[m].n_interval[0];i++)
      fread(pmn_matrix[m].ilist[0]+i,sizeof(Local_Trig_Interval),1,fp);
    /* update ilist_used */
    ilist_used=pmn_matrix[m].ilist[0]+pmn_matrix[m].n_interval[0];

    /* align ilist[1]*/
    pmn_matrix[m].ilist[1]=ilist_used;
    /* read [1]; remember pointers are gibberish */
    for(i=0;i<pmn_matrix[m].n_interval[1];i++)
      fread(pmn_matrix[m].ilist[1]+i,sizeof(Local_Trig_Interval),1,fp);
    /* update ilist_used */
    ilist_used=pmn_matrix[m].ilist[1]+pmn_matrix[m].n_interval[1];

  }

  /* read fftw wisdom. We do this last because we have less control over it*/
  fftw_import_wisdom_from_file(fp);

  /* We have now read everything from the file.
     Some of it is not in a usable state, however:
     In the ilists:
        bell,
	dualbell
	Ftrigtw_plan
     are not set properly.
     The plans do not even exist, although with the (invisible) 
     wisdom, they can be created rapidly.

     To get things working, we:
     1. put a structure on the bell_handle so we can access the proper bell
     2. create the trig plans. This is stolen from dyadic_g_ilist_init.
        It is not clear how to make this more general.
     3. go through all the intervals and align them to the correct bell
        and trig plan.
     */

  /* set halfpts */
  halfpts=numpts/2;

  /* initialize bell_holder. It holds all the bell values */

  /* first index, the level.
   We have twice as many so we can hold dual bells too*/
  bell_holder=(REAL ***)calloc(num_levels*2,sizeof(REAL **));
  assert(bell_holder != NULL);

  /* second index, which bell at that level. convention:
     0 -- symmetric left edge
     1 -- symmetric generic
     2 -- symmetric right edge
     3 -- asymmetric generic
     4 -- asymmetric right edge
     Not all levels will have all of these, but for simplicity we get
     memory for all of them.
     We also get twice as much for the dual bells*/
  bell_holder[0]=(REAL **)calloc(5*num_levels*2,sizeof(REAL *));
  assert(bell_holder[0] != NULL);
  
  /* align first bell. Only this one is used at level==0 */
  bell_holder[0][0]=saves_fordump->bell_handle;

  for(level=1;level<num_levels*2;level++){
    /* loop through the  levels, including extra for dual bells*/

    /* align the next one */
    bell_holder[level]=bell_holder[level-1]+5;

    /* set values based on the offsets */

    for(i=0;i<5;i++){
      /* loop through the types of bells */
      
      /* set pointer */
      bell_holder[level][i]=saves_fordump->bell_handle +
	saves_fordump->bell_offsets[level*5+i];
    }
  }

  /* make all the lct and dct saves */

  /* for each level there is a lct and a dct */
  plan_holder=(Ftrigtw_Plan *)calloc(2*num_levels,sizeof(Ftrigtw_Plan));
  assert(plan_holder != NULL);

  flags=(FFTW_MEASURE | FFTW_USE_WISDOM);

  for(level=0;level<num_levels; level++){
    /* loop through the levels */

    /* set winlength */
    winlength=halfpts>>level;

    /* make lctw plan */
    lctw_init(plan_holder+level,flags,winlength);

    /* make dctw plan. Offset by num_levels */
    dctw_init(plan_holder+level+num_levels,flags,winlength);
  }

  /* loop through the intervals and align */

  for(m=0;m<order_limit;m++){
    /* loop through the order */

    for(parity=0;parity<2;parity++){
      /* loop through even and odd */

      for(i=0;i<pmn_matrix[m].n_interval[parity];i++){
	/* loop through the intervals */

	/* deduce level */
	for(level=0;(pmn_matrix[m].ilist[parity][i].numpts)<(halfpts>>level);
	    level++);
	
	/* get transform plan */
	pmn_matrix[m].ilist[parity][i].transform_plan=
	  plan_holder[level+num_levels*
		     pmn_matrix[m].ilist[parity][i].which_transform];

	if(pmn_matrix[m].ilist[parity][i].left_extra==0){
	  /* left edge.
	     This must be done first because level 0 only has this one */

	  /* set bell and dual bell */
	  pmn_matrix[m].ilist[parity][i].bell=bell_holder[level][0];
	  pmn_matrix[m].ilist[parity][i].dualbell=
	    bell_holder[num_levels+level][0];

	}else if(pmn_matrix[m].ilist[parity][i].left_extra==
		  pmn_matrix[m].ilist[parity][i].numpts/2){
	  /* symmetric bell */

	  if(pmn_matrix[m].ilist[parity][i].right_extra==
		  pmn_matrix[m].ilist[parity][i].numpts/2){
	    /* generic */

	    /* set bell and dual bell */
	    pmn_matrix[m].ilist[parity][i].bell=bell_holder[level][1];
	    pmn_matrix[m].ilist[parity][i].dualbell=
	      bell_holder[num_levels+level][1];

	  } else{
	    /* right edge*/

	    /* set bell and dual bell */
	    pmn_matrix[m].ilist[parity][i].bell=bell_holder[level][2];
	    pmn_matrix[m].ilist[parity][i].dualbell=
	      bell_holder[num_levels+level][2];

	  }
	} else{
	  /* asymmetic */

	  if(pmn_matrix[m].ilist[parity][i].right_extra==
		  pmn_matrix[m].ilist[parity][i].numpts/2){
	    /* generic */

	    /* set bell and dual bell */
	    pmn_matrix[m].ilist[parity][i].bell=bell_holder[level][3];
	    pmn_matrix[m].ilist[parity][i].dualbell=
	      bell_holder[num_levels+level][3];

	  } else{
	    /* right edge*/

	    /* set bell and dual bell */
	    pmn_matrix[m].ilist[parity][i].bell=bell_holder[level][4];
	    pmn_matrix[m].ilist[parity][i].dualbell=
	      bell_holder[num_levels+level][4];

	  }
	}

      } /* i loop */
    } /* parity loop */
  } /* m loop */

  /* free things no longer needed */
  free(bell_holder[0]);
  free(bell_holder);

  /* this may cause memory leaks, since we lose our handle */
  free(plan_holder);

  /* set parameters for output */
  (*numpts_ptr)=numpts;
  (*band_limit_ptr)=band_limit;
  (*num_levels_ptr)=num_levels;
  (*node_ptr)=node;
  (*weight_ptr)=weight;
  (*pmn_matrix_ptr)=pmn_matrix;

#if ENTEREXIT
  printf("exiting load_fast1\n");
#endif
}
