/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** \file fast1_tofile.c

\brief Precompute and save data for fast1 spherical harmonic transform.

Summary:
- fast1_tofile drives the initialization and saving.
- init_fast1 does the actual initialization.
- dump_fast1 does the actual saving.

*/
#include "libftsh.h"
 
#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

REAL fast1_tofile(int  numpts,
		  int band_limit,
		  REAL *node,
		  REAL *weight,
		  REAL epsilon,
		  FILE *fp
		  ){
  /** This routine initializes the fast transform for spherical
      harmonics (version 1) and dumps it to a file. This version is
      designed for the novice user, and so doesn't allow tuning.

      INPUTS: 
      - numpts --  the number of points on the full interval. 
        Must be divisible by 4. More factors of 2 will improve
	performance.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
      - node --  an array of length numpts with the nodes,
        usually Arcsin of the usual gaussian nodes.
      - weight -- an array of length numpts with the quadrature weights
      - epsilon -- the level at which to truncate. The error in the
        transform will be slightly larger than this
     
      OUTPUTS: 
      - fp -- the objects created are written onto the file pointed to by
        fp, in binary format. The file must be opened before
	calling this routine.
      - returns the compression ratio, which is an upper bound on the
        speed up factor of using this algorithm

      NOTES:
      - node and weight are input so the user can select.
      - we hard-wire the largest number of levels legal. This will slow 
        initialization but is the best thing for novice users
      - The safety factor on the bell may not be quite right
      - we leak a little bit of memory from the trig plans
  */
  register int i,m;
  int order_limit,halfpts,num_levels;
  REAL (*bell_choice) (int,int);
  Pm_1D *pmn_matrix;
  Fast1_Save saves_fordump;
  REAL ratio,bell_fudge;

#if ENTEREXIT
  printf("entering fast1_tofile\n");
#endif

  /* set some global parameters*/

  /* set halfpts */
  halfpts=numpts/2;

  /* set order_limit to band_limit. */
  order_limit=band_limit;

  /* choose the bell, dependent on epsilon */

  /* note that the bellmat_ series are bi-orthogonal */

  /* The Pmn's we will be expanding are not normalized for the discretization,
     they are just point values of the analytically normalized functions.
     The main normalization missing is a factor of 1.0/sqrt(halfpts).
     If we truncate by epsilon in absolute terms, we truncate by 
     epsilon/sqrt(halfpts) in relative terms. When we do the multiply itself
     we accumulate sqrt(halfpts) times the truncation error, so the extra
     factors cancel out.

     The decision on which bell is based on relative, not absolute truncation.
     We therefore adjust out bell choice for this factor.
     It is likely that additionsl tuning here will be necessary.*/
  bell_fudge=1.0/sqrt((REAL) halfpts);

  /* safety factor; removed at end */
  epsilon*=bell_fudge;
  if(epsilon>1e-1){bell_choice=bellmat1;}
  else if(epsilon>1e-2){bell_choice=bellmat2;}
  else if(epsilon>1e-3){bell_choice=bellmat3;}
  else if(epsilon>1e-4){bell_choice=bellmat4;}
  else if(epsilon>5e-5){bell_choice=bellmat5;}
  else if(epsilon>1e-5){bell_choice=bellmat6;}
  else if(epsilon>1e-6){bell_choice=bellmat7;}
  else if(epsilon>1e-7){bell_choice=bellmat8;}
  else if(epsilon>1e-8){bell_choice=bellmat9;}
  else if(epsilon>5e-9){bell_choice=bellmat10;}
  else if(epsilon>1e-9){bell_choice=bellmat11;}
  else if(epsilon>1e-10){bell_choice=bellmat12;}
  else if(epsilon>1e-11){bell_choice=bellmat13;}
  else if(epsilon>1e-12){bell_choice=bellmat14;}
  else if(epsilon>5e-13){bell_choice=bellmat15;}
  else if(epsilon>1e-13){bell_choice=bellmat16;}
  else if(epsilon>5e-14){bell_choice=bellmat17;}
  else if(epsilon>1e-14){bell_choice=bellmat18;}
  else if(epsilon>5e-15){bell_choice=bellmat19;}
  else {bell_choice=bellmat20;}
  /* removal of safety */
  epsilon/=bell_fudge;

  /**** main initialization ************************************************/

  /* force initializer to choose maximum number of levels */
  num_levels=-1;

  init_fast1(&pmn_matrix,&saves_fordump,
	     halfpts, band_limit,node,&num_levels,epsilon,
	     bell_choice);

  /* compute the compression ratio, which we will return*/
  i=0;
  for(m=0;m<order_limit;m++){
    i+=pmn_matrix[m].matrix[0].non_zero+pmn_matrix[m].matrix[1].non_zero;
  }
  ratio=(halfpts*(band_limit*(band_limit+1))/2.0)/i;

  /* Dump to file */

  /* dump using subroutine */
  dump_fast1(numpts,band_limit,num_levels,node,weight,
	     &saves_fordump,pmn_matrix,fp);

  /* free everything that we can**
    We miss the trig plans */
  free(saves_fordump.bell_handle);
  free(saves_fordump.bell_offsets);
  for(m=0;m<order_limit;m++){
    /* loop through the orders */
    
    /* free the matrices */
    free_bdmat_memory(pmn_matrix[m].matrix);
    free_bdmat_memory(pmn_matrix[m].matrix+1);

    /* free the ilists. The bell pointers were freed in mass by bell_handle.
       we can't free the trig plans because we don't have a handle */
    free(pmn_matrix[m].ilist[0]);
    free(pmn_matrix[m].ilist[1]);
  }
  free(pmn_matrix);

#if ENTEREXIT
  printf("returning from fast1_tofile\n");
#endif

  /* return compression ration */
  return ratio;

}

void init_fast1(Pm_1D **out,
		Fast1_Save *saves_fordump,
		int halfpts,
		int band_limit,
		REAL *node,
		int *numlevels,
		REAL epsilon,
		REAL (*bell_choice) (int,int)
		){
  /** This routine initializes for the fast transform for spherical harmonic,
      version 1.
     
      Version 1 compresses in 1D (each Pmn separately) and chooses a
      single partition for each m and parity. By choosing a single
      partition we minimize the trig transform cost, but sacrifice
      precision. This is appropriate for small problems, near the
      break-even point, since it reduces overhead.

      INPUTS: 
      - halfpts -- half the number of points on the full interval.
        This implies the full number of points is even,
	and we also assume the nodes are symmetric.
	Currrently halfpts must also be even for our trig transforms to work.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
	For gaussian nodes this must be < 2*halfpts
	and for other nodes it should be even smaller.
	We confirn band_limit < 2*halfpts but leave the rest to the user.
      - node --  an array with the nodes, usually Arcsin of the usual
        gaussian nodes. We use halfpts of these, representing 
	half of the interval.
      - bell_choice -- a pointer to the bell function to use. This should
        be tuned to the desired precision. The choice of
	bell is subtle, and handled elsewhere.
      - numlevels -- the number of dyadic levels to use when trying to
        adapt. halfpts must be divisible by 2^numlevels.
	If numlevels is input too large or negative, this
	routine will choose the maximum allowed.
      - epsilon -- the target precision. The fast algorithm is an 
        approximate method, tunable to the desired precision.
	The algorithm is faster the less precision (larger
	epsilon but < 1) you request. On the other hand,
	doubling the number of digits slows the algorithm
	by less than a factor of 2, so asking for a little 
	extra is safe. If epsilon is chosen too close to
	the round-off error, the algorithm will degenerate
	to direct matrix applications.
 
      OUTPUTS: 
      - out -- memory is allocated for band_limit (==order_limit)
        Pm_1D structures, each representing the compressed 
	matrix for one m. Various bits of memory are
	allocated for these, and values filled in.
      - saves_fordump -- bell values and total length are filled in here
        so we can dump them to file efficiently.
      - numlevels -- is modified to the maximum allowed number if the
	input number was illegal.

      NOTES:
      - halfpts must be even for this routine to function. 
        Performance (speed) of the algorithm will improve if halfpts has 
	more factors of 2. For best performance, the non-power-of-two factor 
	in halfpts should be small (10 or so).
      - output       ? total coefficients,bands,rows,intervals
        ? modified num_levels
	?some holder for the trig plans
      - leaks some memory
      - right FFTW flags? import and export wisdom?
      - min_gap hard wired
      - bdmat_raw may not have enough bands

    */

  register int i,j,m,parity,level,within,index;
  int num_int_dyadic,coefs_inrow,order_limit,max_ints_kept,numbas[2],
    n_ints_kept,num_levels,n_int_lev,winlength,min_gap;
  Local_Trig_Interval *ilist[2],*kept_int,*ilist_raw;
  REAL *pmnraw[2],*coefraw;
  Dense_Matrix dmat_tmp;
  Pmncos_Recurrence_Save pmn_save;
  Dyadic_Gsearch_Save dy_gs_save;
  Pm_1D *pmn_matrix;
  Banded_Matrix bdmat_raw;
  int intervals_used;

#if ENTEREXIT
  printf("entering init_fast1\n");
#endif

  /* hard wire min_gap. Our sparse, banded structure is more efficient
     if we merge bands that only have a small gap between them.
     min_gap gives the minimal gap that we allow.*/
  min_gap=3;


  /*** compute maximum num_levels ***/
  
  /* halfpts must be even */
  assert(halfpts%2 ==0);

  /* count the number of factors of 2 */
  num_levels=1;
  for(i=halfpts/2;(i%2==0);i/=2)
    num_levels++;

  if(((*numlevels) < 1) || ((*numlevels) > num_levels)){
    /* if numlevels input is illegal, change it to the maximum */
    (*numlevels)=num_levels;
  }else{
    /* use the numlevels input */
    num_levels=(*numlevels);
  }

  /* construct the dyadic partition interval list.
     When we start to allow other partitions, they will be constructed here.*/
  dyadic_g_ilist_init(ilist,&num_int_dyadic,&coefs_inrow,
		      &(saves_fordump->bell_handle),
		      &(saves_fordump->bell_length),
		      &(saves_fordump->bell_offsets),
		      halfpts,num_levels,
		      bell_choice,FFTW_MEASURE | FFTW_USE_WISDOM);

  /* order_limit should be the same as band_limit */
  order_limit=band_limit;

  /* memory to hold the raw matrix*/
  pmnraw[0]=(REAL *)calloc(halfpts*band_limit,sizeof(REAL));
  assert(pmnraw[0]!=NULL);
  
  /*give half (rounding down) to odd*/
  pmnraw[1]=pmnraw[0]+halfpts*((band_limit+1)/2);

  /* memory to hold the raw coefficients 
     coefs_inrow * ((band_limit+1)/2)*/
  coefraw=(REAL *)calloc(coefs_inrow*((band_limit+1)/2),sizeof(REAL));
  assert(coefraw != NULL);

  /* holding memory for the chosen intervals of coefficients
   it needs halfpts per row and half (rounding up) band_limit rows */
  dmat_tmp=init_dmat_memory(halfpts*((band_limit+1)/2));
  /* the number of columns is always the same, so we assign it here */
  dmat_tmp.n_col=halfpts;

  /* get memory for and initialize Pmn recurrence */
  Pmncos_resave_init(&pmn_save,node,2*halfpts);

  /* initialize memory for dyadic grow search */
  dyadic_gsearch_init(&dy_gs_save,num_levels);

  /* memory for main compressed matrix holder */
  pmn_matrix=(Pm_1D *)calloc(order_limit,sizeof(Pm_1D));
  assert(pmn_matrix != NULL);

  /* get memory for two worst-case interval lists.
     once we know the actual length of a given list, we will get just 
     the right amount of memory and copy.*/
  max_ints_kept=(1<<(num_levels-1));
  ilist_raw=(Local_Trig_Interval *)
    calloc(max_ints_kept,sizeof(Local_Trig_Interval));
  assert(ilist_raw != NULL);

  /* get memory for the worst case banded matrices.
   after converting to banded format, we will allocate just the right 
   amount of memory for the given banded matrix and copy into that.
   This allows two bands for each fine level interval; probably more
   than enough, but not a firm bound*/
  bdmat_raw=init_bdmat_memory(halfpts*band_limit,band_limit,
			      (1<<num_levels)*band_limit);


  /*** begin main (order) loop *********************************************/
  for(m=0;m<order_limit;m++){
    /* loop through the order */
#if 0
    printf("m=%i\n",m);
#endif
    /* construct the matrix for that order */
    
    /* set the number of basis elements. There may be one less odd one*/
    numbas[0]=(order_limit-m+1)/2;
    numbas[1]=(order_limit-m)/2;

    for(i=0;i<numbas[1];i++){
      /* loop through the basis elements, by two's*/

      /* construct the even basis function*/
      next_rtsinPmncos(pmnraw[0]+i*halfpts,&pmn_save,halfpts,m,2*i);

      /* construct the odd basis function*/
      next_rtsinPmncos(pmnraw[1]+i*halfpts,&pmn_save,halfpts,m,2*i+1);

#if 0
      for(parity=0;parity<halfpts;parity++)
	printf("%i %e \n",parity,pmnraw[0][i*halfpts+parity]);
#endif

    }

    if(numbas[0] > numbas[1]){
      /*  there is one more even function*/

      /* construct the even basis function*/
      next_rtsinPmncos(pmnraw[0]+i*halfpts,&pmn_save,halfpts,m,2*i);
    }

    /* optionally allow change of sign convention */
#if 0
    printf("constructed matrix\n");
#endif

    for(parity=0;parity<2;parity++){
      /* loop through the two parities. 0==even, 1==odd */

      /* clear array of costs */
      for(level=0;level<num_levels;level++){
	/* loop through levels */
	
	/* note number of intervals on this level */
	n_int_lev=1<<level;
	
	for(within=0;within<n_int_lev;within++){
	  /* loop within the level */
	  
	  /* clear cost arrays */
	  dy_gs_save.cost_full[level][within]=0;
	  dy_gs_save.cost_asym[level][within]=0;
	}
	/* note that not all indices in cost_asym are actually used */
      }

      for(i=0;i<numbas[parity];i++){
	/* loop through the basis functions themselves */

	/* expand using  interval list */
	/* for now save all these coefficients */
	expand_intlist(coefraw+i*coefs_inrow,pmnraw[parity]+i*halfpts,
		       ilist[parity],num_int_dyadic);
      
	/* compute costs (number of coefficients) for each interval
	   and add to the running total */

	/* compute costs for the full bells */
	for(level=0;level<num_levels;level++){
	  
	  /* set interval size */
	  winlength=halfpts>>level;

	  /* note number of intervals on this level */
	  n_int_lev=1<<level;
	
	  for(within=0;within<n_int_lev;within++){
	    /* loop within the level */

	    dy_gs_save.cost_full[level][within]+=
	      count_thresh(coefraw+i*coefs_inrow
			   +level*halfpts
			   +within*winlength,
			   winlength,epsilon);
#if 0
	    printf("level=%i within=%i cost=%i\n",level,within,dy_gs_save.cost_full[level][within]);
#endif
	  }
	}
	
	/* compute costs for the asymmetric bells
	   same as above, except:
	   extra offset of halfpts*num_levels in the coefficients
	   no level==0 or level==num_levels -1
	   the costs go in a different place
	   no within==0*/
	for(level=1;level<num_levels-1;level++){
	  
	  /* set interval size */
	  winlength=halfpts>>level;
	  
	  /* note number of intervals on this level */
	  n_int_lev=1<<level;
	  
	  for(within=1;within<n_int_lev;within++){
	    /* loop within the level */

	    dy_gs_save.cost_asym[level][within]+=
	      count_thresh(coefraw+halfpts*num_levels
			   +(level-1)*halfpts
			   +i*coefs_inrow+within*winlength,
			   winlength,epsilon);
#if 0
	    printf("a level=%i within=%i cost=%i\n",level,within,dy_gs_save.cost_full[level][within]);
#endif
	  }
	}
      }
#if 0
      printf("expanded in intervals\n");
#endif

      /* run the search to determine the best partition */
      dyadic_gsearch(&dy_gs_save);
    
      /*** extract the partition choosen ***/

      gsearch_to_ilist(ilist_raw,&intervals_used,
		       max_ints_kept,ilist[parity],
		       &dy_gs_save,halfpts);
      
      /* get just enough memory for the interval list */
      pmn_matrix[m].ilist[parity]=(Local_Trig_Interval *)
	calloc(intervals_used,sizeof(Local_Trig_Interval));
      assert(pmn_matrix[m].ilist[parity] != NULL);

      /* note number of intervals */
      pmn_matrix[m].n_interval[parity]=intervals_used;

      /* copy the interval list*/
      for(i=0;i<intervals_used;i++)
	pmn_matrix[m].ilist[parity][i]=ilist_raw[i];


#if 0
      printf("pmn_matrix[%i].n_interval[%i]=%i\n",m,parity,
	     pmn_matrix[m].n_interval[parity]);
#endif
    
      /* pack the coefficients from the redundant representation into
	 a single partition representation.
	 As we do so, correct the offsets in the interval list.*/
    
      /*** HACK ***/
      n_ints_kept=pmn_matrix[m].n_interval[parity];
      kept_int=pmn_matrix[m].ilist[parity];
      
      for(i=0;i<n_ints_kept;i++){
	/* loop through the chosen intervals */
	
	for(index=0;index <numbas[parity];index++){
	  /* loop through the basis functions */
	  
	  for(j=0;j<kept_int[i].numpts;j++){
	    /* loop through the coefficients */
	    
	    /* read coefficients based of coef_offset and write based on
	       fcn_offset.*/
	    dmat_tmp.matrix[index*halfpts+kept_int[i].fcn_offset+j]
	      =coefraw[index*coefs_inrow+kept_int[i].coef_offset+j];
	  }
	}
	/* re-align coef_offset to agree with fcn_offset */
	kept_int[i].coef_offset=kept_int[i].fcn_offset;
	
      }
    
      /* set row size for the dmat. Column size was set at initialization */
      dmat_tmp.n_row=numbas[parity];
      
      /* convert the full coefficient matrix to banded matrix format */
      dmat_to_bdmat(&bdmat_raw,&dmat_tmp,epsilon,min_gap);
#if 0
printf("kept %i\n",bdmat_raw.non_zero);
#endif

      /* get just enough memory */
      pmn_matrix[m].matrix[parity]=
	init_bdmat_memory(bdmat_raw.non_zero,bdmat_raw.n_row,
			  bdmat_raw.band_p[bdmat_raw.n_row]/3);

      /* copy the banded matrix into place */
      bdmat_copy(&(pmn_matrix[m].matrix[parity]),&bdmat_raw);

#if 0
      dump_ascii_bdmat(&(pmn_matrix[m].matrix[parity]),stdout);
#endif
    }/* end parity loop */

  } /* end order loop */

  /* free the things we don't need any more, and don't need to dump */
#if 1
  /* don't need pmnraw[0/1] */
  free(pmnraw[0]);

  /* don't need coefraw */
  free(coefraw);

  /* don't need dmat_tmp.matrix */
  free(dmat_tmp.matrix);

  /* don't need pmn_save */
  Pmncos_resave_destroy(&pmn_save);

  /* don't need dy_gs_save */
  dyadic_gsearch_free(&dy_gs_save);

  /* don't need bdmat_raw */
  free_bdmat_memory(&bdmat_raw);

  /* don't need ilist_raw */
  free(ilist_raw);

  /* we don't need ilist[0/1], although we do need some things pointed
     to within the structure, so we don't do a complete free now.
     In a real program we need to be more careful, since this creates
     memory leaks*/
  free(ilist[0]);
#endif

  /* align out */
  (*out)=pmn_matrix;

#if ENTEREXIT
  printf("exiting init_fast1\n");
#endif
}

void dump_fast1(int numpts,
		int band_limit,
		int num_levels,
		REAL *node,
		REAL *weight,
		Fast1_Save *saves_fordump,
		Pm_1D *pmn_matrix,
		FILE *fp
		){
  /** This routine dumps the initialization constructed by init_fast1 and
      general parameters for the fast transform for spherical harmonics,
      version 1.

      INPUTS: 
      - numpts --  the number of points on the full interval.
      - band_limit -- the bound on the degree of the expansion.
        In \\(P^m_n\\) notation this bounds n.
      - num_levels -- the number of dyadic levels used in the adaption   
      - node --  an array of length numpts with the nodes,
        usually Arcsin of the usual gaussian nodes.
      - weights -- an array of length numpts with the quadrature weights
      - saves_fordump -- a structure containing the information needed
        for the bells. Constructed by init_fast1
      - pmn_matrix -- an array of band_limit (==order_limit)
        Pm_1D structures, each representing the compressed 
	matrix for one m.

      OUTPUTS: fp -- the objects above are written into the file
      pointed to by fp, in binary format. The file must be opened
      before calling this routine.

  */

  register int i,m,order_limit;
  int num_rows,num_elements,num_bands,num_intervals;


#if ENTEREXIT
  printf("entering dump_fast1\n");
#endif

  /* order_limit is the same as band_limit */
  order_limit=band_limit;

  /* dump numpts and band_limit */
  fwrite(&numpts,sizeof(int),1,fp);
  fwrite(&band_limit,sizeof(int),1,fp);

  /* we could dump epsilon for diagnostics */

  /*  dump num_levels so we can recreate trig stuff */
  fwrite(&num_levels,sizeof(int),1,fp);

  /* dump nodes and weights */
  /* we could probably just dump halfpts of them */
  fwrite(node,sizeof(REAL),numpts,fp);
  fwrite(weight,sizeof(REAL),numpts,fp);

  /* dump saves_fordump, which contains the bells */
  fwrite(&(saves_fordump->bell_length),sizeof(int),1,fp);
  fwrite(saves_fordump->bell_handle,sizeof(REAL),
	 saves_fordump->bell_length,fp);
  fwrite(saves_fordump->bell_offsets,sizeof(int),10*num_levels,fp);

  /* determine overall memory requirements *****/
  num_rows=0;
  num_elements=0;
  num_bands=0;
  num_intervals=0;
  for(m=0;m<order_limit;m++){
    /* loop through the orders */

    /* accumulate number of rows; include one extra per matrix */
    num_rows+=pmn_matrix[m].matrix[0].n_row+pmn_matrix[m].matrix[1].n_row+2;

    /* accumulate number of elements */
    num_elements+=pmn_matrix[m].matrix[0].non_zero+
      pmn_matrix[m].matrix[1].non_zero;

    /* accumulate number of bands */
    num_bands+=(pmn_matrix[m].matrix[0].band_p[pmn_matrix[m].matrix[0].n_row]+
		pmn_matrix[m].matrix[1].band_p[pmn_matrix[m].matrix[1].n_row])
      /3;

    /* accumulate number of intervals */
    num_intervals+=pmn_matrix[m].n_interval[0]+pmn_matrix[m].n_interval[1];
  }

  /* output num_rows,num_elements,num_bands,num_intervals */
  fwrite(&num_rows,sizeof(int),1,fp);
  fwrite(&num_elements,sizeof(int),1,fp);
  fwrite(&num_bands,sizeof(int),1,fp);
  fwrite(&num_intervals,sizeof(int),1,fp);
  
  /* main martix output.
     we write everything needed for the banded matrices,
     but only the parameters for the Local_Trig_Intervals */
  for(m=0;m<order_limit;m++){
    /* loop through the orders */

    /* dump matrix[0] and [1] */
    dump_bin_bdmat(&(pmn_matrix[m].matrix[0]),fp);
    dump_bin_bdmat(&(pmn_matrix[m].matrix[1]),fp);

    /* dump n_interval[0/1]*/
    fwrite(&(pmn_matrix[m].n_interval[0]),sizeof(int),1,fp);
    fwrite(&(pmn_matrix[m].n_interval[1]),sizeof(int),1,fp);

    /* dump interval lists [0/1];
       We dump the entire structure, including pointers. When reloaded these
       pointers will be gibberish*/
    for(i=0;i<pmn_matrix[m].n_interval[0];i++)
      fwrite(pmn_matrix[m].ilist[0]+i,sizeof(Local_Trig_Interval),1,fp);
    for(i=0;i<pmn_matrix[m].n_interval[1];i++)
      fwrite(pmn_matrix[m].ilist[1]+i,sizeof(Local_Trig_Interval),1,fp);
  }

  /* dump fftw wisdom. We do this last because we have less control over it*/
  fftw_export_wisdom_to_file(fp);


#if ENTEREXIT
  printf("exiting dump_fast1\n");
#endif
}
