/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** \file init_Pm_1d_workspace.c

\brief Initialize workspace for the transform.

*/
#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

/*************************************************************************/
void init_Pm_1d_workspace(Pm_1D_Workspace *out,
			  int numpts,
			  int band_limit,
			  REAL *node,
			  REAL *weight
			  ){
  /** This routine initializes a structure Pmncos_1D_Workspace,
      which is used to hold workspace and a few precomputed values 
      for use in the synthesis/analysis routines.

      INPUTS:  
      - numpts -- the number of points in the interval
        (number of latitudes) used here to determine memory needed.
      - band_limit -- the limit on the degree of the expansion.
        Used here to determine the maximal number
	of coefficients possible for one order.
      - node -- an array of numpts REALs specifying the discretization.
        Usually this will be gaussian nodes.
      - weight -- an array of numpts REALs specifying the quadrature
        weight. Usually this will be gaussian weight.

      OUTPUT: out -- memory is allocated for the fields of out, and
      some values are precomputed.
    
      NOTE:
      - probably should pack the memory together
  */

  register int i;

#if ENTEREXIT
  printf("entering init_Pm_1d_workspace\n");
#endif

  /********* note global parameters ******************/

  /* note numpts */
  out->numpts=numpts;

  /* note band_limit */
  out->band_limit=band_limit;

  /***********get dvect memory**********/

  /* get all of it and give to dvect_coef[0].
     We need (band_limit+1)/2 for dvect_coef[0]
             (band_limit/2 for dvect_coef[1]
	     (numpts+1)/2 for dvect_fcn[0]
	     numpts/2 for dvect_fcn[1]
	     */
  out->dvect_coef[0]=init_dvect_memory(band_limit+numpts);

  /* set the size for out->dvect_coef[0]*/
  out->dvect_coef[0].n_row=(band_limit+1)/2;

  /* give the rest to dvect_coef[1]*/
  give_dvect_memory(out->dvect_coef+1,out->dvect_coef);

  /* set the size for out->dvect_coef[1]*/
  out->dvect_coef[1].n_row=band_limit/2;

  /* give the rest to dvect_fcn[0]*/
  give_dvect_memory(out->dvect_fcn,out->dvect_coef+1);

  /* set the size for out->dvect_fcn[0]*/
  out->dvect_fcn[0].n_row=(numpts+1)/2;

  /* give the rest to dvect_fcn[1]*/
  give_dvect_memory(out->dvect_fcn+1,out->dvect_fcn);

  /* set the size for out->dvect_fcn[1]*/
  out->dvect_fcn[1].n_row=numpts/2;


  /******* get REAL array memory *********/

  /* get all of it and give it to eval_fcn[0]
     we need (numpts+1)/2 for eval_fcn[0]
              numpts/2 for eval_fcn[1]
	      numpts for over_rootsin
	      numpts for mod_gweight
	      */
  out->eval_fcn[0]=(REAL *)calloc(3*numpts,sizeof(REAL));
  assert(out->eval_fcn[0] != NULL);

  /* leave (numpts+1)/2 for eval_fcn[0] and give the rest to [1] */
  out->eval_fcn[1]=out->eval_fcn[0]+(numpts+1)/2;

  /* leave numpts/2 for eval_fcn[1] and give the rest to over_rootsin */
  out->over_rootsin=out->eval_fcn[1]+numpts/2;

  /***** precompute over_rootsin so we can quickly remove this factor ******/
  /* start at 1.0 */
  for(i=0;i<numpts;i++)
    out->over_rootsin[i]=1.0;
  /* remove sqrt(sin()) at the given nodes */
  remove_rootsin(out->over_rootsin,out->over_rootsin,node,numpts);

  /* leave numpts for over_rootsin and give the rest to mod_weight */
  out->mod_weight = out->over_rootsin + numpts;

  /* make mod_weight into weight*over_rootsin */
  pointwise_multiply(out->mod_weight,out->over_rootsin,weight,numpts);

#if ENTEREXIT
  printf("exiting init_Pm_1d_workspace\n");
#endif

}
