/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** \file lctw.c

\brief Local Cosine Transform based on FFTW.

Summary:
- lctw_init initializes; do once. 
- lctw does the transform.

*/


#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void lctw(REAL *fcn,
	  Ftrigtw_Plan *main_plan
	  ){
  /** Computes the local cosine transform of fcn and puts the coeficients 
      back into fcn. 

      The local cosine transform is also known as 
      the type 4 discrete cosine transform.

      This is the LCT at half integer points. Is its own inverse.

      INPUTS:
      - fcn   array containing the data
      - main_plan -- a structure holding precomputed trig functions,
        the plan used by fftw, and the number of points; 
	initialized by lctw_init.
		      
      OUTPUT: fcn.  Puts the coefficients in fcn.  specifically:
      \\[ f(k) =norm \\sum_{j=0}^{numpts-1} f(j)
      \\cos\\left(\\pi\\frac{(k+1/2)(j+1/2)}{numpts}\\right)
      \\]

                   numpts-1              (k+0.5)  (j+0.5)
            fcn[k]= SUM   fcn[j] cos( PI ----------------  ) *norm
	           j=0                         numpts   
      set for \\(L^2\\) normalization.

      NOTES:
      - comment better
      - requires even numpts. checked by lctw_init
  */

  register int i,halfn,numpts;
  REAL *lctw_save;
  fftw_complex *work_in,*work_out;

#if ENTEREXIT
  printf("entering lctw\n");
#endif

  /* note number of points */
  numpts=main_plan->numpts;

  /* align the initial trig functions */
  lctw_save=main_plan->trig_save;

  /* note half the number of points */
  halfn=numpts>>1;

  /* split workspace in two pieces */
  work_in=main_plan->complex_work;
  work_out=work_in+halfn;
  
  /*     lctw_save[2*i]=cos( (i+0.25)*PI/numpts)
     lctw_save[2*i+1]=sin( (i+0.25)*PI/numpts)*/ 

  /* form work_in[k]=(fcn[2k]+i fcn[numpts-1-2k])exp(-i(k+1/4)PI/numpts)*/ 
  for(i=0;i<halfn;i++){
    work_in[i].re=fcn[2*i]*lctw_save[2*i]
      +fcn[numpts-1-2*i]*lctw_save[2*i+1];
    work_in[i].im=-fcn[2*i]*lctw_save[2*i+1]
      +fcn[numpts-1-2*i]*lctw_save[2*i];
  }

  /* do a complex fft of work_in and put the result in work_out **/
  /* this makes work_out[k]= SUM_{j=0}^{j=N/2-1} work_in[j] exp(-4iPIkj/N)*/
  fftw_one(main_plan->w_plan,work_in,work_out);

  /* shift the trig saves to the second half */
  lctw_save+=numpts;
  /* now
     lctw_save[2*i]=cos( i*PI/numpts) * norm
     lctw_save[2*i+1]=sin( i*PI/numpts) * norm */

  /* construct
     fcn[2k]=REAL(exp(-ikPI/N)work_out[k])
     fcn[N-1-2k]= -IMAG( same ) 
     and normalize*/
  for(i=0;i<halfn;i++){
    fcn[2*i]=work_out[i].re*lctw_save[2*i]+work_out[i].im*lctw_save[2*i+1];
    fcn[numpts-1-2*i]= work_out[i].re*lctw_save[2*i+1]
      -work_out[i].im*lctw_save[2*i];
  }
  
  /* The fact that this process fcn -> work_in -> work_out -> fcn
     yields the lct is a matter of trig identities */
#if ENTEREXIT
  printf("exiting lctw\n");
#endif

}

void lctw_init(Ftrigtw_Plan *main_plan,
		int flags,
		int numpts
		){
  /** This routine initializes for lctw

  We initialize trigonometic things used by the lct and dct portions,
  and call the initailizer for the fftw real transform (forward).

  INPUTS: 
  - main_plan -- a structure that will hold the things we generate
    Memory for its fields is allocated in this routine.
  - flags -- flags passed directly to the rfftw initializer
  - numpts -- the number of points on which we will transform
    MUST be even

  OUTPUTS: main_plan is filled in:
     

  NOTES:
  - comment better
     
  ********/
  
  register int i,halfn;
  register REAL *local_save;
  /* keep these always as double so we don't lose precision **/
  register double arg, norm;

#if ENTEREXIT
  printf("entering lctw_init\n");
#endif

  /* make sure numpts is even */
  assert(numpts%2==0);

  /* note length of the transform */
  main_plan->numpts=numpts;

  /* some things are half the length */
  halfn=numpts>>1;

  /* initialize workspace. This will be split into two workspaces
     of half the length */
  main_plan->complex_work=(fftw_complex *)calloc(numpts,sizeof(fftw_complex));

  /* create fftw_plan */
  main_plan->w_plan=fftw_create_plan(halfn, FFTW_FORWARD,flags);

  /* get memory for trig_save
   numpts/2 real/imaginary pairs
   + another numpts/2 real/imaginary pairs*/
  local_save=(REAL *)calloc(2*numpts,sizeof(REAL));

  /* align trig_save with local_save */
  main_plan->trig_save=local_save;

  /* intialize trig functions.
     trig_save[2*i]=cos( (i+0.25)*PI/numpts)
     trig_save[2*i+1]=sin( (i+0.25)*PI/numpts)*/
  for(i=0;i<halfn;i++){
    arg=(i+0.25)*PI/numpts;
    local_save[2*i]=cos(arg);
    local_save[2*i+1]=sin(arg);
  }

  /* align for second half */
  local_save+=numpts;

  /* compute normalizing factor */
  norm=sqrt(1.0/halfn);

  /* intialize trig functions.
     trig_save[2*i]=cos( i*PI/numpts) *norm
     trig_save[2*i+1]=sin( i*PI/numpts) * norm*/
  for(i=0;i<halfn;i++){
    arg=i*PI/numpts;
    local_save[2*i]=cos(arg)*norm;
    local_save[2*i+1]=sin(arg)*norm;
  }
#if ENTEREXIT
  printf("exiting lctw_init\n");
#endif
}







