/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/** \file libftsh.h

\brief Header for a library for doing the Fast Transform for Spherical Harmonics.
*/


/* make sure we haven't already included this library **/
#ifndef LIBFTSH_PRESENT
#define LIBFTSH_PRESENT


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <time.h>


#ifndef PI
#define PI 3.14159265358979323846264338328
#endif
 
/** USE_double sould be defined by a compiler flag
    - 0 for float
    - 1 for double 
 */
#ifndef USE_double
#define USE_double 1
#endif

/** REAL is the chosen floating-point data type (float or double) */
#if USE_double
#define REAL double
#else
#define REAL float
#endif

/* Deal with possible ways fftw was compiled */
#if 0
  /** fftw has been compiled with both precisions */
#if USE_double
  /* double precision header */
#include "dfftw.h"
#include "drfftw.h"
#else
  /* single precision header */
#include "sfftw.h"
#include "srfftw.h"
#endif
#else /*fftw in only one precision. pray it is the right precision */
/*  Make sure it is compiled with fftw_real == REAL */
#include "fftw.h"
#include "rfftw.h"
#endif

/*************************************************************************/
/**** data structures ****************************************************/
/*************************************************************************/


/** \brief Holder for precomputed FFTW objects.

This structure holds all the precomputed objects necessary for doing a
fast trigonometric transform (without bells), using fftw.  The
different transforms (Local Cosine, Discrete Cosine) may use different
fields in different ways. We keep a single structure so we can pass it
through higher level routines without difficulty.  We organize this
into a structure to make it easier to pass around and modify.

NOTES:
- currently numpts must be even for lctw
- The lct is self-inverting but the dctw is not
*/
typedef struct{
  /** The length of the transform.   */
  int numpts;
  /** Complex workspace for interacting with the fft.   */
  fftw_complex *complex_work;
  /** The plan for the fftw routine.  */
  fftw_plan w_plan;
  /** The plan for the rfftw routine in forward mode.   */
  rfftw_plan rw_plan_forward;
  /** The plan for the rfftw routine in reverse mode.   */
  rfftw_plan rw_plan_reverse;
  /** Precomputed trig things for lctw or dctw.   */
  REAL *trig_save; 
}Ftrigtw_Plan; 

/** \brief An ordinary (one dimensional) matrix.
*/
typedef struct 
{
  /** The amount of memory allocated for this matrix. This may be
      quite a lot larger than the amount needed for this particular
      matrix, but allows us to check if we have enough room when we
      form a matrix. Needs to be at least n_col*n_row.
   */
  int max_size;
  /** The number of colums in the matrix.  */
  int n_row;
  /** The number of rows in the matrix.   */
  int n_col;
  /** The values of the entries. The indexing convention is that the
      element (row,col) is in matrix[col+row*n_col]
   */
  REAL *matrix;
} Dense_Matrix;

/** \brief An ordinary (one dimensional) vector. 

We label it as if it was a column vector, but this doesn't really
matter.
*/
typedef struct 
{
  /** The amount of memory allocated for this vector. This may be
      quite a lot larger than the amount needed for this particular
      vector, but allows us to check if we have enough room when we
      form a vector. Needs to be at least n_row.
   */
  int max_size;
  /** The number of rows in this column vector. This is just the length
      of the vector.
  */
  int n_row;
  /** The values of the entries. The (row) entry is in vector[row].  */
  REAL *vector;
} Dense_Vector;

/** \brief A one dimensional sparse matrix, with banded structure.

NOTES:
- In normal usage:
  +  band_p points to max_rows int's,
  + band_list points to 3*max_bands int's and
  + matrix to max_size REALs. 
  If they have less than that, some routines
  may write past the allocated memory.
- The encoding of bands is different than in the first ftsh1.
*/
typedef struct 
{
  /** The amount of memory allocated for this matrix. This may be
      quite a lot larger than the amount needed for this particular
      matrix, but allows us to check if we have enough room when we
      form a matrix. Needs to be at least non_zero.
   */
  int max_size;
  /** The amount of memory allocated for row_p. Keeping track of 
      this separately from max_size gives us greater flexibility
   */
  int max_rows;
  /** (half) the amount of memory available for band_list.
      Needs to be at least band_p[n_row].
   */
  int max_bands;
  /** The (theoretical) number of rows in the matrix. */
  int n_row;
  /** The (theoretical) number of colums in the matrix. */
  int n_col;
  /** The actual number of non-zero entries in the matrix.   */
  int non_zero;
  /** A list of n_row+1 integers that keep track of which bands are in
      which row of the matrix. band_p[i] holds (three times) the total
      number of bands used on all previous rows.  It is used as an
      offset within bandlist. band_p[i] gives the beginning of the
      bands for that row, and band_p[i+1] gives a marker just past the
      end of the bands for that row.  The number of bands on a row is
      given by (band_p[i+1]-band_p[i])/2. Note that we need access to
      band_p[n_row].
   */
  int *band_p;
  /** A list of ints that encodes the bands kept.
      For each band we have three consecutive entries:
      - the first index in the band;
      - ONE PAST the last index in the band;
      - an offset telling where the coefficients start in matrix
        (at matrix[offset])
      If you subtract the band start from the band end and add over
      all the bands, you should get non_zero.
      There are 3*band_p[n_row] ints used.
   */
  int *band_list;
  /** A list of non_zero REALs that gives the value of an entry.   */
  REAL *matrix;
} Banded_Matrix;

/** \brief Workspace for Banded_Matrix routines.

Bdmat_Op_Workspace holds workspace used by several bdmat routines.
Although not all routines use all the workspace, we gather all the
possible workspaces together in this structure so that we can pass it
through other routines more easily. It also carries information on the
workspace size so we can check that it's enough.  If we add other
workspaces into this structure, the routines that pass it won't need
to be modified.

NOTES: 
- Many routines assume the workspaces start all zero.
  Any routine which uses them must therefore clean up after itself.
- Is twice as many ints really appropriate?
*/
typedef struct{
  /** How much memory we have in this structure.  */
  unsigned int max_size; 
  /** An array of int's of length 2*max_size .   */
  int *int_work;
  /** An array of REALs of length max_size.   */
  REAL *real_work;
} Bdmat_Op_Workspace;

/** \brief Workspace for "best non-decreasing dyadic partition" search.

The structure Dyadic_Gsearch_Save is used to hold memory and some
precomputed orderings for the "best non-decreasing dyadic partition"
search. It is normaly initialized by dyadic_gsearch_init, used
multiple times by dyadic_gsearch, and then freed by
dyadic_gsearch_free.
*/
typedef struct{
  /** Tells at which level to start our partition at the left
      edge. From there we can follow branching to get the partition.
   */
  int best_level; 
  /** The total number of levels to use.  */
  int num_levels;
  /** Used to keep track of the partition. Normally: For a particular
      [level][within] a value of 
      - 0 means at the right edge 
      - 1 means keep the samesize and 
      - 2 means go up a size 
      The first index goes from 0 to num_levels-1 
      and the second from 0 to 2^level -1
  */
  unsigned short int **branching;
  /** Stores the cost of a full bell expansion. Indexed like branching.  */
  unsigned int **cost_full;
  /** Stores the cost of a asymmetric bell expansion.  Indexed like
      cost_full, but the entries [0][_], [_][0] and [num_levels-1][_]
      are not used
   */
  unsigned int **cost_asym;
  /** Keeps track of the ordering in which to process the nodes. Here
      we keep just the level of the nodes
   */
  unsigned int *order_level;
  /** Here we keep the index of the ordered nodes within their
      respective levels. When combined with orderlev this identifies
      which interval is next.
   */
  unsigned int *order_within;
}Dyadic_Gsearch_Save;

/** \brief Extra things to save for fast1 algorithm.

This structure holds things computed during init_fast1 that need to be
saved to file separately from the main matrices.
   
Notes:
- We could save some structure of offsets within bell_handle. We can't 
  save pointers but this gives a compromise. Otherwise we need to recreate
  the structure of the bells, and so assume some (probably dyadic)
  partition.
*/
typedef struct{
  /** The total length of all the bells to save.  */
  int bell_length;
  /** A list of integer offsets to use within bell_handle to get 
      particular bells. There are 5*2*num_levels entries.
      The 2 is for bells and dual bells.
      The num_levels is because we have different bells for each level
      The five is for the possible types of bells at each level:
      - 0 -- symmetric left edge
      - 1 -- symmetric generic
      - 2 -- symmetric right edge
      - 3 -- asymmetric generic
      - 4 -- asymmetric right edge
      Not all levels will have all of these. If a bell is not used
      we give it offset 0
  */
  int *bell_offsets;
  /** An array of length bell_length holding all bell values. */
  REAL *bell_handle;
}Fast1_Save;

/** \brief Information to do a local trigonometric transform on an interval. 

NOTES:
- It assumes the function values and coeffiecients are each stored in 
  a contiguous array, and so is not compatible with 
  multidimensional blocking.
- Eventually allow 
  + which_transform==2 --> Discrete Sine Transform and
  + which_transform==3 --> Local Sine Transform
 */ 
typedef struct{
  /** The number of points in the selected interval.
      Also is the number of coefficients produced.
  */
  int numpts;
  /** The amount the bell spills to the left of the main interval.  */
  int left_extra;
  /** The amount the bell spills to the right of the main interval. */
  int right_extra;
  /** An offset telling where our subinterval starts within the full
      array of function values. Generally this represents a physical
      location.
   */
  int fcn_offset;
  /** An offset telling where our subinterval starts within the full
      array of coefficient values. Generally this does NOT represent a
      physical location, because we leave room for various other
      redundant representations of the same physical location.
   */
  int coef_offset;
  /** A pointer to the precomputed bell values for the expansion
      transform. When in use, points to left_extra+numpts+right_extra REALs
   */
  REAL *bell;
  /** A pointer to the precomputed bell values for the evaluation
      transform. When in use, points to left_extra+numpts+right_extra
      REALs. If the bell used was orthogonal then the bell and dualbell
      values are the same.
  */
  REAL *dualbell;
  /** A flag telling us which local trigonometric transform
      to use. This also tells us the parity of the folds
      at the two edges. Current legal values are:
      - 0 -- Local Cosine Transform (dct4). Even fold at the left
        and odd fold at the right
      - 1 -- Discrete Cosine Transform. Even folds at both edges
      - 9 -- no transform, no bell; simple copy
   */
  int which_transform;
  /** A structure that holds precomputed things used by
      the trig transform itself.
   */
  Ftrigtw_Plan transform_plan;
}Local_Trig_Interval;

/** \brief Hold a compressed matrix of Associate Legendre functions of a fixed order. 

The matrix has been separated into even and odd parts, and each of
these has been compressed in a local trig basis.
*/
typedef struct{
  /** The maximum number of intervals we can hold. 
      We must have n_ints[0]+n_ints[1] <= max_intervals.
      Often we will get a large amount of memory, use some,
      and then pass the rest to the next object.
  */
  int max_intervals;
  /** The length of ilist[ ] */
  int n_interval[2];
  /** The compressed version of the even (respectively odd) part, 
      in Banded_Matrix format. */
  Banded_Matrix matrix[2];
  /** The intervals used in this representation. This can give
      a single basis or an over-representation, as long as it
      is consistent with matrix[ ]. */
  Local_Trig_Interval *ilist[2];
}Pm_1D;

/** \brief Hold a banded matrix of Associate Legendre functions of a fixed order.

The matrix is not compressed, but by using the banded format we can
discard areas that are near zero.  The matrix has been separated into
even and odd parts.
*/
typedef struct{
  /** 
      - matrix[0] -- The even part, in Banded_Matrix
      - matrix[1] -- same, but for the odd part of the matrix.
  */
  Banded_Matrix matrix[2];
}Pm_Direct_bd;

/** \brief Hold a dense matrix of Associate Legendre functions of a fixed order.

The matrix has been separated into even and odd parts.
*/
typedef struct{
  /**
  - matrix[0] -- The even part, in Dense_Matrix
  -  matrix[1] -- same, but for the odd part of the matrix. */
  Dense_Matrix matrix[2];
}Pm_Direct_d;

/** \brief Hold workspace and a few precomputed values for the fast synthesis/analysis.
   
In normal usage,
- dvect_coef[0/1] holds the even and odd index
  spherical harmonic coefficients,
- dvect_fcn[0/1] holds the even and odd parity function values in
  local cosine coordinates,
- eval_fcn[0/1] holds the even and odd parity function values in ordinary
  coordinates,
- over_rootsin hold 1/sqrt(sin()) at the gaussian nodes,
- mod_weight holds the gaussian quadrature weights times over_rootsine
*/
typedef struct{
  /** The number of points in the full interval, used here to 
      specify the memory needed and set over_rootin */
  int numpts;
  /** The limit on the degree n allowed, used here to
      specify the memory needed. */
  int band_limit;
  /** An array of two Dense_Vectors, of length 
      (band_limit+1)/2 and band_limit/2 */
  Dense_Vector dvect_coef[2];
  /** An array of two Dense_Vectors, of length (numpts+1)/2 and numpts/2 */
  Dense_Vector dvect_fcn[2];
  /** An array of two REAL arrays, of length (numpts+1)/2 and numpts/2 */
  REAL *eval_fcn[2];
  /** An array of REALs of length numpts */
  REAL *over_rootsin;
  /** An array of REALs of length numpts */
  REAL *mod_weight;
}Pm_1D_Workspace;


/** \brief Hold data and workspace to do the three-term recurrence for a vector of sample points.
*/
typedef struct{
  /** The number of sample points. For spherical harmonic usage this
      is number of latitudes. We assume the samples are symmetric
      about PI/2 (we are on [0,PI]). We only compute values at
      (numpts+1)/2 values, since the others are obtained by reflection. */
  int numpts;
  /** An array of size 2x(numpts+1)/2 (ie sofar[2][(numpts+1)/2]).
      This will contain samples of the previous two Pmn's 
      (at n-1 and n-2) with their overall size stored separately in size. */
  REAL **sofar;
  /** An array of size (numpts+1)/2 containing an extracted power
      of 2. To get true values from sofar, multiply by 2^size for the
      corresponding point. */
  int *size;
  /** An array of size (numpts+1)/2, holding cos(theta_i) where theta_i
      are your sampling points. */
  REAL *ex;
  /** An array of size (numpts+1)/2, holding sin(theta_i) where theta_i
      are your sampling points. */
  REAL *why;
}Pmncos_Recurrence_Save;

/*************************************************************************/
/******* END Data Structures *********************************************/
/*************************************************************************/


/*************************************************************************/
/******* BEGIN function prototypes ***************************************/
/*************************************************************************/

/* dvect.c */
Dense_Vector init_dvect_memory(unsigned  int n
			       );
void give_dvect_memory(Dense_Vector *current,
		       Dense_Vector *previous
		       );
/* dmat.c */
Dense_Matrix init_dmat_memory(unsigned  int n
			       );
void give_dmat_memory(Dense_Matrix *current,
		       Dense_Matrix *previous
		       );
void free_dmat_memory(Dense_Matrix *in
		      );
void dump_bin_dmat(Dense_Matrix *x,
		   FILE *out
		   );
void load_bin_dmat(Dense_Matrix *x,
		   FILE *in
		   );
/* dmat_ops.c */
void dmat_dvect_multiply(Dense_Vector *out,
			 Dense_Matrix *inmat,
			 Dense_Vector *invect
			 );
void dmat_t_dvect_multiply(Dense_Vector *out,
			   Dense_Matrix *inmat,
			   Dense_Vector *invect
			   );
/* bdmat.c */
Banded_Matrix init_bdmat_memory(int max_size,
				int max_rows,
				int max_bands
				);
void give_bdmat_memory(Banded_Matrix *current,
		       Banded_Matrix *previous
		       );
void free_bdmat_memory(Banded_Matrix *in
		       );
void bdmat_copy(Banded_Matrix *out,
		Banded_Matrix *in
		);
void dump_bin_bdmat(Banded_Matrix *x,
		    FILE *out
		    );
void load_bin_bdmat(Banded_Matrix *x,
		    FILE *in
		    );
/* bdmat_ops.c */
void bdmat_dvect_multiply(Dense_Vector *out,
			  Banded_Matrix *inmat,
			  Dense_Vector *invect
			  );
void bdmat_t_dvect_multiply(Dense_Vector *out,
			    Banded_Matrix *inmat,
			    Dense_Vector *invect
			    );
void dmat_to_bdmat(Banded_Matrix *out,
		   Dense_Matrix *in,
		   REAL cutoff,
		   int min_gap
		   );
/* dyadic_gsearch.c */
int dyadic_gsearch(Dyadic_Gsearch_Save *dy_gs_save
		   );
void dyadic_gsearch_init(Dyadic_Gsearch_Save *dy_gs_save,
			 int num_levels
			 );
void dyadic_gsearch_free(Dyadic_Gsearch_Save *dy_gs_save
			 );
/* fast1_fromfile.c*/
void fast1_fromfile(int *numpts_ptr,
		    int *band_limit_ptr,
		    REAL **node_ptr,
		    REAL **weight_ptr,
		    Pm_1D **pmn_matrix_ptr,
		    Pm_1D_Workspace *pmnwork,
		    FILE *fp
		    );
void load_fast1(int *numpts_ptr,
		int *band_limit_ptr,
		int *num_levels_ptr,
		REAL **node_ptr,
		REAL **weight_ptr,
		Fast1_Save *saves_fordump,
		Pm_1D **pmn_matrix_ptr,
		FILE *fp
		);
/* direct1_fromfile.c*/
void direct1_fromfile(int *numpts_ptr,
		      int *band_limit_ptr,
		      REAL **node_ptr,
		      REAL **weight_ptr,
		      Pm_Direct_d **pmn_matrix_ptr,
		      Pm_1D_Workspace *pmnwork,
		      FILE *fp
		      );
void load_direct1(int *numpts_ptr,
		  int *band_limit_ptr,
		  REAL **node_ptr,
		  REAL **weight_ptr,
		  Pm_Direct_d **pmn_matrix_ptr,
		  FILE *fp
		  );
/* dctw.c */
void dctw(REAL *fcn,
	  Ftrigtw_Plan *main_plan,
	  int dir
	  );
void dctw_init(Ftrigtw_Plan *main_plan,
	       int flags,
	       int numpts
	       );
void rfftw_to_NR_format(REAL *data_nr,
			REAL *data_rfftw,
			int numpts,
			int dir
			);
/* lctw.c */
void lctw(REAL *fcn,
	  Ftrigtw_Plan *main_plan
	  );
void lctw_init(Ftrigtw_Plan *main_plan,
	       int flags,
	       int numpts
	       );
/* bell_utils.c */
REAL dualabell(REAL (*origbell) (int,int),
		 int laplength,
		 int t
		 );
void bell_1size_init(REAL *bellout,
		     REAL (*whichbell) (int,int),
		     int lapleft,
		     int center,
		     int lapright,
		     int dual_flag
		     );
void fold_e_e_1s(REAL *coef,
		 REAL *fcn,
		 int numpts,
		 REAL *left,
		 int n_left,
		 REAL *right,
		 int n_right,
		 REAL *bell,
		 int dir
		 );
void fold_e_o_1s(REAL *coef,
		 REAL *fcn,
		 int numpts,
		 REAL *left,
		 int n_left,
		 REAL *right,
		 int n_right,
		 REAL *bell,
		 int dir
		 );
/* Pmn_construct.c */
REAL Pmncos(int order,
	    int index,
	    REAL phi
	    );
void Pmncos_resave_init(Pmncos_Recurrence_Save *pmn_save,
			REAL *node,
			int numpts
			);
void next_rtsinPmncos(REAL *outfcn,
		      Pmncos_Recurrence_Save *pmn_save,
		      int halfpts,
		      int order,
		      int index
		      );
void Pmncos_resave_destroy(Pmncos_Recurrence_Save *pmn_save
			   );
/* direct1_tofile.c */
void direct1_tofile(int  numpts,
		    int band_limit,
		    REAL *node,
		    REAL *weight,
		    FILE *fp
		    );
void init_direct1(Pm_Direct_d **out,
		  int halfpts,
		  int band_limit,
		  REAL *node
		  );
void dump_direct1(int numpts,
		  int band_limit,
		  REAL *node,
		  REAL *weight,
		  Pm_Direct_d *pmn_matrix,
		  FILE *fp
		  );
/* fast1_tofile.c */
REAL fast1_tofile(int  numpts,
		  int band_limit,
		  REAL *node,
		  REAL *weight,
		  REAL epsilon,
		  FILE *fp
		  );
void init_fast1(Pm_1D **out,
		Fast1_Save *saves_fordump,
		int halfpts,
		int band_limit,
		REAL *node,
		int *numlevels,
		REAL epsilon,
		REAL (*bell_choice) (int,int)
		);
void dump_fast1(int numpts,
		int band_limit,
		int num_levels,
		REAL *node,
		REAL *weight,
		Fast1_Save *saves_fordump,
		Pm_1D *pmn_matrix,
		FILE *fp
		);
/* direct1_transform.c */
void Pm_direct1_analyse(REAL *outcoef,
			Pm_Direct_d *inmat,
			REAL *infcn,
			Pm_1D_Workspace *pmnwork
			);
void Pm_direct1_synthesize(REAL *outfcn,
			   Pm_Direct_d *inmat,
			   REAL *incoef,
			   Pm_1D_Workspace *pmnwork
			   );
/* matbells.c  */
REAL bellmat1(int laplength,
		int t
		);
REAL bellmat2(int laplength,
		int t
		);
REAL bellmat3(int laplength,
		int t
		);
REAL bellmat4(int laplength,
		int t
		);
REAL bellmat5(int laplength,
		int t
		);
REAL bellmat6(int laplength,
		int t
		);
REAL bellmat7(int laplength,
		int t
		);
REAL bellmat8(int laplength,
		int t
		);
REAL bellmat9(int laplength,
		int t
		);
REAL bellmat10(int laplength,
		int t
		);
REAL bellmat11(int laplength,
		int t
		);
REAL bellmat12(int laplength,
		int t
		);
REAL bellmat13(int laplength,
		int t
		);
REAL bellmat14(int laplength,
		int t
		);
REAL bellmat15(int laplength,
		int t
		);
REAL bellmat16(int laplength,
		int t
		);
REAL bellmat17(int laplength,
		int t
		);
REAL bellmat18(int laplength,
		int t
		);
REAL bellmat19(int laplength,
		int t
		);
REAL bellmat20(int laplength,
		int t
		);
/* fast1_transform.c */
void Pm_1d_analyse(REAL *outcoef,
		   Pm_1D *inmat,
		   REAL *infcn,
		   Pm_1D_Workspace *pmnwork
		   );
void Pm_1d_synthesize(REAL *outfcn,
		      Pm_1D *inmat,
		      REAL *incoef,
		      Pm_1D_Workspace *pmnwork
		      );
/* mk_gnode_gweight.c */
void mk_gnode_gweight(REAL *gnode,
		      REAL *gweight,
		      REAL epsilon,
		      int numpts
		      );
void mk_cosi_gnode(REAL *gnode,
		   REAL epsilon,
		   int numpts
		   );
void cosi_gnode_to_gweight(REAL *gweight,
			   REAL *gnode,
			   int numpts
			   );
/* second.c */
double second();
/* getoption.c */
char* getoption(const char* flag,
		int argc,
		char **argv
		);
/* pointwise_ops.c */
void pointwise_multiply(REAL *out,
			REAL *in_1,
			REAL *in_2,
			int numpts
			);
void reflect_evenodd(REAL *even,
		     REAL *odd,
		     REAL *whole,
		     int numpts,
		     int dir
		     );
void remove_rootsin(REAL *out,
		    REAL *in,
		    REAL *node,
		    int numpts
		    );
/* dyadic_trig_gsearch.c */
void dyadic_g_ilist_init(Local_Trig_Interval *ilist[2],
			 int *ints_used,
			 int *coefs_used,
			 REAL **bell_handle,
			 int *bell_length,
			 int **bell_offsets,
			 int halfpts,
			 int num_levels,
			 REAL (*bell_choice) (int,int),
			 int flags
			 );
void ltrigint_init(Local_Trig_Interval *out,
		   int numpts,
		   int left_extra,
		   int right_extra,
		   int fcn_offset,
		   int coef_offset,
		   REAL (*bell_choice)(int,int),
		   int which_transform,
		   int fftw_flags
		   );
void gsearch_to_ilist(Local_Trig_Interval *ilist_out,
		      int *intervals_kept,
		      int max_intervals,
		      Local_Trig_Interval *ilist_in,
		      Dyadic_Gsearch_Save *dy_gs_save,
		      int halfpts
		      );
/* trig_int_ops.c */
void evaluate_intlist(REAL *outfcn,
		      REAL *incoef,
		      Local_Trig_Interval *interval_list,
		      int num_intervals
		      );
void expand_d_intlist(REAL *outcoef,
		      REAL *infcn,
		      Local_Trig_Interval *interval_list,
		      int num_intervals
		      );
void expand_intlist(REAL *outcoef,
		   REAL *infcn,
		   Local_Trig_Interval *interval_list,
		   int num_intervals
		   );
/* count_thresh.c */
int count_thresh(REAL *lambda,
		 int howmany,
		 REAL thresh
		 );
/* init_Pm_1d_workspace */
void init_Pm_1d_workspace(Pm_1D_Workspace *out,
			  int numpts,
			  int band_limit,
			  REAL *node,
			  REAL *weight
			  );

/* */
void ftgaqd(int *nlat,
	    REAL *theta,
	    REAL *wts,
	    REAL *work,
	    int *lwork,
	    int *ierror
	    );
void gnode_to_Pmninit(REAL ***sofar,
		      REAL **ex,
		      REAL **why,
		      int **size,
		      REAL *gnode,
		      int hemipts
		      );

/*************************************************************************/
/******* END function prototypes *****************************************/
/*************************************************************************/

#endif /* LIBFTSH_PRESENT */
