/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "libftsh.h"


#ifndef PI
#define PI 3.14159265358979323846264338328
#endif

/** \file matbells.c

\brief Contains the Matviyenko bells for local trigonometric transforms.

The bells have the form   bell(laplength,t)
   
INPUTS: 
- laplength-- the number of POINTS the bell spills out of the 
  interval.  It also spills in an equal number of
  points.  Thus the bell takes 2*laplength points to 
  rise from 0 to 1.
- t  --     Which point you want to evaluate the bell at.  This
  should range from 0 to 2*laplength-1.  
  bell(laplength,0)=~0 then it rises until 
  bell(laplength,2*laplength-1)=~1.
  The edge of the interval is between points, at 
  t=(laplength-1) +0.5

OUTPUT:   The bell function returns a REAL which is the bell's value
at that point.

The bellmatXX sequence of bells is due to Gregory Matviyenko
'Optimized Local Trigonometric Bases' Research report
YALEU/DCS/RR-1041 July 15, 1994 and Applied and Computational Harmonic
Analysis 3 (1996)
   
The are not standard local cosine bells, but instead come in pairs,
with one bell used for expansion and the other (dual) bell used for
reconstruction.

These particular bells have been optimized after fixing a parameter N
(?smoothness).  Each N gives the best results for some cutoff
epsilon_N.  They are optimized in the sense that for a sin of
arbitrary phase and frequency, they will, on average, yield the fewest
coefficients above epsilon_N

(It's not clear from the paper whether s should begin at 0 or 1.
formula (151) has a g_0 but table 1 starts with g_1.  It should begin
at 0.)

*/

REAL bellcutoff(int laplength,
		int t
		){
  /** bellcutoff is a characteristic function. 

  It is useful for tests when
  we want to quickly disable the effect of the bell */
  if(t<laplength)
    return 0.0;
  else
    return 1.0;
}


/***************************************************************************/


REAL bellmat1(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.1002143947640111085};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<1;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat2(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.1723768006269012949,
                    0.1855148479250006034};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<2;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat3(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2031447668472587192,
                    0.2487749850071917170,
                    0.0475141123801596348};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<3;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat4(int laplength,
		int t
		){
  /* Best for epsilon=1e-3 */
  REAL x,aux;
  int s;
  const REAL g[]={1.2196727213232474166,
		      0.2853161868567129887,
		      0.0789155484618257136,
		      0.0135550276613148530};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<4;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat5(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2299263341780548351,
		    0.3091936162560012950,
		    0.1022547499107077172,
		    0.0270080848154667718,
		    0.0040643028367457815};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<5;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat6(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2368960390161255733,
		    0.3260454684070418003,
		    0.1202144284772571658,
		    0.0392824672307590613,
		    0.0094646390135188831,
		    0.0012540398189263445};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<6;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat7(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2419380540495429195,
		    0.3385842628287665721,
		    0.1344330305316184567,
		    0.0501467090476806012,
		    0.0153116614831942916,
		    0.0033448841387432324,
		    0.0003942039965801714};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<7;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat8(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2457536263723204704,
		    0.3482817014907560773,
		    0.1459560151659402844,
		    0.0596831230559218041,
		    0.0211609941444967639,
		    0.0059658781972807312,
		    0.0011854380315612793,
		    0.0001255469017738763};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<8;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat9(int laplength,
		int t
		){
  /* Best for epsilon=1e-7 */
  REAL x,aux;
  int s;
  const REAL g[]={1.2487411472878298253,
		      0.3560069611425606312,
		      0.1554771240803420597,
		      0.0680536775001451859,
		      0.0267991398186155608,
		      0.0088894805175620360,
		      0.0023126782210685839,
		      0.0004203150366958263,
		      0.0000403732905776057};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<9;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat10(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2511435101979626018,
                    0.3623069592485838316,
                    0.1634729471600794903,
                    0.0754255125823668245,
                    0.0321305559573675648,
                    0.0119592651970663211,
                    0.0036995960643158676,
                    0.0008907386396059369,
                    0.0001489420206098608,
                    0.0000130803773243655};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<10;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat11(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2531171572760240244,
                    0.3675432476836647534,
                    0.1702809341651916143,
                    0.0819484046917260704,
                    0.0371207535134114532,
                    0.0150712511279723392,
                    0.0052752483692935777,
                    0.0015235383929696997,
                    0.0003408100998412218,
		    0.0000527237866578197,
                    0.426301975610614e-5};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<11;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat12(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2547673684254528441,
                    0.3719644997534442686,
                    0.1761463604899294465,
                    0.0877499693755897336,
                    0.0417667654440872536,
                    0.0181579310814718085,
                    0.0069802273692443469,
                    0.0022969056584170223,
                    0.0006209177060710873,
                    0.0001295783756168840,
                    0.0000186407482263636,
                    0.13960350016921e-5};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<12;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat13(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2561675816035252352,
                    0.3757474138575439717,
                    0.1812516131847340424,
                    0.0929368147130266026,
                    0.0460814756936615327,
                    0.0211765890308308775,
                    0.0087672206344614442,
                    0.0031866825949287158,
                    0.0009873442127753202,
                    0.0002505909631277034,
                    0.0000489789669579668,
                    0.658213072318501e-5,
                    0.459014688967825e-6};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<13;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat14(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2573705781037171767,
                    0.3790210170105753012,
                    0.1857350426141542667,
                    0.0975974055372798749,
                    0.0500851855065012145,
                    0.0241011932143827745,
                    0.0105996145856369849,
                    0.0041695173336949991,
                    0.0014340335627767149,
                    0.0004193026849064740,
		    0.0001002259697370695,
		    0.0000184143983824276,
		    0.232126216952405e-5,
		    0.151428882710068e-6};


  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<14;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*****************************/
REAL bellmat15(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2584152648035577478,
                    0.3818817363437593838,
                    0.1897034477160271649,
                    0.1018051162625708735,
                    0.0538010556771918339,
                    0.0269168795850161325,
                    0.0124496130486740615,
                    0.0052243034619146703,
                    0.0019527481969350935,
		    0.0006365376324034804,
		    0.0001760791467702309,
		    0.0000397571893420880,
		    0.688941970620248e-5,
		    0.817633371026949e-6,
		    0.501000810832705e-7};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<15;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*********************************/
REAL bellmat16(int laplength,
		int t
		){
  /* Best for epsilon=1e-13 */
  REAL x,aux;
  int s;
  const REAL g[]={1.2593309500457438031,
		      0.3844030891706786254,
		      0.1932405574881542732,
		      0.1056209666415645249,
		      0.0572526660933526002,
		      0.0296162216589041811,
		      0.0142964336695234923,
		      0.0063326892901278450,
		      0.0025343672760701497,
		      0.0009011351818553216,
		      0.0002789521059156221,
		      0.0000731823780789826,
		      0.0000156526972929684,
		      0.256611058982070e-5,
		      0.287672400786250e-6,
		      0.166167484200405e-7};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<16;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*********************************/
REAL bellmat17(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2601401306174552497,
		    0.3866421087683765354,
		    0.1964129247635806935,
		    0.1090959178203754145,
		    0.0604627480475656409,
		    0.0321967100439478970,
		    0.0161247652829157535,
		    0.0074791070877270307,
		    0.0031697006503196943,
		    0.0012105885859207094,
		    0.0004101047807666002,
		    0.0001208077758892354,
		    0.0000301301370711848,
		    0.612055142365566e-5,
		    0.951935962296931e-6,
		    0.101105205879303e-6,
		    0.552324489986036e-8};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<17;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*********************************/
REAL bellmat18(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2608603612954268205,
		    0.3886437212387816550,
		    0.1992741058079349344,
		    0.1122727446284484620,
		    0.0634525707675491656,
		    0.0346590468664805595,
		    0.0179235156823607947,
		    0.0086505731134735900,
		    0.0038499687030354075,
		    0.0015615520391816049,
		    0.0005698390644760264,
		    0.0001842373381822920,
		    0.0000517518322815702,
		    0.0000122981314895988,
		    0.237837400561446e-5,
		    0.351830571000310e-6,
		    0.354989120817137e-7,
		    0.183937626456071e-8};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<18;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*********************************/
REAL bellmat19(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2615055408626026438,
		    0.3904437975327983132,
		    0.2018676789041392205,
		    0.1151875431220117202,
		    0.0662416923898752204,
		    0.0370059895821450853,
		    0.0196848223338375282,
		    0.0098363945921014495,
		    0.0045670612090700171,
		    0.0019502209236860229,
		    0.0007577056719391676,
		    0.0002645671093113601,
		    0.0000817808996953258,
		    0.0000219482849485672,
		    0.498006589408313e-5,
		    0.918946754735698e-6,
		    0.129595480966266e-6,
		    0.124523708171851e-7,
		    0.613594334772978e-9};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<19;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}

/*********************************/
REAL bellmat20(int laplength,
		int t
		){
  REAL x,aux;
  int s;
  const REAL g[]={1.2620868165186817200,
		    0.3920713262058702161,
		    0.2042294606490294914,
		    0.1178709389362636122,
		    0.0688479112856821096,
		    0.0392415664712484188,
		    0.0214032823468272275,
		    0.0110278550409693667,
		    0.0053136548560657154,
		    0.0023726050667205839,
		    0.0009726951425515420,
		    0.0003624288874381589,
		    0.0001212704239660943,
		    0.0000359012169826721,
		    0.922285969549705e-5,
		    0.200203409841185e-5,
		    0.353205918170597e-6,
		    0.475881033651720e-7,
		    0.436427442377405e-8,
		    0.204997258761082e-9};

  x=(t+0.5)/(2.0*laplength) -0.5;

  aux=1.0;
  for(s=0;s<20;s++)
    aux+=g[s]*sin((s+0.5)*PI*x);
  aux*=0.5;

  return aux;
}


