/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file mk_gnode_gweight.c

\brief Construct the Gaussian nodes and weights.

Summary:
- mk_gnode_gweight drives the construction.
- mk_cosi_gnode constructs the nodes.
- cosi_gnode_to_gweight constructs the weights from the nodes.

*/

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

#ifndef LNOF2
#define LNOF2 0.693147180559945309417232121458
#endif

void mk_gnode_gweight(REAL *gnode,
		      REAL *gweight,
		      REAL epsilon,
		      int numpts
		      ){
  /** This routine constructs the gaussian nodes and weights.
     
  We make cosine inverse of the ordinary gaussian nodes.

  INPUTS:
  - gnode -- the array of length numpts in which to put the 
    gaussian nodes. Memory must already be allocated. 
  - gweight --  the array in which to put the gaussian weights.
    Memory must already be allocated.
  - epsilon -- the precision to which to compute the gaussian nodes.
  - numpts --  the order of the nodes and weights we compute
    this is the length of gnode and gweight.

  OUTPUTS:
  - gnode -- the numpts gaussian node values are written on gnode
    We make cosine inverse of the ordinary gaussian nodes.
  - gweight -- numpts values of gweight are filled in

  */

#if ENTEREXIT
  printf("entering mk_gnode_gweight\n");
#endif

  /* call the gaussian node maker ****/
  mk_cosi_gnode(gnode,epsilon,numpts);

  /* call the node to weight converter **/
  cosi_gnode_to_gweight(gweight,gnode,numpts);
  
#if ENTEREXIT
  printf("entering mk_gnode_gweight\n");
#endif

}

void mk_cosi_gnode(REAL *gnode,
		    REAL epsilon,
		    int numpts
		    ){
  /* Here we find the numpts zeroes of P^0_(2*hemipts)(cos(_)) on [0,PI]. 

  These zeros give us cos^{-1} of the ordinary gaussian nodes.

  We use a primitive bisection scheme. 
  Probably a newton's method is better.
  
  INPUTS: 
  - gnode -- the array of length numpts in which to put the 
    gaussian nodes. Memory must already be allocated.
  - epsilon -- the accuracy to which we compute the nodes
  - numpts --  the order of the nodes we compute

  OUTPUTS: the numpts gaussian node values are written on gnode

  NOTES: 
  - This routine is slow, but really tough. The user is encouraged to 
    compute the nodes once and save to disk
  -  minimum value for numpts.
  */

  int i,lowersign;
  REAL lower,upper,mid;

#if ENTEREXIT
  printf("entering mk_cosi_gnode\n");
#endif
  
  /* make sure numpts logical */
  assert(numpts>0);

  /* a dumb case, but we can handle it*/
  if(numpts==1){
    gnode[0]=PI/2;
    return;
  }

  /* note that numpts is both the number of points 
     AND the index=n to use for P_n****/    

  /* the first zero is the hardest to get.  
    We take baby steps to make sure we don't miss it.*/
  upper=epsilon;
  /* our sign convention has P^m_n start positive.
   Since frequency DECREASES as we move, we know that doubling
   upper will never leap us over 2 zeroes*/
  while(Pmncos(0,numpts,upper)>0)
    upper*=2;
  /* set lower to the previous one, which will still have positive value*/
  lower=upper*0.5;
  
  /* we have now bracketed the first zero and can bisect**/
  /* we only keep track of the sign at the lower end **/
  lowersign=1;
  while(upper-lower>epsilon){
    /* gap still too big */

    /* bisect */
    mid=(upper+lower)*0.5;
    if(Pmncos(0,numpts,mid)*lowersign > 0)
      /* move lower to mid, since they have the same sign */
      lower=mid;
    else
      /* move upper */
      upper=mid;
  }

  /* set the first node */
  gnode[0]=(lower+upper)*0.5;
  

  /* since frequency DECREASES, stepping by the distance between
    previous zeroes is safe.*/
  /* get safe (small) starting valuse for upper */
  upper=gnode[0]*2.0;
  /* lower changes sign */
  lowersign*=-1;

  while(Pmncos(0,numpts,upper)*lowersign>0)
    /* have not bracket a zero yet */

    /* move upper */
    upper+=gnode[0];

  /* set lower to the previous value of upper, which has sign lowersign*/ 
  lower=upper-gnode[0];

  /*now we've bracketed the second zero.*/


  /*********** general step *****************/
  /* we compute just under half the nodes.
    we then have a couple of special cases, 
    and then some computed by symmetry ******/

  /* see comments above for procedure */
  for(i=1;i<(numpts>>1)-1;i++){
    while(upper-lower>epsilon){
      mid=(upper+lower)*0.5;
      if(Pmncos(0,numpts,mid)*lowersign > 0)
	lower=mid;
      else
	upper=mid;
    }
    
    gnode[i]=(lower+upper)*0.5;
    /* since frequency DECREASES, stepping by the distance between
      previous zeroes is safe.*/
    upper=gnode[i]+0.5*(gnode[i] -gnode[i-1]);
    lowersign*=-1;
    while(Pmncos(0,numpts,upper)*lowersign>0)
      upper+=0.5*(gnode[i] -gnode[i-1]);
    lower=upper-0.5*(gnode[i] -gnode[i-1]);
  }

  /* zero [numpts/2-1], already bracketed.
    we do this as a special case because trying to bracket 
    zero [numpts/2] may cause problems**/

  while(upper-lower>epsilon){
    mid=(upper+lower)*0.5;
    if(Pmncos(0,numpts,mid)*lowersign > 0)
      lower=mid;
    else
      upper=mid;
  }
  gnode[numpts/2-1]=(lower+upper)*0.5;


  /* if numpts is odd then we have a zero at PI/2**/
  if(numpts%2)
    gnode[numpts/2]=PI/2.0;

  /* get the rest by reflection across PI/2 *********/
  for(i=0;i<numpts/2;i++)
    /* safe whether numpts even or odd */

    gnode[numpts-i-1]=PI-gnode[i];

#if ENTEREXIT
  printf("exiting mk_cosi_gnode\n");
#endif

}

void cosi_gnode_to_gweight(REAL *gweight,
			 REAL *gnode,
			 int numpts
			 ){
  /** This routine computes the gaussian weights that go with a set of 
      gaussian nodes.

      The nodes input here are \\(\\cos^{-1}(\\text{the usual nodes} )\\).

      INPUTS: 
      - gweight --  the array in which to put the gaussian weights.
        Memory must already be allocated.
      - gnode -- the gausian nodes, theta values between 0 and PI
        This is \\(\\cos^{-1}(\\text{the usual nodes} )\\).
      - numpts --  the order of the weights we compute
        this is the length of gnode and gweight.

      OUTPUTS: numpts values of gweight are filled in

      NOTES: 
      - We use the Christoffel function method, mainly because we can avoid
        derivatives.
      - Do I use normalized Pmn?
  */

  register int point,i;
  register REAL oneback,twoback,sum,ex;
  REAL sofar[2];

#if ENTEREXIT
  printf("entering cosi_gnode_to_gweight\n");
#endif

  /* The weight is computed by the Christofffel function method via
                (N-1   (         ) 2 ) -1                             
    weight   = ( SUM  ( P (node ) )   )             
          i     (j=0   ( j     i )   )                                  
	  */

  /* note the degree is the same as numpts*/

  for(point=0;point<(numpts+1)/2;point++){
    /* loop through half (or slightly more if numpts odd) gaussian nodes */
    
    /* determine the equivalent x value */
    ex=cos(gnode[point]);
    
    /* set the starting constant P_0*/
    twoback=sqrt(0.5);
    
    /* initialize the recurrence */
    sofar[0]=twoback;

    /* start the running sum */
    sum=(sofar[0])*(sofar[0]);

    /* another special case, requires some initialization*/
    /* this is the oneback recurrence coefficient */
    oneback=sqrt(3.0);
    sofar[1]=oneback* sofar[0]*ex;
    
    /* add to the running sum */
    sum+=(sofar[1])*(sofar[1]);
    
    /* general case */
    for(i=2;i<numpts;i++){
      /* This is the main recurrence.  First we compute the coefficients,
	 then compute and load in the values */
      oneback=2.0*sqrt((1.0-0.5/i)*(1.0+0.5/i));
      twoback=-sqrt((1.0-1.0/i)*(1.0-1.0/i)
		    *(1.0+4.0/(2.0*i-3.0)));
      sofar[i%2]=oneback*ex*sofar[(i-1)%2]+twoback*sofar[(i-2)%2];
    
      /* add to the running sum */
      sum+=(sofar[i%2])*(sofar[i%2]);
      
    }
    
    /* set the weight itself */
    gweight[point]=1.0/sum;
     
  }

  /* set the othe half (or less if numpts odd) by symmetry */
  for(point=0;point<numpts/2;point++)
    gweight[numpts-point-1]=gweight[point];


#if ENTEREXIT
  printf("exiting cosi_gnode_to_gweight\n");
#endif
}






