/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file pointwise_ops.c

\brief Contains functions to do pointwise operations.

Summary:
- pointwise_multiply multiplies two arrays.
- reflect_evenodd forms even and odd parts of an array.
- remove_rootsin divides an array by sqrt(sin(phi))

*/

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void pointwise_multiply(REAL *out,
			REAL *in_1,
			REAL *in_2,
			int numpts
			){
  /** Pointwise multiply two vectors to form the third.

  INPUTS: 
  - out -- a vector of length numpts in which to put the output.
    May be the same as in_1 or in_2
  - in_1 -- the first input vector, of length numpts
  - in_2 -- the second input vector, of length numpts
  - numpts --  the length of the vectors

  OUTPUTS: out -- is written onto; out[i]=in_1[i]*in_2[i];

  */

  register int i;
#if ENTEREXIT
  printf("entering pointwise_multiply\n");
#endif

  for(i=0;i<numpts;i++){
    /* loop through points */
      
    /* multiply */
    out[i]=in_1[i]*in_2[i];
  }
   
#if ENTEREXIT
  printf("exiting pointwise_multiply\n");
#endif
}

void reflect_evenodd(REAL *even,
		     REAL *odd,
		     REAL *whole,
		     int numpts,
		     int dir
		     ){
  /** Form the even and odd parts of the input vector.
     
  It can also reverse the process.

  INPUTS: 
  - even -- In forward mode, the even portion will be put here.
    In reverse mode, the even portion is read from here.
    Of length (numpts+1)/2.
  - odd -- In forward mode, the odd portion will be put here.
    In reverse mode, the odd portion is read from here.
    Of length numpts/2.
  - whole -- in forward mode, the fcn to read from.
    In reverse mode, where to put the function.
    Of length numpts.
  - numpts --  the length of whole and twice the length of even and odd. 
    If numpts is odd, the extra point goes to even.
  - dir -- the direction to go in
    + dir==1  fold whole into even/odd
    + dir==-1 unfold even/odd into whole
    + else error
    
  OUTPUTS: 
  - In forward mode even and out are written onto, with the even/odd
    reflections of whole: 
    + even[i]=whole[i]+whole[numpts-1-i]
    + odd[i]=whole[i]-whole[numpts-1-i]

    If the number of points is odd, the center point is copied
    into even, and omitted from odd.
  - in reverse mode undoes this process

  NOTES:
  - the exception handling for odd numpts might slow this down
  */

  register int i,halfpts;
#if ENTEREXIT
  printf("entering reflect_evenodd\n");
#endif

  /* note the half-length, rounding down */
  halfpts=numpts/2;

  if(dir==1){
    /* whole --> even odd */

    for(i=0;i<halfpts;i++){
      /* loop through points */
      
      /* even reflection */
      even[i]=whole[i]+whole[numpts-1-i];

      /* odd reflection */
      odd[i]=whole[i]-whole[numpts-1-i];
    }
    
    /* check for middle point */
    if(numpts%2)
      /* have middle, copy to even */
      even[halfpts]=whole[halfpts];
      
  }else if(dir==-1){
    /* unreflect even/odd --> whole */

    
    for(i=0;i<halfpts;i++){
      /* loop through points */

      /* first half, add even and odd */
      whole[i]=even[i]+odd[i];

      /*second half, subtract odd from even */
      whole[numpts-1-i]=even[i]-odd[i];
    }

    /* check for middle point */
    if(numpts%2)
      /* have middle, copy from even */
      whole[halfpts]=even[halfpts];

  }
  else
    /*************** error *******************/
    fprintf(stderr,"reflect_evenodd: WARNING: illegal dir==%i\n",dir);

#if ENTEREXIT
  printf("exiting reflect_evenodd\n");
#endif
}

void remove_rootsin(REAL *out,
		    REAL *in,
		    REAL *node,
		    int numpts
		    ){
  /** Divide the vector in by sqrt(sin( node values)).

  INPUTS: 
  - out -- a pointer to where to put the outputs.
    May be the same as in.
  -  in -- the input zector, of length numpts
  - node -- and array holding the sampling nodes, corresponding
    to numpts physical points. Should be in (0,PI)
  - numpts -- the number of sample points. (plus 1 if odd)

  OUTPUTS : out -- is written onto out[i]=in[i]/sqrt(sin(node[i]))

  NOTES:
  - For safety we do sqrt(fabs(sin(node[i])))
    This is so if the node value is out of range the routine doesn't crash
  - We could still crash by dividing be zero   
  */

  register int i;

#if ENTEREXIT
  printf("entering remove_rootsin\n");
#endif

  for(i=0;i<numpts;i++)
    out[i]=in[i]/sqrt(fabs(sin(node[i])));    

#if ENTEREXIT
  printf("exiting remove_rootsin\n");
#endif

}






  





  
