/*
 * Copyright (c) 2000, 2012 Martin J. Mohlenkamp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** \file trig_int_ops.c

\brief Operations based on Local_Trig_Interval struct.

Summary:
- evaluate_intlist evaluate on a list of intervals (dual bell).
- expand_intlist expand on a list of intervals.
- expand_d_intlist expand on a list of intervals using dual bell.

*/

#include "libftsh.h"

#ifndef ENTEREXIT
#define ENTEREXIT 0
#endif

void evaluate_intlist(REAL *outfcn,
		      REAL *incoef,
		      Local_Trig_Interval *interval_list,
		      int num_intervals
		      ){
  /** This routine does a LCT/DCT evaluation based on a list of intervals.

  INPUTS:  
  - outfcn --  where to put the function constructed.
    We ADD onto outfcn
  - incoef     -  the coefficients to evaluate. destroyed
  - interval_list-- An array of Local_Trig_Interval's,
    encoding what intervals to evaluate,
    how to evaluate,
    and where to put the result
  - num_intervals - the number of intervals

  OUTPUTS: outfcn is filled in. incoef is destroyed

  NOTES: 
  - evaluates using the dual bell

  */
    register int i,j;
#if ENTEREXIT
  printf("entering evaluate_intlist\n");
#endif

  for(i=0;i<num_intervals;i++){
    /* loop through the intervals */

    /*printf("i=%i out of %i\n",i,num_intervals);*/

    if(interval_list[i].which_transform==0){
      /********* do the LCT ****************/

      /* LCT the interval in incoef, in place */
      lctw(incoef+interval_list[i].coef_offset,
	   &(interval_list[i].transform_plan));

      /* fold into outfcn, even on left, odd on right */
      fold_e_o_1s(incoef+interval_list[i].coef_offset,
		  outfcn+interval_list[i].fcn_offset,
		  interval_list[i].numpts,
		  outfcn+interval_list[i].fcn_offset-1,
		  interval_list[i].left_extra,
		  outfcn+interval_list[i].fcn_offset+interval_list[i].numpts,
		  interval_list[i].right_extra,
		  interval_list[i].dualbell,-1);

    }
    else if (interval_list[i].which_transform==1){
      /*********** do the DCT *******************/

      /* inverse DCT the interval in incoef, in place */
      dctw(incoef+interval_list[i].coef_offset,
	   &(interval_list[i].transform_plan),-1);

      /* fold into outfcn, even on left, even on right */
      fold_e_e_1s(incoef+interval_list[i].coef_offset,
		  outfcn+interval_list[i].fcn_offset,
		  interval_list[i].numpts,
		  outfcn+interval_list[i].fcn_offset-1,
		  interval_list[i].left_extra,
		  outfcn+interval_list[i].fcn_offset+interval_list[i].numpts,
		  interval_list[i].right_extra,
		  interval_list[i].dualbell,-1);

    }
    else if(interval_list[i].which_transform==9){
      /* simple copy, no bell **/

      for(j=0;j<interval_list[i].transform_plan.numpts;j++){
	/* loop through the points in the interval */

	/* copy the value */
	outfcn[interval_list[i].fcn_offset+j]=
	  incoef[interval_list[i].coef_offset+j];
      }
    }
    else
      /*************** error *******************/
      fprintf(stderr,"evaluate_intlist: unknown transform "
	      "interval_list[%i].which_transform=%i\n",
	      i,interval_list[i].which_transform);
  }

#if ENTEREXIT
  printf("exiting evaluate_intlist\n");
#endif
}

void expand_intlist(REAL *outcoef,
		   REAL *infcn,
		   Local_Trig_Interval *interval_list,
		   int num_intervals
		   ){
  /** This routine does a LCT/DCT expansion on a list of intervals.

  INPUTS:  
  - outcoef --  where to put the coefficients constructed.
  - infcn     -  the fcn to expand
  - interval_list-- An array of Local_Trig_Interval's,
    encoding what intervals to expand, how to expand,
    and where to put the result
  -num_intervals - the number of intervals

  OUTPUTS: outcoef is filled in. infcn is untouched

  */
  
  register int i,j;
#if ENTEREXIT
  printf("entering expand_intlist\n");
#endif

  for(i=0;i<num_intervals;i++){
    /* loop through the intervals */

    /*printf("i=%i out of %i\n",i,num_intervals);*/

    if(interval_list[i].which_transform==0){
      /********* do the LCT ****************/

      /* fold into outcoef, even on left, odd on right */
      fold_e_o_1s(outcoef+interval_list[i].coef_offset,
		  infcn+interval_list[i].fcn_offset,
		  interval_list[i].numpts,
		  infcn+interval_list[i].fcn_offset-1,
		  interval_list[i].left_extra,
		  infcn+interval_list[i].fcn_offset+interval_list[i].numpts,
		  interval_list[i].right_extra,
		  interval_list[i].bell,1);

      /* LCT the interval in outcoef, in place */
      lctw(outcoef+interval_list[i].coef_offset,
	   &(interval_list[i].transform_plan));
    }
    else if (interval_list[i].which_transform==1){
      /*********** do the DCT *******************/

      /* fold into outcoef, even on left, even on right */
      fold_e_e_1s(outcoef+interval_list[i].coef_offset,
		  infcn+interval_list[i].fcn_offset,
		  interval_list[i].numpts,
		  infcn+interval_list[i].fcn_offset-1,
		  interval_list[i].left_extra,
		  infcn+interval_list[i].fcn_offset+interval_list[i].numpts,
		  interval_list[i].right_extra,
		  interval_list[i].bell,1);

      /* DCT the interval in outcoef, in place */
      dctw(outcoef+interval_list[i].coef_offset,
	   &(interval_list[i].transform_plan),1);
    }
    else if(interval_list[i].which_transform==9){
      /* simple copy, no bell **/

      for(j=0;j<interval_list[i].transform_plan.numpts;j++){
	/* loop through the points in the interval */

	/* copy the value */
	outcoef[interval_list[i].coef_offset+j]=
	  infcn[interval_list[i].fcn_offset+j];
      }
    }
    else
      /*************** error *******************/
      fprintf(stderr,"expand_intlist: unknown transform "
	      "interval_list[%i].which_transform=%i\n",
	      i,interval_list[i].which_transform);
  }

#if ENTEREXIT
  printf("exiting expand_intlist\n");
#endif
}

void expand_d_intlist(REAL *outcoef,
		      REAL *infcn,
		      Local_Trig_Interval *interval_list,
		      int num_intervals
		      ){
  /** This routine does a LCT/DCT expansion on a list of intervals.

  This version expands using the DUAL bell. 
  Otherwise it is just like expand_intlist.

  See expand_intlist for a full description.
  */
  
  register int i;
  register REAL *bell_swap;

#if ENTEREXIT
  printf("entering expand_d_intlist\n");
#endif

  for(i=0;i<num_intervals;i++){
    /* loop through the intervals */

    /* swap bell and dualbell */
    bell_swap=interval_list[i].bell;
    interval_list[i].bell=interval_list[i].dualbell;
    interval_list[i].dualbell=bell_swap;

  }

  /* call expand_intlist */
  expand_intlist(outcoef,infcn,interval_list,num_intervals);


  for(i=0;i<num_intervals;i++){
    /* loop through the intervals */

    /* RE-swap bell and dualbell */
    bell_swap=interval_list[i].bell;
    interval_list[i].bell=interval_list[i].dualbell;
    interval_list[i].dualbell=bell_swap;

  }



#if ENTEREXIT
  printf("exiting expand_d_intlist\n");
#endif
}







